/***********************************************************************/
/*                                                                     */
/*                           Objective Caml                            */
/*                                                                     */
/*         Xavier Leroy and Damien Doligez, INRIA Rocquencourt         */
/*                                                                     */
/*  Copyright 1996 Institut National de Recherche en Informatique et   */
/*  Automatique.  Distributed only by permission.                      */
/*                                                                     */
/***********************************************************************/

/* $Id: asmstartup.c,v 1.1.1.1 1998/04/20 11:32:09 lefessan Exp $ */

/* Start-up code */

#include <stdio.h>
#include <stdlib.h>
#include "fail.h"
#include "gc.h"
#include "gc_ctrl.h"
#include "misc.h"
#include "mlvalues.h"
#include "sys.h"
#ifdef HAS_UI
#include "ui.h"
#endif

int with_native = 1;

header_t atom_table[256];
char * static_data_start, * static_data_end;
char * code_area_start, * code_area_end;

/* Initialize the atom table and the static data and code area limits. */

struct segment { char * begin; char * end; };

static void minmax_table(struct segment *table, char **min, char **max)
{
  int i;
  *min = table[0].begin;
  *max = table[0].end;
  for (i = 1; table[i].begin != 0; i++) {
    if (table[i].begin < *min) *min = table[i].begin;
    if (table[i].end > *max)   *max = table[i].end;
  }
}
  
static void asm_init_atoms(void)
{
  int i;
  extern struct segment caml_data_segments[], caml_code_segments[];

  for (i = 0; i < 256; i++) atom_table[i] = Make_header(0, i, White);
  minmax_table(caml_data_segments, &static_data_start, &static_data_end);
  minmax_table(caml_code_segments, &code_area_start, &code_area_end);
}

extern void caml_start_program (void);
extern void init_signals (void);

/*
void caml_startup(char **argv)
{
  caml_main(argv);
}
*/

void asm_caml_main(char **argv)
{
  asm_init_atoms();
  init_signals();
  caml_start_program();
}
