(***********************************************************************)
(*                                                                     *)
(*                       Jocaml                                        *)
(*                                                                     *)
(*      Fabrice Le Fessant, projet SOR/PARA INRIA Rocquencourt         *)
(*                                                                     *)
(*                 mail: fabrice.le_fessant@inria.fr                   *)
(*                 web:  http://www-sor.inria.fr/~lefessan             *)
(*                                                                     *)
(*  Copyright 1997 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(*                                                                     *)
(*         File written in the Objective-CAML language                 *)
(***********************************************************************)

(* open Typedtree *)
type pattern = int
type process = int


type automate = pattern list * (bool array * process) list


(* le nbre de messages dans un seul automate est limite a 15 pour
   tenir dans un seul int ( 31 sur une arch 64 )
*)

type state = 
{
  state: int;
  mesgs: int;
  trans: state option array;
  mutable eq: (state * process) option
}


let states = ref 0
let nouvel_etat = ref false

let new_state msgs n =
  let rec state =
    {
      state = !states;
      mesgs = msgs;
      trans = Array.create n None;
      eq = None
    }
  in
    nouvel_etat := true;
    states := !states + 1;
    state


let create_automate automata =
  let (patterns,lignes) = automata in
  let n = List.length patterns in
    (* le premier etat *)
  let etat0 = new_state 0 n in
  let stats = ref [ (0,etat0) ] in
  let goods = ref [] in
    (* les etats de declenchement *)
  let _ = List.iter (
      function (tab,proc) ->
        let v = ref 0 in
        let c = ref 1 in
          for i = 0 to n-1 do
            if tab.(i) then
              v := !v lor !c;
            c := !c lsl 2
          done;
          if List.mem_assoc !v !stats then 
            begin
              print_string "Process never triggered"; print_newline ()
            end
          else
            begin
              let etat = new_state !v n in
                etat.eq <- Some (etat0,proc);
                stats := (!v,etat) :: !stats;
                goods := (!v,proc) :: !goods
            end
  ) lignes in
  let new_stats = ref stats in
  let _ =
    while !nouvel_etat do
      let new_created_stats = ref [] in
        nouvel_etat := false;
        List.iter (
          function (msgs,etat) -> (* pour chaque etat *)
            for i = 0 to n-1 do (* pour chaque message *)
              if etat.trans.(i) = None then
                let old_v = (msgs land (3 lsl (2*i))) lsr (2*i) in
                  match old_v with
                      0 -> 
                    | 1 ->
                    | 2 ->
                    | _ (* 3 *) ->
            done) !new_stats;
        new_stats := new_created_stats
    done in
    !goods

let _ = create_automate ([0;1;2],
                         [[|true;true;false|],0;
                          [|false;true;false|],1;
                          [|true;false;true|],2
                         ])
;;

