/***********************************************************************/
/*                                                                     */
/*                           Join-Calculus                             */
/*                                                                     */
/*     Fabrice Le Fessant, projet PARA/SOR, INRIA Rocquencourt         */
/*                                                                     */
/*  Copyright 1998 Institut National de Recherche en Informatique et   */
/*  Automatique.  Distributed only by permission.                      */
/*                                                                     */
/***********************************************************************/

#include <string.h>
#include "config.h"
#include "fail.h"
#include "misc.h"
#include "memory.h"
#include "alloc.h"
#include "mlvalues.h"
#include "stacks.h"
#include "callback.h"
#include "instruct.h"
#include "fix_code.h"
#include "modules.h"

/* To unable module migration */

#define INITIAL_MODULES_TABLE_SIZE 32
value module_table;
int32 module_table_size,next_slot;
value module_begin,module_end;

void init_modules(void)
{
  mlsize_t i;

  module_table_size = INITIAL_MODULES_TABLE_SIZE;
  module_table = alloc_shr(module_table_size,0);
  module_begin = (value) stat_alloc(module_table_size * sizeof(value));
  module_end = (value) stat_alloc(module_table_size * sizeof(value));
  
  for(i=0; i<module_table_size; i++){
      initialize(&Field (module_table, i), Val_unit);
  }
  next_slot = 0;
}

void realloc_modules(void)
{
  mlsize_t requested_size,i;
  value new_module_table;
  value new_begin;
  value new_end;
  
  requested_size = module_table_size*2;
  gc_message ("Growing modules table to %lu entries.\n", requested_size);
  new_module_table = alloc_shr(requested_size, 0);
  
  for (i = 0; i < module_table_size; i++){
    initialize(&Field(new_module_table, i), Field(module_table, i));
  }
  
  for (i = module_table_size; i < requested_size; i++){
    initialize(&Field (new_module_table, i), Val_unit);
  }

  new_begin = (value) stat_alloc( requested_size * sizeof(value));
  for (i = 0; i < module_table_size; i++){
    Field(new_begin,i) = Field(module_begin,i);
  }
  stat_free((char*)module_begin);
  module_begin = new_begin;

  new_end   = (value) stat_alloc( requested_size * sizeof(value));
  for (i = 0; i < module_table_size; i++){
    Field(new_end,i) = Field(module_end,i);
  }
  stat_free((char*)module_end);
  module_end = new_end;

  module_table = new_module_table;
  module_table_size = requested_size;
}
    
int32 find_free_slot(void)
{
  int32 i;

  if(next_slot ==  module_table_size)    realloc_modules();
  i = next_slot;
  next_slot++;
  return i;
}

value module_prepare_install(value nbr) /* ML */
{
  while(next_slot+Int_val(nbr) > module_table_size)
    realloc_modules();

  return Val_unit;
}

value add_module=Val_unit;
value set_add_module(value f)   /* ML */
{
  add_module = f;
  register_global_root(&add_module);

  return Val_unit;
}

void begin_module(code_t pc_code, value initial_data)
{
  int32 module_slot,i;
  code_t begin_code;
  value reloc;

  Begin_root(initial_data);
  module_slot = find_free_slot();
  End_roots();
  DPRINT2("Install in slot %d module :%s \n",module_slot,
	  String_val(Field(initial_data,MODULE_NAME)));
  Field(initial_data,MODULE_SLOT) = Val_int(module_slot);
  Tag_val(initial_data) = Reloc_tag;

  Modify(&Field(module_table,module_slot),initial_data);
  
  DPRINT1("offset %d\n",Int_val(Field(initial_data,OFFSET)));
  begin_code = pc_code-Int_val(Field(initial_data,OFFSET));
  Field(initial_data,_BEGIN_CODE) = Val_unit;
  Field(initial_data,_END_CODE) = Val_unit;
  
  Field(module_begin,module_slot) = (value) begin_code;
  Field(module_end,module_slot) = 
    (value)(begin_code+Int_val(Field(initial_data,LONGUEUR)));
  DPRINT2("Code : %lx - %lx\n",Field(module_begin,module_slot),
	  Field(module_end,module_slot));

  /* relocation of module bytecode */
  reloc = Field(initial_data,RELOC);
  if(Is_block(reloc)){
    for(i=0;i<Wosize_val(reloc);i++){
      DPRINT1("Modifying : %p\n",begin_code+Int_val(Field(reloc,i)));
      *(begin_code+Int_val(Field(reloc,i))) = module_slot;
    }
  }

  if(!(add_module==Val_unit)){
    callback2(add_module,initial_data,module_table);
  }
}


value find_module=Val_unit;
value set_find_module(value f)   /* ML */
{
  find_module = f;
  register_global_root(&find_module);

  return Val_unit;
}
value find_prim=Val_unit;
value set_find_prim(value f)   /* ML */
{
  find_prim = f;
  register_global_root(&find_prim);

  return Val_unit;
}

void resolv_module(code_t pc,int32 opcode,int offset)
{
  value module;

  module = Field(module_table,pc[0]);
  DPRINT1("Find : %s\n",String_val(Field(module,
				      pc[offset]+MODULE_DATA_SIZE)));
  module = Field(module_table,Int_val(callback(find_module,Field(module,
				      pc[offset]+MODULE_DATA_SIZE))));
  DPRINT1("Found : %s\n",String_val(Field(module,MODULE_NAME)));
  /* c'est peut-etre un DUMMY module ! */
  pc[-1] = THREADED(NOP);
  pc[0] = THREADED(opcode);
  pc[offset] = (int32) (Int_val(Field(module,INTERFACE)));
}

void resolv_prim(code_t pc,int32 opcode,int offset)
{
  value prim;

  prim = callback(find_prim,Field(Field(module_table,pc[0]),
				  pc[offset]+MODULE_DATA_SIZE));
  pc[-1] = THREADED(NOP);
  pc[0] =  THREADED(opcode);
  pc[offset] = (int32) (Int_val(prim));
}

code_t resolv_codeptr(code_t pc)
{
  value module;
  
  DPRINT2("RESOLV CODEPTR %d offset %d\n",pc[0],pc[1]);
  module = Field(module_table,pc[0]);
  if(Field(module_begin,pc[0]) == (value) NULL)
    { failwith("Unresolved code ptr"); }
  return ((code_t)(Field(module_begin,pc[0])))+pc[1];
}

/*
  Installe un nouveau module dans la table des modules et prepare le 
  bytecode pour etre execute.
*/

value install_module(value initial_data,value module)  /* ML */
{
  int32 module_slot,i,code_size;
  code_t begin_code;
  value reloc;

  if(Field(module,LONGUEUR) != Val_int(0))  /* ce module est deja present */
    return Val_unit;

  module_slot = Int_val(Field(module,MODULE_SLOT));
  for(i=2;i<Wosize_val(initial_data);i++)
    Modify(&Field(module,i),Field(initial_data,i));

  Field(module,INTERFACE) = Val_int(-1); /* dummy module for GETGLOBAL ... */
  /* relocation du bytecode */
  code_size = 4*Int_val(Field(module,LONGUEUR));
  DPRINT2("codesize : %d chars bytecode %d values\n",code_size,Wosize_val(Field(module,BYTECODE))); 
  begin_code = (code_t) stat_alloc(code_size);
  Field(module,_BEGIN_CODE) = Val_unit;
  Field(module,_END_CODE) = Val_unit;

  Field(module_begin,module_slot) = (value)begin_code;
  Field(module_end,module_slot) = (value)(begin_code+(code_size/4));

  DPRINT2("Code : %lx - %lx\n",Field(module_begin,module_slot),
	  Field(module_end,module_slot));

  memcpy(begin_code,String_val(Field(module,BYTECODE)),code_size);
#ifdef ARCH_BIG_ENDIAN
  fixup_endianness(begin_code, code_size);
#endif
#ifdef THREADED_CODE
  thread_code(begin_code, code_size);
#endif
  reloc = Field(module,RELOC);
  if(Is_block(reloc)){
    for(i=0;i<Wosize_val(reloc);i++){
      *(begin_code+Int_val(Field(reloc,i))) = module_slot;
    }
  }

  return Val_unit;
}

/*
  Recherche si v est un pointeur de code.
    Retourne Val_unit si ce n'est pas un pointeur de code.
    Retourne le module auquel il appartient si c'est un pointeur de code.
       Dans ce cas, offset est modifie en fonction de l'offset depuis le
         debut du code du module.
*/


value find_code_ptr (code_t v,int32 *offset)
{
  mlsize_t i;
  value module;

  DPRINT1("Looking for code ptr of %p\n",v);
  if(*v == THREADED(CODEPTR) || *v == THREADED(REPLACECODEPTR)){
    *offset = v[2];
    DPRINT("Code ptr is a forwarder\n");
     return Field(module_table,v[1]);
  }  
 
  for(i=0;i<next_slot;i++){
    if(Field(module_begin,i) != (value)NULL &&
       v >= (code_t) Field(module_begin,i) &&
       v < (code_t) Field(module_end,i) ){
      
      module = Field(module_table,i);
      *offset =  v - (code_t) Field(module_begin,i);
      DPRINT2("find module %s with offset %d\n",
              String_val(Field(module,MODULE_NAME)),*offset);
      return module;
      
    }
  }
  
  return Val_unit;
}

/*
  On occupe un slot de la table des modules avec une donnee non
removable (a moins de finaliser le morceau de code qui pointe vers elle).
*/

extern obj_block(value tag,value size);

value install_dummy_module(value name,value size) /* ML */
{
  value module, module_slot;

  Begin_root(name);
  module = obj_block(Val_int(Reloc_tag),size);
  Assign(Field(module,MODULE_NAME),name);
  End_roots();
  
  Begin_root(module);
  module_slot = find_free_slot();
  End_roots();
  
  Field(module,MODULE_SLOT) = Val_int(module_slot);
  Assign(Field(module_table,module_slot),module);
  Field(module_begin,module_slot) = (value)NULL;
  Field(module_end,module_slot) = (value)NULL;
  
  return module;
}


#include "stub.h"
#include <stdio.h>
void failed_message(char *msg,char * expr,char *file, int line)
{
  fprintf (stderr, "%s: Assertion failed: %s; file %s; line %d\n",
           msg, expr, file, line);
  exit (100);
}


void good_stub(value stub,char *msg,char *file,int line)
{
  value tmp;
  int tag;
  
  fprintf(stderr,"Good stub(%s:%d(%s)) : %x\n",file,line,msg,stub);
  if(Is_blue_hd(Hd_val(stub)))
    failed_message(msg,"Stub is a forwarded",file,line);
    
  if(!Is_block(stub))
    failed_message(msg,"Stub is a block",file,line);
    /* le tag */
  tag = Tag_val(stub);
  
  if(tag!=Thread_tag && tag != Location_tag && tag!=Automata_tag
     && tag!=Object_tag
     )
    failed_message(msg,"Stub tag",file,line);
  
  /* le locator */
  Assert(Field(stub,STUB_LOCATOR) != stub);
  tmp = Field(stub,STUB_LOCATOR);
  fprintf(stderr,"Locator : %x\n",tmp);
  if(tmp!=Val_unit){
    if(!Is_block (tmp))
      failed_message(msg,"option block",file,line);
    if(Tag_val(tmp)!=0){
      fprintf(stderr,"stub tag: %d size %d \n",Tag_val(stub),Wosize_val(stub));
      fprintf(stderr,"tag: %d size %d \n",Tag_val(tmp),Wosize_val(tmp));
      failed_message(msg,"option tag",file,line);
    }
    if(Wosize_val(tmp)!=1)
      failed_message(msg,"option size",file,line);
    tmp = Field(tmp,0);
    if(!Is_block (tmp))
      failed_message(msg,"Locator block",file,line);
    if(Tag_val(tmp)!=0)
      failed_message(msg,"locator tag",file,line);
    if(Wosize_val(tmp)!=4)
      failed_message(msg,"locator size",file,line);
    /* distance */
    if(!Is_long(Field(tmp,0)))
      failed_message(msg,"locator.age int",file,line);
    if(!Is_long(Field(tmp,1)))
      failed_message(msg,"locator.distance int",file,line);
    /* on affiche l'UID  */
    tmp = Field(tmp,2); /* Uid(space_id,ts) */
    fprintf(stderr,"UID(%d:",Int_val(Field(tmp,1)));
    tmp = Field(tmp,0); /* space_id */
    tmp = Field(tmp,0);
    fprintf(stderr,"%d:%d)\n",Int_val(Field(tmp,0)),Int_val(Field(tmp,1)));
  }
  tmp = Field(stub,STUB_LOCAL);
  if(tmp!=Val_unit){
    fprintf(stderr,"stub local (%x)\n",tmp);
    if(Is_blue_hd(Hd_val(tmp)))
      failed_message(msg,"local is a forwarded",file,line);
    fprintf(stderr,"tag: %d size %d \n",Tag_val(tmp),Wosize_val(tmp));
  }
  else
    fprintf(stderr,"stub remote\n");

  tmp = Field(stub,STUB_STUB);
  if(tmp!=Val_unit)
    fprintf(stderr,"GC stub : %lx\n",tmp);
    
}

value find_method = Val_unit;
value set_find_method(value f)   /* ML */
{
  find_method = f;
  register_global_root(&find_method);

  return Val_unit;
}

value resolv_method(code_t pc)
{
  value module;
  value meth_name;
  value label;
  
  module = Field(module_table,pc[0]);
  meth_name = Field(module, pc[1]+MODULE_DATA_SIZE)

  DPRINT1("Resolv method %s\n",String_val(meth_name));
  label = callback(find_method, meth_name);
  
  pc[-1] = THREADED(PUSHCONSTINT);
  pc[0] = Int_val(label);
  pc[1] = THREADED(GETMETHOD);
  return label;
}

/*
Les exceptions: 
  Il faut pouvoir rebinder les exceptions circulant entre deux runtimes.
On distinguera deux cas:
- exception native: elles peuvent circuler entre les runtimes, mais ne sont
    pas reconnues.
- exception bytecode/mixte: elles peuvent circuler entre les runtimes, et
    restent uniques sur un meme runtime.

Le principal probleme, c'est que les modules n'existent pas dans le code
natif. Dans le code mixte, les exceptions sont generees dans le code
natif, puis reprises dans le bytecode. Comment retrouver ces exceptions, voir
partager ces exceptions entre le bytecode et le code natif ? comment ?

*/

value add_exn = Val_unit;

value set_add_exn(value f) /* ML */
{
  if(add_exn != Val_unit || f == Val_unit) failwith("set_add_exn");
  
  add_exn = f;
  register_global_root(&add_exn);
  
  return Val_unit;
}

value create_exception(value name,value module_data) /* ML */
{
  value exc;
  
  Begin_roots2(name,module_data);
  
  exc = alloc(3,Exn_tag);
  Field(exc,0) = name;
  Field(exc,1) = Field(module_data,MODULE_NAME);
  Field(module_data,CLASSES) = Field(module_data,CLASSES) + 2;
  Field(exc,2) = Field(module_data,CLASSES);
  Begin_root(exc);
  if(add_exn == Val_unit){
    value list;
    
    list = alloc(2,0);
    Field(list,0) = exc;
    Field(list,1) = Field(module_data,EXCEPTIONS);
    Assign(Field(module_data,EXCEPTIONS),list);
  } else {
    /* si l'exception existe deja, l'ancienne remplace la nouvelle */
    exc = callback(add_exn,exc);
  }
  End_roots();
  End_roots();
  
  return exc;
}
