/***********************************************************************/
/*                                                                     */
/*                           Objective Caml                            */
/*                                                                     */
/*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         */
/*                                                                     */
/*  Copyright 1996 Institut National de Recherche en Informatique et   */
/*  Automatique.  Distributed only by permission.                      */
/*                                                                     */
/***********************************************************************/

/* $Id: register.c,v 1.1.1.1 1998/04/20 11:32:11 lefessan Exp $ */

/* Callbacks from C to Caml */

#include <string.h>
#include "callback.h"
#include "memory.h"
#include "mlvalues.h"
#include "common.h"

/* Naming of Caml values */

struct named_value {
  value val;
  struct named_value * next;
  char name[1];
};

#define Named_value_size 13

static struct named_value * named_value_table[Named_value_size] = { NULL, };

static unsigned int hash_value_name(char *name)
{
  unsigned int h;
  for (h = 0; *name != 0; name++) h = h * 19 + *name;
  return h % Named_value_size;
}

value register_named_value(value vname, value val) /* ML */
{
  struct named_value * nv;
  char * name = String_val(vname);
  unsigned int h = hash_value_name(name);

  nv = (struct named_value *)
         stat_alloc(sizeof(struct named_value) + strlen(name));
  strcpy(nv->name, name);
  nv->val = val;
  nv->next = named_value_table[h];
  named_value_table[h] = nv;
  register_global_root(&nv->val);
  return Val_unit;
}

value * caml_named_value(char *name)
{
  struct named_value * nv;
  for (nv = named_value_table[hash_value_name(name)];
       nv != NULL;
       nv = nv->next) {
    if (strcmp(name, nv->name) == 0) return &nv->val;
  }
  return NULL;
}

int callback_depth = 0;

value callback(value closure, value arg)
{
   value res;
   callback_depth++;
   if(Tag_val(closure) == Closure_byte_tag)
    res = byte_callback(closure,arg);
   else
     res = asm_callback(closure,arg);
   callback_depth--;
   return res;
}


value callback2(value closure, value arg1,value arg2)
{
  value res;
  callback_depth++;
  if(Tag_val(closure) == Closure_byte_tag)
    res = byte_callback2(closure,arg1,arg2);
  else
    res = asm_callback2(closure,arg1,arg2);
  callback_depth--;
  return res;
}


value callback3(value closure, value arg1,value arg2,value arg3)
{
  value res;
  callback_depth++;
  if(Tag_val(closure) == Closure_byte_tag)
    res = byte_callback3(closure,arg1,arg2,arg3);
  else
    res = asm_callback3(closure,arg1,arg2,arg3);
  callback_depth--;
  return res;
}

