(***********************************************************************)
(*                                                                     *)
(*                           Objective Caml                            *)
(*                                                                     *)
(*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         *)
(*                                                                     *)
(*  Copyright 1996 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

(* $Id: errors.ml,v 1.1.1.1 1998/04/20 11:32:12 lefessan Exp $ *)

(* Error report *)

open Format
open Location

(* Report an error *)

let report_error exn =
  open_box 0;
  begin match exn with
    Lexer.Error(err, start, stop) ->
      Location.print {loc_start = start; loc_end = stop};
      Lexer.report_error err
  | Syntaxerr.Error err ->
      Syntaxerr.report_error err
  | Env.Error err ->
      Env.report_error err
  | Typemisc.Error(location, err) ->
      Location.print location; Typemisc.report_error err
  | Typetexp.Error(location, err) ->
      Location.print location; Typetexp.report_error err
  | Typedecl.Error(location, err) ->
      Location.print location; Typedecl.report_error err
  | Includemod.Error err ->
      Includemod.report_error err
  | Typemod.Error(location, err) ->
      Location.print location; Typemod.report_error err
  | Translcore.Error(location, err) ->
      Location.print location; Translcore.report_error err
  | Symtable.Error code ->
      Symtable.report_error code
  | Bytelink.Error code ->
      Bytelink.report_error code
  | Bytelibrarian.Error code ->
      Bytelibrarian.report_error code
  | Sys_error (err,msg) ->
      print_string "I/O error: "; print_string msg
  | Typeclass.Error(location, err) ->
      Location.print location; Typeclass.report_error err
  | x ->
      close_box(); raise x
  end;
  close_box(); print_newline()
