(***********************************************************************)
(*                                                                     *)
(*                     the join-calculus language                      *)
(*                                                                     *)
(*   Cedric Fournet and Luc Maranget, projet PARA, INRIA Rocquencourt  *)
(*                                                                     *)
(*  Copyright 1997 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

open graphics
open ml


do open_graph (" 800x800+0+0")

let sx = size_x ()
let sy = size_y ()


let xmin = float_of_string("-1.5")
and xmax =  float_of_string("0.5")
and ymin =  float_of_string("-1.25")
and ymax =  float_of_string("+1.25")

let pixel = 1

let limit = 32

do  set_font("-*-times-bold-r-*-*-20-*-*-*-*-*-*-*")

let print_name(i,j,w,h,name,color) =
    let wt,ht = text_size(name) in
    if color <> black then (
      set_color(black) ;      
      fill_rect(i+2,j+h/2-2,wt+4,ht+4)
    ) ;
    set_color(color) ;
    moveto(i+4,j+h/2-4) ;
    draw_string(name) ;
    reply
;;

let draw_square(name,i,j,w,h,data)  =
    spawn
      {draw.quad(data,i,j,w,h,limit) ;
      print_name(i,j,w,h,name,white) ;} ;
    reply  to draw_square


let make_comp (there,killover) =
  loc mandel
  with square(i0,j0,w,h) =
  let r = quad.empty(w,h,limit) in
    let iter(i,j) =
      let ii = i+i0 and jj = j+j0 in
      if i >= w then {reply}
      else {
        if j >= h then {
          iter(i+1,0) ; reply
        } else {
          let x = xmin +. ((xmax -. xmin) *. float(ii*pixel)) /. float(sx)
          and y = ymin +. ((ymax -. ymin) *. float(jj*pixel)) /. float(sy) in
          let m = calc.mandel(x,y,limit) in
          quad.set(r,i,j,m) ;
          iter(i,j+1) ; reply
        }
      } in
      iter(0,0) ; reply r to square
  and kill () = print_string("Here I die") ; print_newline () ; killover()
  init
    print_string("Here I go") ; print_newline () ;
    go(there)  ; print_string("Here I am") ; print_newline () ;
  end in
  reply square,kill

let w = 100 and h = 100 (* true sizes *)

let do_it () = 

let S(n,m) | next(name,job,kill) =
  if m + pixel > sy then {S(n+w,0) |  next(name,job,kill)}
  else if n+pixel > sx then {kill() | KillAll()}
  else {
      S(n,m+h) |
      let w = min(w,sx-n) and h = min(h,sy-m) in
      print_name(n,m,w,h,name,black) ;
      let r = job(n/pixel,m/pixel,w/pixel,h/pixel) in
      draw_square(name,n,m,w,h,r) ;
      next(name,job,kill)
  }
and KillAll() | next(name,job,kill) =
  KillAll() | kill() in


let register_me(local,kill,name) =
  let name =
   if string.length(name) > 8 then
     string.sub(name,0,8)
   else name in
  let job,kill = make_comp(local,kill) in
  next(name,job,kill)

in ns.register("register_me",register_me) ; S(0,0)

spawn do_it()


