(***********************************************************************)
(*                                                                     *)
(*                     the join-calculus language                      *)
(*                                                                     *)
(*   Cedric Fournet and Luc Maranget, projet PARA, INRIA Rocquencourt  *)
(*                                                                     *)
(*  Copyright 1997 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

open Graphics

let _ = open_graph (" 800x800")

let sx = size_x ()
let sy = size_y ()


let xmin = float_of_string("-1.5")
and xmax =  float_of_string("0.5")
and ymin =  float_of_string("-1.25")
and ymax =  float_of_string("+1.25")

let pixel = 1

let limit = 32

let _ = set_font("-*-times-bold-r-*-*-20-*-*-*-*-*-*-*")

let def print_name! (i,j,w,h,name,color) =
  Debug.print ("name of window "^name);
  (let wt,ht = text_size(name) in
  if color <> black then (
    set_color(black) ;      
    fill_rect (i+2) (j+h/2-2) (wt+4) (ht+4)
      );
  set_color(color) ;
  moveto (i+4) (j+h/2-4) ;
  draw_string(name) );
  print_string ("Client : "^name);
  print_newline();
    reply ()
;;

let def draw_square! (name,i,j,w,h,data)  =
  {
  Debug.print ("----------> "^name);
  Draw.quad data i j w h limit ;
  print_name (i,j,w,h,name,white);} |
  reply  to draw_square

let def make_comp! (there,killover) =
  let loc mandel(square,kill) = struct
    let def square! (i0,j0,w,h) =
      let r = Quad.empty w h limit in
      let def iter! (i,j) =
        Debug.print "iter\n";
        let ii = i+i0 and jj = j+j0 in
        if i >= w then
          reply (Debug.print "End iter\n") 
        else 
          if j >= h then
            reply iter ( (i+1),0 )
          else
          {
          let x = xmin +. ((xmax -. xmin) *. float(ii*pixel)) /. float(sx)
          and y = ymin +. ((ymax -. ymin) *. float(jj*pixel)) /. float(sy) in
          let m = Calc.mandel x y limit in
          Quad.set r i j m;
          reply iter (i,j+1)
          } 
       in
         iter (0,0) ;reply r to square
  and kill () = print_string("Here I die") ; print_newline () ; killover ()
  ;;
  print_string("Here I go") ; print_newline () ;
  Join.go there  ; print_string("Here I am") ; print_newline () ;;
end in
  reply (square,kill)

let w = 100 and h = 100 (* true sizes *)

let def s (n,m) | next (name,job,kill) =
  if m + pixel > sy then  { s(n+w,0) |  next(name,job,kill) }
  else if n+pixel > sx then  { kill() | killAll() }
  else  {
      { Debug.print "s" ; s(n,m+h) } |
      let w = min w (sx-n) and h = min h (sy-m) in
      print_name (n,m,w,h,name,black) ;
      let r = job (n/pixel,m/pixel,w/pixel,h/pixel) in
      Debug.print ("draw square:"^name);
      draw_square (name,n,m,w,h,r);
      Debug.print ("next name:"^name);
    next(name,job,kill)
} 
or killAll () | next (name,job,kill) = killAll() | kill() 

let def register_me (local,kill,name) =
  let name =
   if String.length(name) > 8 then
     String.sub name 0 8
   else name in
  let job,kill = make_comp (local,kill) in
   Debug.print ("next_name "^name) ; next(name,job,kill)

let _ = 
  Ns.register "mandel:register_me" register_me vartype;
(*  Join.setLocalService "register_me" register_me vartype; *)
  spawn { Debug.print "s" ; s(0,0) }

let _ = 
  Join.server ();
  Nfs.register "mandel"

