(***********************************************************************)
(*                                                                     *)
(*                     the join-calculus language                      *)
(*                                                                     *)
(*   Cedric Fournet and Luc Maranget, projet PARA, INRIA Rocquencourt  *)
(*                                                                     *)
(*  Copyright 1997 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

open Graphics

type tree = Nodex of cell | Nodey of cell | Leaf of Graphics.color
and cell = {mutable left:tree ; mutable right:tree}

type t = tree ref * int * int

let empty w h c = ref (Leaf c),w,h

let normalize = function
  Nodex {left=Leaf cl ; right=Leaf cr} when cl=cr -> Leaf cl
| Nodey {left=Leaf cl ; right=Leaf cr} when cl=cr -> Leaf cl
| a -> a

let set t i j v =
  let t,w0,h0 = t in
  let rec do_rec (i0,j0) (w,h) a =
(*
    print_string ("do_rec "^string_of_int i0^" "^
      string_of_int j0^" "^
      string_of_int w^" "^
      string_of_int h) ;
    print_newline () ;
*)
    match a with
    Leaf v0 as leaf ->
      if  v=v0 then Leaf v
      else
        if w = 1 && h = 1 then
          Leaf v
        else if w > h then
          let c = {left = leaf ; right = leaf} in
          do_rec (i0,j0) (w,h) (Nodex c)
        else
          let c = {left = leaf ; right = leaf} in
          do_rec (i0,j0) (w,h) (Nodey c)
   | Nodex c as n ->
      let w1 = w/2 in
      let w2 = w-w1 in
      if i < i0 + w1 then begin
        let new_left = do_rec (i0,j0) (w1,h) c.left in
        c.left <- new_left
      end else begin
        let new_right = do_rec(i0+w1,j0)  (w2,h) c.right in
        c.right <- new_right
      end ;
      normalize n
  | Nodey c as n ->
      let h1 = h/2 in
      let h2 = h - h1 in
      if j < j0 + h1 then begin
        let new_left = do_rec (i0,j0) (w,h1) c.left in
        c.left <- new_left
      end else begin
        let new_right = do_rec(i0,j0+h1)  (w,h2) c.right in
        c.right <- new_right
      end ;
      normalize n in
  if i < 0 || i >= w0 || j < 0 || j > h0 then
     raise (Invalid_argument "Quad") ;
(*
  print_string ("set: "^string_of_int i^" "^
    string_of_int j^" "^
    string_of_int v) ;
  print_newline () ;
*)
  let new_t = do_rec (0,0) (w0,h0) !t in
  t := new_t
        

(*
let _ = open_graph ""

let t = empty 256 128

let _ = set t 250 78  6 

let do_it j c =
  for i = 0 to 255 do
   set t i j c
  done

let _ = do_it 25 0
let _ = do_it 50 128

let _ = draw t

let rec f () = f ()

let _ = f ()

*)
