(***********************************************************************)
(*                                                                     *)
(*                     the join-calculus language                      *)
(*                                                                     *)
(*   Cedric Fournet and Luc Maranget, projet PARA, INRIA Rocquencourt  *)
(*                                                                     *)
(*  Copyright 1997 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

open Graphics

let _ = open_graph (" 800x800")

let sx = size_x ()
let sy = size_y ()


let xmin = float_of_string("-1.5")
and xmax =  float_of_string("0.5")
and ymin =  float_of_string("-1.25")
and ymax =  float_of_string("+1.25")

let pixel = 1

let limit = 32

let _ = set_font("-*-times-bold-r-*-*-20-*-*-*-*-*-*-*")

let print_name (i,j,w,h,name,color) =
  (let wt,ht = text_size(name) in
  if color <> black then (
    set_color(black) ;      
    fill_rect (i+2) (j+h/2-2) (wt+4) (ht+4)
      );
  set_color(color) ;
  moveto (i+4) (j+h/2-4) ;
  draw_string(name)
  );
  print_string ("Client : "^name);
  print_newline()

let draw_square (name,i,j,w,h,data)  =
  Draw.quad data i j w h limit ;
  print_name (i,j,w,h,name,white)

let  square (i0,j0,w,h) =
  let r = Quad.empty w h limit in
  for i = 0 to w - 1 do
    for j = 0 to h - 1 do
      let ii = i+i0 and jj = j+j0 in
      let x = xmin +. ((xmax -. xmin) *. float(ii*pixel)) /. float(sx)
      and y = ymin +. ((ymax -. ymin) *. float(jj*pixel)) /. float(sy) in
      let m = Calc.mandel x y limit in
      Quad.set r i j m
    done 
  done;
  r

let w = 100 and h = 100 (* true sizes *)

let name = String.sub "specialix" 0 8

let rec s (n,m) =
  if m + pixel > sy then  s (n+w,0) 
  else if n+pixel > sx then ()
  else  
    let w = min w (sx-n) and h = min h (sy-m) in
    print_name (n,m,w,h,name,black) ;
    let r = square (n/pixel,m/pixel,w/pixel,h/pixel) in
    draw_square (name,n,m,w,h,r);
    s (n,m+h)


let _ =  s (0,0)

