(***********************************************************************)
(*                                                                     *)
(*                     the join-calculus language                      *)
(*                                                                     *)
(*   Cedric Fournet and Luc Maranget, projet PARA, INRIA Rocquencourt  *)
(*                                                                     *)
(*  Copyright 1997 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

(* encoding of data-structures:
   - object-style FIFO
   - depth-first traversal of a tree
   - breadth-first traversals of a tree using FIFO *)

let tuple (p,c,gv,gs) =
  let get () =   reply p,c,gv,gs to get in
  reply get to tuple

let print_node (t) =
  let p,c,gv,gs = t () in
  p () ; reply to print_node

let leaf(i) =
  let  get_value () =  reply i to get_value
  and  pretty () =  print(i) ; reply to pretty 
  and  get_sons () =
    let x = failwith("get_sons of leaf") in reply x,x to get_sons
  and  code () =  reply 0 to code in
  reply tuple(pretty,code,get_value,get_sons) to leaf

let node(l,r) =
  let  get_value () = reply failwith("get_value of node") to get_value
  and  pretty () =
     print_string("<") ;
     print_node(l) ;
     print_string(".") ;
     print_node(r) ;
     print_string (">") ;
     reply to pretty
  and  get_sons () =  reply l,r to get_sons
  and  code () =  reply 1 to code in
  reply tuple(pretty,code,get_value,get_sons) to node

let new_fifo () =

  let new_cons(x) =
   let N() | setprev(n) = S(n) | reply to setprev
   and S(p) | remove() = reply x,p,false to remove
   and N() | remove () = reply x,self,true to remove
   and self () = reply setprev,remove to self in
   N() | reply self to new_cons in

  let E() | empty () = reply true to empty
  and E() | enfile(x) =
    let fst = new_cons(x) in
    S(fst,fst) | reply to enfile
  and E() | defile () = reply failwith("Defile vide\n") to defile
  and S(fst,lst) | empty () =  S(fst,lst) | reply false to empty
  and S(fst,lst) | enfile(x) =
     let new_fst = new_cons(x) in
     let set,rem = fst() in
     set(new_fst) ;
     S(new_fst,lst) |
     reply to enfile
  and S(fst,lst) | defile () =
     let set,rem = lst() in
     let x,p,last = rem () in
     if last then {E()} else {S(fst,p)} |
     reply x to defile in
  E() | reply empty,enfile,defile to new_fifo
   

let t = node(node(leaf(1),node(leaf(2),leaf(3))),leaf(4))
let t = node(node(t,leaf(0)),t)

do print_node(t) ; print_newline()

let dfs(t) =
  let  p,c,gv,gs = t () in
  if c() = 0 then {
    print(gv()) ; print_string(" ") ; reply to dfs
  } else {
    let l,r = gs () in
    dfs(l) ; dfs(r)  ; reply to dfs
  }


let bfs(t) =
  let v,e,d = new_fifo () in
  let Loop() = 
    if v() then {Fini()}
    else {
      let t = d() in
      let p,c,gv,gs = t () in
      if c() = 0 then {
        print(gv()) ;
        print_string(" ") ;
        Loop ()
      } else {
        let l,r = gs() in
        e(l) ; e(r) ; Loop ()
      }
    }
   and Fini() | k() = reply to k in
    e(t) ; Loop () | k() ; reply to bfs


do dfs(t) ; print_newline() ;
   bfs(t) ; print_newline()

