(***********************************************************************)
(*                                                                     *)
(*                     the join-calculus language                      *)
(*                                                                     *)
(*   Cedric Fournet and Luc Maranget, projet PARA, INRIA Rocquencourt  *)
(*                                                                     *)
(*  Copyright 1997 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

(* draws a fractal leave in a concurrent manner (from the root to the
   leaves); uses a lock so that series of graphical primitives do not
   mix; this exhibits a hybrid scheduling strategy, and suggest how
   fair it is in practice. (compare with leave.j!) *)

(* the program terminates once the leave is complete; to prevent this,
   you can start it with flag -s *)

open ml
open pair
open graphics
  
let pi = float(4) *. atan(float(1)) (* well-known formula *)
and epsilon = float(4)
  
let split(p) = reply fst(p),snd(p)
let trunc(p) =
  let h,v = split(p) in reply truncate(h),truncate(v)
  
let trace(scale,p,q) | once() =
  
  let c =
    if      scale > float(40) then rgb(10,48,10)
    else if scale > float(15) then rgb(10,96,10)
    else if scale > float(12) then rgb(20,128,20)
    else if scale > float(7)  then rgb(40,196,10)
    else                           rgb(40,255,0) in

  let w =
    if      scale > float(15) then 3
    else if scale > float(7)  then 2
    else 1 in
  
  let x0,y0 = trunc(p) in
  let x1,y1 = trunc(q) in

  set_color(c);
  set_line_width(w);
  moveto(x0,y0);
  lineto(x1,y1 );
  once() | reply to trace

spawn once()
  
let leave(hom1,hom2,h,v) =

  let r1,e1 = split(hom1) in
  let tige = float_of_string("1") -. e1 in
  let branche = tige *. float_of_string("0.5") in
  let racine = create(float(h/2),float(0)) in
  let orientation = create(float(-5) *. r1 , float(v)) in

  let leaf(p,bent,draw,h) =
    let angle,scale = split(h) in
    if scale < epsilon then { reply } else

    { let rotate(vers_la_gauche,h) =
        let r,e = split(h) in
        reply create(
          if vers_la_gauche then angle +. r else angle -. r,
          scale *. e) in 

      let shift(p,d) =
        let x,y = split(p) in
        reply create(
          x +. d *. scale *. sin(angle),
          y +. d *. scale *. cos(angle) ) in
      let q = shift(p,tige) in
      let r = shift(p,branche) in

      trace(scale,p,q);
      { leaf(r,(not(draw)),draw,rotate (draw,hom2)); }
      | leaf(q,bent,(not(draw)),rotate (bent,hom1)); reply } in

  leaf(racine,true,false,orientation);
;;
spawn
  open_graph("");
  set_color(black);
  leave(
  create(float_of_string("0.04"),float_of_string("0.86")),
  create(float_of_string("0.7"),float_of_string("0.53")),
  size_x(),
  size_y() )




