(***********************************************************************)
(*                                                                     *)
(*                     the join-calculus language                      *)
(*                                                                     *)
(*   cedric fournet and luc maranget, projet para, inria rocquencourt  *)
(*                                                                     *)
(*  copyright 1997 institut national de recherche en informatique et   *)
(*  automatique.  distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

(* encoding of data-structures:
   - object-style lists
   - merge-sort *)

let triple (x,y,z>> =
  let  get <<>> =   reply x,y,z to get in
  reply get to triple

let new_cons <<h,t>> =
  let  head <<>> =  reply h to head
  or  tail <<>> =  reply t to tail
  or  null <<>> =  reply false to null in
  reply triple <<null,head,tail>> to new_cons
or new_nil <<>> =
  let  head <<>> = reply failwith <<"head of nil">> to head
  or  tail <<>> = reply failwith <<"tail of nil">> to tail
  or  null <<>> = reply true to null  in
  reply triple <<null,head,tail>> to new_nil

let pspace <<>> = print_string " " ; reply to pspace
or skip <<>> = print_string "\n" ; reply to skip

let print_list<<p,sep,l>> =
  let print<<l>> =
    let n,h,t = l <<>> in
    if n<<>> then {fini <<>>}
    else {
      let l2 = t<<>> in
      let nt,ht,tt = l2 <<>> in
      if nt<<>> then {p<<h<<>>>> ; fini <<>>}
      else {p<<h<<>>>> ; sep<<>> ; print<<t<<>>>>}
    }
  or fini<<>> | k<<>> = reply to k in
  print<<l>> | k<<>> ; reply to print_list

let pint<<x>> = print<<x>> ; reply to pint

let pints<<l>> = print_list<<pint,pspace,l>> ; reply to pints

let pintss<<l>> =
  print_list<<pints,skip,l>> ; print_newline<<>> ;
  print_newline<<>> ; reply to pintss


let merge<<l1,l2>> =
  reply
    let n1,h1,t1 = l1 <<>> in
    if n1 <<>> then l2
    else 
      let n2,h2,t2 = l2 <<>> in
      if n2 <<>> then l1
      else if h1 <<>> < h2 <<>> then
        new_cons<<h1 <<>>,merge<<t1 <<>>,l2>>>>
      else
        new_cons<<h2 <<>>,merge<<l1,t2 <<>>>>>>    
  to merge

let sort<<l>> =

  let first_pass<<l>> =
    let n,h,t = l<<>> in
    if n<<>> then
       {reply new_nil <<>> to first_pass}
    else
       {reply
         new_cons<<new_cons<<h<<>>,new_nil<<>>>>,first_pass<<t <<>>>>>>
       to first_pass} in

  let merge_aux<<l>> =
    let n,h,t = l<<>> in
    if n <<>> then reply new_nil <<>> to merge_aux
    else {
      let l2 = t<<>> in
      let nt,ht,tt = l2<<>> in
      if nt<<>> then reply new_cons<<h<<>>,new_nil<<>>>>  to merge_aux
      else {
        reply new_cons<<merge<<h<<>>,ht<<>>>>,merge_aux<<tt<<>>>>>> to merge_aux
      }
    } in

  let step<<l>> =
    pintss<<l>> ;
    let n,h,t = l <<>> in
    if n<<>> then {
      reply new_nil <<>>  to step
    } else {
      let l2 = t <<>> in
      let nt,ht,tt = l2<<>> in
      if nt <<>> then {
         reply h<<>> to step
      } else {
        reply step<<merge_aux<<l>>>> to step
      }
    } in
  reply step<<first_pass<<l>>>> to sort

let inter<<m,n>> =
  reply
    if m < n then new_nil <<>>
    else new_cons<<m,inter<<m-1,n>>>>
  to inter

do
  let r = sort<<inter<<10,1>>>> in print_newline <<>>
