(***********************************************************************)
(*                                                                     *)
(*                     the join-calculus language                      *)
(*                                                                     *)
(*   Cedric Fournet and Luc Maranget, projet PARA, INRIA Rocquencourt  *)
(*                                                                     *)
(*  Copyright 1997 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

(* encoding of data-structures:
   - object-style FIFO
   - depth-first traversal of a tree
   - breadth-first traversals of a tree using FIFO *)

let def tuple ! (p,c,gv,gs) =
  let def get ! () =   reply p,c,gv,gs to get in
  reply get to tuple

let def print_node ! (t) =
  let p,c,gv,gs = t  () in
  p  () ; reply to print_node

let def leaf! (i) =
  let def  get_value ! () =  reply i to get_value
  and  pretty ! () =  Join.print i ; reply to pretty 
  and  get_sons ! () =
    let x = failwith "get_sons of leaf" in reply x,x to get_sons
  and  code ! () =  reply 0 to code in
  reply tuple (pretty,code,get_value,get_sons) to leaf

let def node! (l,r) =
  let def  get_value ! () = reply failwith "get_value of node" to get_value
  and  pretty ! () =
     print_string "(" ;
     print_node (l) ;
     print_string"." ;
     print_node (r) ;
     print_string ")" ;
    flush stdout;
     reply to pretty
  and  get_sons ! () =  reply l,r to get_sons
  and  code ! () =  reply 1 to code in
  reply tuple (pretty,code,get_value,get_sons) to node

let def new_fifo ! () =

  let def new_cons! (x) =
   let def n() | setprev! (n') = s (n') | reply to setprev
   or s(p) | remove! () = reply (x,p,false) to remove
   or n() | remove ! () = reply (x,self,true) to remove
   or self ! () = reply ( (setprev,remove)) to self in
   n () | reply self to new_cons in

  let def e() | empty ! () = reply true to empty
  or e() | enfile! (x) =
    let fst = new_cons (x) in
    s (fst,fst) | reply to enfile
  or e() | defile ! () = reply failwith "defile vide\n" to defile
  or s(fst,lst) | empty ! () =  s (fst,lst) | reply false to empty
  or s(fst,lst) | enfile! (x) =
     let new_fst = new_cons (x) in
     let set,rem = fst () in
     set (new_fst) ;
     s (new_fst,lst) |
     reply to enfile
  or s(fst,lst) | defile ! () =
     let set,rem = lst () in
     let x,p,last = rem  () in
     if last then e () else {s (fst,p)} |
     reply x to defile in
  e () | reply empty,enfile,defile to new_fifo
   

let t = node (node (leaf (1),node (leaf (2),leaf (3))),leaf (4))
let t = node (node (t,leaf (0)),t)

let _ = print_node (t) ; print_newline ()

let def dfs! (t) =
  let  p,c,gv,gs = t  () in
  if c () = 0 then {
    print_int (gv ()) ; print_string (" ") ; reply to dfs
  } else {
    let l,r = gs  () in
    dfs (l) ; dfs (r)  ; reply to dfs
  }


let def bfs! (t) =
  let v,e,d = new_fifo  () in
  let def loop () = 
    if v () then {fini ()}
    else {
      let t = d () in
      let p,c,gv,gs = t  () in
      if c () = 0 then {
        print_int (gv ()) ;
        print_string (" ") ;
        loop  ()
      } else {
        let l,r = gs () in
        e (l) ; e (r) ; loop ()
      }
    }
   or fini () | k! () = reply to k in
    e (t) ; loop  () | k () ; reply to bfs


let _ =
  dfs (t) ; print_newline () ;
  bfs (t) ; print_newline ()

;;
