(***********************************************************************)
(*                                                                     *)
(*                     the join-calculus language                      *)
(*                                                                     *)
(*   Cedric Fournet and Luc Maranget, projet PARA, INRIA Rocquencourt  *)
(*                                                                     *)
(*  Copyright 1997 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

(* encoding of data-structures:
   - object-style lists
   - merge-sort *)

type 'a t = T of ( <<>> => ((<<>> => bool) * (<<>> => 'a) * (<<>> => ('a t))))

let print i = print_int i; print_newline ()

let triple <<x,y,z>> =
  let  get <<>> =   reply x,y,z in
  reply get

let new_cons <<h,t>> =
  let  head <<>> =  reply h
  and  tail <<>> =  reply t
  and  null <<>> =  reply false in
  reply triple <<null,head,tail>>
and new_nil <<>> =
  let  head <<>> = failwith "head of nil" ; reply ()
  and  tail <<>> = failwith "tail of nil"; reply ()
  and  null <<>> = reply true  in
  reply triple <<null,head,tail>>

let pspace <<>> = print_string " " ;reply ()
and skip <<>> = print_string "\n" ; reply ()

let print_list<<p,sep,l>> =
  let print <l> =
    let n,h,t = l <<>> in
    if n <<>> then fini <<>>
    else  {
      let T l2 = t <<>> in
      let nt,ht,tt = l2 <<>> in
      if nt<<>> then { p<< h<<>> >> ;fini <<>>  }
      else { p<<h<<>>>> ; sep<<>> ; print <<let T t =t<<>> in t>>  }
    } 
  and fini <> | k<<>> = reply to k in
  print<<l>> | {k<<>> ;reply to print_list}

let pint<<x>> = print_int x ;reply to pint

let pints<<l>> = print_list<<pint,pspace,l>> ; reply to pints

let pintss<<l>> =
        print_list<<pints,skip,l>> ; print_newline () ;
      print_newline() ; reply to pintss


let merge<<l1,l2>> =
  reply
    let n1,h1,t1 = l1 <<>> in
    if n1 <<>> then l2
    else 
      let n2,h2,t2 = l2 <<>> in
      if n2 <<>> then l1
      else if h1 <<>> < h2 <<>> then
        new_cons<<h1 <<>>,merge<<t1 <<>>,l2>>>>
      else
        new_cons<<h2 <<>>,merge<<l1,t2 <<>>>>>>    
  to merge

let sort<<l>> =

  let first_pass<<l>> =
    let n,h,t = l<<>> in
    if n<<>> then
        { reply new_nil <<>> to first_pass }
    else
        {reply
         new_cons<<new_cons<<h<<>>,new_nil<<>>>>,first_pass<<t <<>>>>>>
       to first_pass } in

  let merge_aux<<l>> =
    let n,h,t = l<<>> in
    if n <<>> then reply new_nil <<>> to merge_aux
    else  {
      let l2 = t<<>> in
      let nt,ht,tt = l2<<>> in
      if nt<<>> then reply new_cons<<h<<>>,new_nil<<>>>>  to merge_aux
      else  {
        reply new_cons<<merge<<h<<>>,ht<<>>>>,merge_aux<<tt<<>>>>>> to merge_aux
      } 
    }  in

  let step<<l>> =
    {
      pintss<<l>> 
      ;
    let n,h,t = l <<>> in
    if n<<>> then  { 
      reply new_nil <<>>  to step
    } else  { 
      let l2 = t <<>> in
      let nt,ht,tt = l2<<>> in
      if nt <<>> then  { 
        reply h<<>> to step
      } else  { 
        reply step<<merge_aux<<l>>>> to step
      }
    } }  in
  reply step<<first_pass<<l>>>> to sort
;;
let inter<<m,n>> =
  reply
    if m < n then new_nil <<>>
    else new_cons<<m,inter<<m-1,n>>>>
  to inter
;;
let r = sort<<inter<<10,1>>>> in print_newline ()
