(***********************************************************************)
(*                                                                     *)
(*                     the join-calculus language                      *)
(*                                                                     *)
(*   Cedric Fournet and Luc Maranget, projet PARA, INRIA Rocquencourt  *)
(*                                                                     *)
(*  Copyright 1997 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

(* encoding for the lambda-calculus and its reductions *)

let tuple<<x3,x,y,z,t,u>> =
  let  get <<>> =   reply x3,x,y,z,t,u to get in
  reply get to tuple

let pair<<x,y>> =
  let get <<>> = reply x,y to get in
  reply get to pair
;;
let create_var <<>> =
  let s < x > | get << >> = s<< (x+1) >> | reply x in
  s<<0>> | reply get
;;
let new_var = create_var <<>>

let print_term<<t>> =
  let (i,gl,ga,gv,s,p) = t <<>> 
  in 
    p <<>>;  reply () to print_term

let subst_term<<t,v,u>> =
  let i,gl,ga,gv,s,p = t <<>> in reply s<<v,u>> to subst_term

let nature<<t>> =
 let i,gl,ga,gv,s,p = t <<>> in reply i<<>> to nature

let get_lam<<t>> =
 let i,gl,ga,gv,s,p = t <<>> in
 let v,t = gl <<>> in
 reply  v,t to get_lam


let get_app<<t>> =
 let i,gl,ga,gv,s,p = t <<>> in
 let f,a = ga <<>> in
 reply  f,a to get_app


let get_var<<t>> = {
 let i,gl,ga,gv,s,p = t <<>> in
 reply gv<<>> to get_var}
      
type 'a t = T of 'a * ('a t)

let node<<>> =
  let l<v,t> | no <<>> = l<<v,t>> | reply 1 to no
  or a<f,a> | no <<>> = a<<f,a>> | reply 2 to no
  or v<x>   | no <<>> = v<<x>>   | reply 3 to no

  or l<<v,t>> | pretty <<>> = l<<v,t>> |
    print_string "<<\\" ;
    print v ;
    print_string "." ;
    print_term<< t >>;
    print_string ">>";
    reply () to pretty
  or a<<f,a>> | pretty <<>> = a<<f,a>> |
    print_string "<<";
    print_term <<f>>;
    print_string " ";
    print_term << a>>;
    print_string ">>";
    reply () to pretty
  or v<<x>> | pretty <<>> = v<<x>> |
    print<<x>>;
    reply () to pretty
    
  or l<<v,t>> | get_lam <<>> =  l<<v,t>> | reply v,t to get_lam
  or a<<f,a>> | get_app <<>> = a<<f,a>> | reply  f,a to get_app
  or v<<x>> | get_var <<>> = v<<x>> | reply x to get_var

  or l<<v,t>> | subst<<w,u>> = l<<v,t>> |
    if w=v then do reply self <<>> to subst done
    else
       do reply lambda<<v,subst_term<<t,w,u>>>> to subst done
  or a<<f,a>> | subst<<w,u>> = a<<f,a>> |
    reply app<<subst_term<<f,w,u>>,subst_term<<a,w,u>>>> to subst
  or v<<x>> | subst<<w,u>> = v<<x>> |
    if x=w then do reply alphaconv<<u>> to subst done 
    else do reply var<<x>> to subst done 


  or self <<>> =
    reply tuple<<no,get_lam,get_app,get_var,subst,pretty>> to self in

  reply l,a,v,self to node 


or lambda<<x,t>> =
   let l,a,v,s = node <<>> in l<<x,t>> | reply s<<>> to lambda
or app<<f,a>> =
   let l,a,v,s = node <<>> in a<<f,a>> | reply s<<>> to app
or var<<x>> =
   let l,a,v,s = node <<>> in v<<x>> | reply s<<>> to var

or alphaconv<<t>> =
  reply
    if nature<<t>> = 1 then
      let v,t2 = get_lam<<t>> in
      let nv = new_var<<>> in
      lambda
        <<nv,
      (subst_term << 
       (alphaconv<<t2>>),
       v,
       (var<<nv>>)
         >>) 
        >>
    else  if nature<<t>> = 2 then
      let f,a = get_app<<t>> in
      app<<alphaconv<<f>>,alphaconv<<a>>>>
    else
      t
  to alphaconv

let wred<<t>> =
  reply
    if nature<<t>> = 1 then
      t
    else if nature<<t>> = 2 then
      let f,a = get_app<<t>> in
      let f = wred<<f>> in
      if nature<<f>> = 1 then
        let v,t = get_lam<<f>> in
        wred<<subst_term<<t,v,a>>>>
      else
        app<<f,a>>
    else t
  to wred

let one<<t>> =
  let p =
    if nature<<t>> = 1 then
      let v,t = get_lam<<t>> in
      let b,t = one<<t>> in
      pair <<b,lambda<<v,t>>>>
    else if nature<<t>> = 2 then
      let f,a = get_app<<t>> in
      if nature<<f>> = 1 then
        let v,t = get_lam<<f>> in
        pair<<true,subst_term<<t,v,a>>>>
      else
        let b,f = one<<f>> in
        if b then
           pair<<true,app<<f,a>>>>
        else
          let b,a = one<<a>> in
          pair<<b,app<<f,a>>>>
    else pair<<false,t>> in
  let b,t = p<<>> in
  reply b,t to one

let doit<<u>> =
  let step<<t>> =
    reply
      let b,t = one<<t>> in
      if b then step<<t>> else t
    to step in
  let v = step<<alphaconv<<u>>>> in
  reply v to doit
;;

spawn { 
let v1 = new_var <<>>
and v2 = new_var <<>>
and v3 = new_var <<>>
and v4 = new_var <<>> in
let zero = lambda<<v1,lambda<<v2,var<<v2>>>>>>
and succ =
    lambda<<v1,lambda<<v2,lambda<<v3,
       app<<var<<v2>>,app<<app<<var<<v1>>,var<<v2>>>>,var<<v3>>>>>>>>>>>> in
let un = doit<<app<<succ,zero>>>> in
let deux = doit<<app<<succ,un>>>> in
let trois = doit<<app<<succ,deux>>>> in
let quatre = doit<<app<<succ,trois>>>> in

let k =  lambda<<v1,lambda<<v2,var<<v1>>>>>> in
let add =
  lambda<<v1,lambda<<v2,lambda<<v3,lambda<<v4,
    app<<app<<var<<v1>>,var<<v3>>>>,app<<app<<var<<v2>>,var<<v3>>>>,var<<v4>>>>>>>>>>>>>>
and mult =
  lambda<<v1,lambda<<v2,lambda<<v3,
    app<<var<<v1>>,app<<var<<v2>>,var<<v3>>>>>>>>>>>>
and test =
  doit<<lambda<<v1,lambda<<v2,lambda<<v3,
    app<<app<<var<<v1>>,app<<k,var<<v3>>>>>>,var<<v2>>>>>>>>>>>> in

let pair =
  lambda<<v1,lambda<<v2,lambda<<v3,
    app<<app<<var<<v3>>,var<<v1>>>>,var<<v2>>>>>>>>>>
and fst = lambda<<v1,lambda<<v2,var<<v1>>>>>>
and snd = lambda<<v1,lambda<<v2,var<<v2>>>>>> in


let next =
  lambda<<v1,
    let n = app<<var<<v1>>,snd>> in
    app<<app<<pair,n>>,app<<succ,n>>>>>> in
let pred =
  lambda<<v1,
    app<<app<<app<<var<<v1>>,next>>,app<<app<<pair,zero>>,zero>>>>,fst>>>> in

let a =
  lambda<<v1,lambda<<v2,
    app<<var<<v2>>,app<<app<<var<<v1>>,var<<v1>>>>,var<<v2>>>>>>>>>> in

let fix = app<<a,a>> in

let montest = doit<<lambda<<v1,app<<app<<test,var<<v1>>>>,zero>>>>>> in
let factaux =
  doit<<lambda<<v1,lambda<<v2,
    app<<app<<montest,var<<v2>>>>,app<<app<<mult,var<<v2>>>>,app<<var<<v1>>,app<<pred,var<<v2>>>>>>>>>>>>>>>> in

let fact = app<<fix,factaux>> in

let r = doit<<app<<fact,zero>>>> in
print_term <<r>> ;
print_newline () ;

} 

