(***********************************************************************)
(*                                                                     *)
(*                     the join-calculus language                      *)
(*                                                                     *)
(*   Cedric Fournet and Luc Maranget, projet PARA, INRIA Rocquencourt  *)
(*                                                                     *)
(*  Copyright 1997 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

(* an attempt to test many cases in the migration protocol:
   m11 with one big def,
   m12 with numerous small ones.

   * synchronous vs asynchronous; arity n -> m for n,m <3 
   * one big let or several small ones
   * caller/callee * local/remote at call-time, at return-time
   ? different kinds of join-sync
   ? nested calls or not (larger stacks)
   ? nested locations or not
   ? top-level or not

   ? double, triple, round migrations 
*)

do shutdown.ready()

let test(protocol,def_before,def_inside,agent_before,agent_inside) = 

  let name = "\n"^protocol^" " in

  let wait() | trigger() = trigger() | reply to wait
  and inside(s) = print_newline();{ decr(s) | wait();reply to inside }
  and decr(s) | count(n,f) = 
    if n = 0 then { print_string(name^"trigger"^s);print_newline();f();{ trigger() | reply to count } }
    else { print_string(name^"decr"^s);print(n);print_newline();count(n-1,f); reply to count } in

  loc defining 
  with
    get_them() =

      let t0a()        = print_string(name^" t0a ");
                         inside(" t0a "); null() in
      let t1a(x)       = print_string(name^" t1a ");print(x);
                         inside(" t1a "); null() in
      let t2a(x,y)     = print_string(name^" t2a ");print(x);print_string(name^" ");print(y); 
                         inside(" t2a "); null() in

      let t00()        = print_string(name^"t00 ");
                         inside(" t00 "); reply to t00 in
      let t10(x)       = print_string(name^"t10 ");print(x);
                         inside(" t10 "); reply to t10 in
      let t20(x,y)     = print_string(name^"t20 ");print(x);print_string(name);print(y); 
                         inside(" t20 "); reply to t20 in

      let t01()        = print_string(name^"t01 ");
                         inside(" t01 "); reply 19    to t01 in
      let t11(x)       = print_string(name^"t11 ");print(x);
                         inside(" t11 "); reply x     to t11 in
      let t21(x,y)     = print_string(name^"t21 ");print(x);print_string(name);print(y); 
                         inside(" t21 "); reply x-y   to t21 in

      let t02()        = print_string(name^" t02 ");
                         inside(" t02 "); reply 29,28   to t02 in
      let t12(x)       = print_string(name^" t12 ");print(x);
                         inside(" t12 "); reply 129,x   to t12 in
      let t22(x,y)     = print_string(name^" t22 ");print(x);print_string(name^" ");print(y); 
                         inside(" t22 "); reply x-y,x+y to t22 in

      reply t0a,t1a,t2a,t00,t10,t20,t01,t11,t21,t02,t12,t22 to get_them

    and move(a) = go(a);reply to move
  end in

  let t0a,t1a,t2a,t00,t10,t20,t01,t11,t21,t02,t12,t22 = get_them() in

  loc testing
  with f() = 
    print_string(name^"enter inside");print_newline();
    move (def_inside);go(agent_inside);
    print_string(name^"leave inside");print_newline();
    reply to f

  init
    print_string(name^"enter before");print_newline();
    move(def_before);go(agent_before); 
    print_string(name^"leave before");print_newline();
    let result(u) = print_string(name^"-> ");print(u);print_newline();reply in
    { null()

    | t0a()
    | t1a(1)
    | t2a(2,3)

    | t00();print_newline();null() 
    | t10(100);print_newline();null() 
    | t20(200,201);print_newline();null() 

    | { let u = t01() in result(u);null() }
    | { let u = t11(110) in result(u);null() }
    | { let u = t21(210,211) in result(u);null() }

    | { let u,v = t02() in result(u);result(v);null() }
    | { let u,v = t12(120) in result(u);result(v);null() }
    | { let u,v = t22(220,221) in result(u);result(v);null() }      
    } 
  end in

  count(11,f); reply to test 



(* we wait until we are inside every call, then we can trigger
   migrations, then we reply; locations are taken as parameters:

   DEFINING LOC,   CALLING LOC
   before, inside, before, inside *)

loc here end 
let there = ns.lookup("server")

do test("llll",here,here,here,here)
do test("lllr",here,here,here,there)
do test("llrl",here,here,there,here)
do test("llrr",here,here,there,there)

do test("lrll",here,there,here,here)
do test("lrlr",here,there,here,there)
do test("lrrl",here,there,there,here)
do test("lrrr",here,there,there,there)

do test("rlll",there,here,here,here)
do test("rllr",there,here,here,there)
do test("rlrl",there,here,there,here)
do test("rlrr",there,here,there,there)

do test("rrll",there,there,here,here)
do test("rrlr",there,there,here,there)
do test("rrrl",there,there,there,here)
do test("rrrr",there,there,there,there)

spawn shutdown.global()
