open Join
open Array
open Graphics

let ts = ref 0

(* game initializer *)
let def join_game! () | free(n) =
  free(n+1) |  reply to join_game

or start_game! (talk_me,ts) | waiting_as(n,other,m) =

  printnl_string ("start_game ("^(string_of_int m)^")") ;
  {if n+1 < m then {
    waiting_as(n+1,talk_me,m)
  } else {
    start_as(0,talk_me,m)
  } |
  reply m,wait (),other to start_game}

or start_game! (talk_me,ts) | free m =

  printnl_string ("start_game ("^(string_of_int m)^")") ;
  if (m <= 1) then {
    free 1 |
    reply 1,0,talk_me to start_game
  } else {
    waiting_as(1,talk_me,m) |
    let n,talk_other = wait_last () in
    reply m,n,talk_other to start_game
  } 

or start_as (n,talk_last,m) | wait! () =
  {if n+1 < m then {
    start_as(n+1,talk_last,m)
  } else {
    free m
  }} | reply n to wait

or start_as(n,talk_last,m) | wait_last! () =
    {if n+1 < m then {
      start_as(n+1,talk_last,m)
    } else {
      free m
    }} | reply n,talk_last to wait_last
        
;;

spawn { free 0 }
;;
    
(* synchronizing barrier *)
let bidon = (white,0,0,0,0,0)
;;

let def waiting (n,nplayers,status,n_deads) | sync_status! (me,pos,is_dead) =
  Array.set status me pos ;
  let n_deads = if is_dead then n_deads+1
  else n_deads in
  {if n+1 >= nplayers then
    {start(0,nplayers,status,n_deads)}
  else
    {waiting(n+1,nplayers,status,n_deads)}} |
  let n_deads = wait () in
  reply status,nplayers=n_deads to sync_status

or waiting(n,nplayers,status,n_deads)| init_status! (new_nplayers,new_status) =
    waiting(0,new_nplayers,new_status,0) |
    reply to init_status

or start(n,nplayers,status,n_deads) | wait! () =
  {if n+1 >= nplayers then
    {waiting(0,nplayers,status,n_deads)}
  else
    {start(n+1,nplayers,status,n_deads)}} |
  reply n_deads to wait
;;

spawn { waiting(0,0,Array.create 0 bidon,0)}
;;

let width = 8
and skip_init = 6 (* skip length   *)
;;

let colors = [|red;blue;green;yellow;magenta;cyan|]

let def  new_game! my_loc =

  let loc game
    with
      message! (color,s) =
         printnl_string s;
         set_color background ;
         fill_rect 0 (size_y()-20) (size_x()) 20;
         set_color color ;
         moveto 10 (size_y()-20) ;
         draw_string s;
         reply ()

     or true_game! () =
       open_graph "";

       set_line_width width ;
       let xmax = size_x()/width+1
       and ymax = size_y()/width+1 in

         
       let board = create_matrix xmax ymax true in

       let def get_board! (x,y) = reply board.(x).(y)
       and set_board! (x,y) = (board.(x)).(y) <- false; reply
       and clear_board! () =
         let def clear_col! l = Array.fill l 0 ymax true ; reply
         in
         let def do_clear! i =
           if i >= 0 then
             { clear_col board.(i); reply (do_clear (i-1))}
           else reply
         in
         do_clear (xmax-1) ; reply 
       and inside! (x,y,s) =
         if (x >= 0) && (x < xmax) && (y >= 0) && (y < ymax) then {
           if s <= 0 then reply get_board (x,y) else reply true
         } else reply false
       in

       let def left! (dx,dy)  =
         if dx>0 then reply 0,dx
         else if dx < 0 then reply 0,dx
         else if dy > 0  then reply 0-dy,0
         else reply 0-dy,0

       and right! (dx,dy) =
         if dx>0 then reply 0,0-dx
         else if dy < 0 then reply dy,0
         else if dx>0 then reply 0,0-dx
         else reply dy,0 in

       let def init_pos! (i,m) =
          let y0 = (ymax*(1+i))/(m+1) and x0 = 0 in
          let col = colors.(i) in
          let pos = (col,x0,y0,x0,y0,0) in
          reply pos,col in

       let def get_color! (status,n) =
         let  col,x,y,dx,dy,s = status.(n) in
         reply col in

       let def draw_me! (pos) =
         let col,x,y,dx,dy,s = pos in
         (if s <= 0 then
           (set_board (x,y) ;
            set_color col;
            moveto (width*x) (width*y) ;
            lineto (width*dx) (width*dy)
              ));
         reply in              
       let def draw_game! status =
         let def do_it! (i) =
           if i < 0 then {reply}
           else {
             draw_me (status.(i)) ;
             reply (do_it (i-1))
             } in
           reply (do_it ((Array.length status)-1)) in  

       let def one_step! (n,status,dx,dy) =
         let pos = status.(n) in
         let col,no1,no2,x,y,s = pos in
         let nx = x+dx and ny = y+dy in
         if inside (nx,ny,s) then {
           let new_pos = (col,x,y,nx,ny,if s > 0 then s-1 else s) in
           status.(n) <- new_pos ;
           draw_me (new_pos) ;
           reply true
         } else {
           reply false
         } in
       let def treat_key! (c,n,status,dx,dy) =
         if c="j" then {
           let dx,dy = left (dx,dy) in
           reply dx,dy
         } else if c="k" then {
           let dx,dy = right (dx,dy) in
           reply dx,dy
         } else if c=" " then {
           let pos = status.(n) in
           let col,x,y,nx,ny,s = pos in
           (if s <= 0 then
             status.(n) <- (col,x,y,nx,ny,skip_init));
            reply dx,dy
         } else {reply dx,dy} in

       let def read! () =
         if key_pressed() then {
           let s = String.make 1 (read_key()) in
           reply s}
         else {reply ""} in
       let def empty_keys! () =
         if key_pressed() then {
           let c = read_key () in empty_keys () ; reply
         } else {reply} in

(* starting *)      
       let def wait_start(status,n) =
          draw_game (status) ;
          let new_status,all_dead =
            sync_status (n,status.(n),false) in
          if read () = "" then {
             wait_start(new_status,n)
          } else {
             play_game(new_status,n,1,0)
          } 

(* Playing *)
       or play_game(status,n,dx,dy) =
         draw_game (status) ;
         let dx,dy = treat_key (read (),n,status,dx,dy) in
         if one_step (n,status,dx,dy) then {
            let new_status,all_dead =
              sync_status (n,status.(n),false) in
            play_game(new_status,n,dx,dy)            
         } else {
           let new_status,all_dead =
             sync_status (n,status.(n),true) in
            message (get_color (status,n),"T'es mort !!!") ;
            game_over(new_status,all_dead,n)
         }

(* Stopping = waiting for everybody to die *)
       or game_over(status,all_dead,n) =
         if all_dead then {
           start()}
         else {
           draw_game (status) ;
           let new_status,all_dead =
             sync_status (n,status.(n),false) in
           game_over(new_status,all_dead,n)
         }

     or talk_start n | other(start_other,m) =
         {if n+1 < m then start_other(n+1) } |
          let init_status,col = init_pos (n,m) in
          dgc_restart ();
          dgc_stop ();
          let new_status,all_dead =
            sync_status (n,init_status,false) in
          message (col,"j =)  gauche, k =)  droite, SPC =) dmarrer/sauter" );
          empty_keys () ;
          wait_start(new_status,n)
          

       or start() =
           clear_graph() ;
           clear_board () ;
           message (black,"Un peu de patience") ;
           incr ts;
           let ts = !ts in
         let nplayers,me,start_other = start_game (talk_start,ts) in
         { if me=0 then 
           { let bidon = (white,0,0,0,0,0) in
           init_status (nplayers,Array.create nplayers bidon) ;
           talk_start me}
         } |
         { 
           other(start_other,nplayers)} in
       start() | reply
    or local_init! () =
       Join.go my_loc ; 
       join_game () ; 
       true_game () ; reply
     = struct end
  in
  local_init () ; reply
;;


  
let _ = 
  Nfs.register "tron-server";  
  Join.setLocalService "play" new_game vartype;
  Join.server ()
;;

