open ml
open graphics
open array

let me = if ns.server then 0 else 1
;;

let width = 8
and k0 = 0 (* time constant *)
and s0 = 6 (* skip length   *)
;;


do open_graph("")
;;

do set_line_width(width)
;;

let xmax = size_x()/width+1
and ymax = size_y()/width+1


let col0 = if me=0 then red else blue
and y0 = if me=0 then ymax/3 else 2*ymax/3
;;

let board = create_matrix(xmax,ymax,true)

let get_board(x,y) =
  let l = array.get(board,x) in
  reply array.get(l,y)

let set_board(x,y) =
  let l = array.get(board,x) in
  array.set(l,y,false) ;
  reply
;;

let clear_board() =

  let clear_col(l) = array.fill(l,0,ymax,true) ; reply in

  let do_clear(i) =
    if i >= 0 then {clear_col(array.get(board,i)) ; do_clear(i-1) ; reply}
    else {reply} in
  do_clear(xmax-1) ; reply
;;        

let inside(x,y,s) =
  if (x >= 0) && (x < xmax) && (y >= 0) && (y < ymax) then {
    if s <= 0 then {reply get_board(x,y)} else {reply true}
 } else reply false
;;

let left(dx,dy) =
  if dx=1 then reply 0,1
  else if dx = -1 then reply 0,-1
  else if dy=1 then reply -1,0
  else reply 1,0

let right(dx,dy) =
  if dx=1 then reply 0,-1
  else if dx = -1 then reply 0,1
  else if dy=1 then reply 1,0
  else reply -1,0

let message(color,s) =
  set_color(background) ;
  fill_rect(0,size_y()-20,size_x(),20);
  set_color(color) ;
  moveto(10,size_y()-20) ;
  draw_string(s) ;
  reply
;;
do ns.register("message"^string_of_int(me),message)

let other_message = ns.lookup("message"^string_of_int(if me=0 then 1 else 0))
;;

let waiting =
  if ns.server then
    (let waiting() | No() = Yes() | reply true to waiting
    and waiting() | Yes() = No() | reply false to waiting in
    spawn {No()} ; ns.register("waiting",waiting) ; waiting)
  else ns.lookup("waiting")
;;

let new_cell() =
 let E() | set(x) = F(x) | reply to set
 and F(x) | get() = F(x) | reply x to get in
 E() | reply set,get
;;

let set,get = new_cell()
;;

do ns.register("get"^string_of_int(me),get)
;;

let get_other = ns.lookup("get"^string_of_int(if me=0 then 1 else 0))
;;


let my_set,my_get = new_cell()
;;

let sync =
  if ns.server then
    let sync1() | sync2() = reply to sync1 | reply to sync2 in
    ns.register("sync",sync2) ;
    sync1
  else
   ns.lookup("sync")
;;

let trace(other_other) | pos(col,x,y,dx,dy,s) =
   if s <= 0 then {
     if inside(x+dx,y+dy,s) then {
       set_color(col) ; moveto(width*x,width*y) ;
       lineto(width*(x+dx),width*(y+dy)) ;
       set_board(x+dx,y+dy) ;
       other_other(col,x,y,dx,dy) ;
       pos(col,x+dx,y+dy,dx,dy,s) | reply true to trace
     } else {reply false to trace}
   } else {
      if inside(x+dx,y+dy,s) then {
        pos(col,x+dx,y+dy,dx,dy,s-1) | reply true to trace
      } else {reply false to trace}
   }

and other(col,x,y,dx,dy) =
    set_color(col) ; moveto(width*x,width*y) ;
    lineto(width*(x+dx),width*(y+dy)) ;
    set_board(x+dx,y+dy) ;

and key(c) | pos(col,x,y,dx,dy,s) =
  if c = "j" then {
    let dx,dy = left(dx,dy) in
    pos(col,x,y,dx,dy,s)
  } else if c = "k" then {
    let dx,dy = right(dx,dy) in
    pos(col,x,y,dx,dy,s)
  } else if (c = " ") && (s <= 0) then {
    pos(col,x,y,dx,dy,s0)
  } else if (c = "l") then {
    if ml.abs(dx)+ml.abs(dy) > 1 then {
      pos(col,x,y,dx/2,dy/2,s)
    } else {
      pos(col,x,y,dx*2,dy*2,s)
    }
  } else {pos(col,x,y,dx,dy,s)}

and sleep(k) =
  if k <= 0 then {
    if trace(my_get()) then {sleep(k0)}
    else {
      message(col0,"Domage !!!") ;
      other_message(col0,"Je me suis viand") ;
      sync() ;
      message(col0,"Pour une nouvelle bourre ``n'', pour aller se prendre une mousse ``q''") ;
      over()}
  } else {sleep(k-1)}

and over() | key(c) =
  if c="n" then {
    {if waiting() then
      {other_message(col0,"On s'en refait une petite ??? (touche ``n'')");}} |
     clear_graph() ; sync() ; start()}
  else if c="q" then {
    {other_message(col0,"Je me casse, salut") ;} | failwith("Adieu !!!")
  } else {over()}

and wait_start(col,x,y,dx,dy) | key(c) =
  other_message(col0,"Je t'aurai au tournant...");
  sleep(0) | pos(col,x,y,dx,dy,0)

and start() =
  clear_board() ;
  set_color(col0) ;
  fill_rect(0,y0*width-width/2,width,width) ;
  message(col0,"Pour mettre les gaz, presser une touche.");
  wait_start(col0,0,y0,1,0)
;;

do set(other) ; let x = get_other() in my_set(x)
;;


let reader () =  
  if key_pressed() then {
    print_string("=> ") ; ml.flush(ml.stdout) ;
    let c = string.make(1,read_key()) in
    print_string(c) ; print_newline() ;
    key(c) | reader()
  } else {
    reader()
  }

spawn start() | reader()
;;

