open Jtypes
open Jprims
open Space

external dump_heap : unit -> unit = "dump_heap"

let doit = ref true

let local_gc () = 
  let spaces = ref [] in
  let stubs = ref 0 in
  critical_section (fun () ->
    Hashtbl.iter 
      (fun addr space ->
        spaces := space :: !spaces;
        ) Space.spaces;
    );
  List.iter 
    (fun space ->
      let life_stubs = ref [] in
      let lifes = ref 0 in
      let deads = ref 0 in
      critical_section (fun () ->
        Whashtbl.iter (fun id gc_stub_opt ->
          Debug.print2 'g' "GC test : %s date %d\n" 
            (Space.string_of_uid (fst id)) (snd id);
          match gc_stub_opt with
            None -> 
              begin
                incr deads;
                Whashtbl.remove space.stubs_table id;
                try
                  let proxy = 
                    Whashtbl.find Space.stubs_table (fst id) in
                  match proxy with
                    None -> 
                      Debug.print0 'g' "Proxy dead\n"
                  | Some pointer ->
                      Debug.print0 'g' "Proxy still alive\n"
                with
                  Not_found ->
                    Debug.print0 'g' "Proxy not found\n"
              end
          | Some gc_stub ->
              Debug.print0 'g' "Still alive\n";
              life_stubs := id :: !life_stubs;
              incr lifes;
              ) 
          ) space.stubs_table;
      Debug.print2 'g' "GC: Deads : %d Lifes: %d" !deads !lifes;
            (* on envoie le message *)
      
      Space.send space 
        (Jmarshal.marshal space 
           (Life (space.threshold,
                  List.map (fun (uid,age) -> uid,age) !life_stubs))
           )
        ) !spaces;
    critical_section (fun () ->
      Whashtbl.iter (fun uid stub_op ->
        match stub_op with
          None -> Whashtbl.remove stubs_table uid
        | Some _ -> incr stubs
              ) stubs_table );
  Debug.print1 'g' "NBre de stubs : %d \n" !stubs
    ;;

let threads = ref 0
    
let init () =
  try
    Sys.getenv "JNODGC"; ()
  with
    Not_found ->
      Thread.create 
        (fun () ->
          
            while true do
              wait_on_gc ();
              Debug.print0 'g' "Major GC\n";
              if !doit then local_gc ()
            done
            ) ();
(*
      incr threads;
      Thread.create
        (fun () ->
          while true do
            Thread.delay 600.; (* 60 secondes pour les tests, 10 min. apres *)
            if !doit then 
              Gc.major ()
          done
          ) ();
*)
      incr threads


        
let life_handler space threshold life_stubs =
  Debug.print0 'g' "LIFE";
  let deads = ref 0 in
  let lifes = ref 0 in
  let suspects = ref 0 in
  let life_scions = ref [] in
  critical_section (fun () ->
    Hashtbl.iter (fun id gc_scion ->
      if List.mem id life_stubs then
        life_scions := (id,gc_scion) :: !life_scions
      else
        if gc_scion.scion_ts >= threshold then
          begin
            life_scions := (id,gc_scion) :: !life_scions;
            incr suspects
          end
        else
          begin
            Debug.print2 'g' "Delete scion %s age %d\n" (Space.string_of_uid (fst id)) (snd id);
            let scion = gc_scion.scion in
            incr deads;
            scion.gc_scion_counter <-
               scion.gc_scion_counter - 1;
            if scion.gc_scion_counter = 0 then
              begin
                Debug.print0 'g' "-";
                Hashtbl.remove Space.scions_table id;
              end          
          end
            ) space.scions_table;
    Hashtbl.clear space.scions_table;
    let failed = ref [] in 
    List.iter (fun (id,gc_scion) -> 
      incr lifes;
      let o = gc_scion.scion.object in
      if (Obj.magic o:stub).position == None then
        failed := (fst id) :: !failed;
      Hashtbl.add space.scions_table id gc_scion;
      ) !life_scions;
    if !failed <> [] then
      Space.send space 
        (Jmarshal.marshal space 
           (Failed !failed))
        );
  
  Debug.print3 'g' "LIFE: Deads %d Lifes: %d Suspects:%d" !deads !lifes !suspects

let fail_handler space list =
  List.iter (fun uid ->
    find_stub uid
    ) list

let stop () = doit := false
let restart () = doit := true; local_gc ()
