(***********************************************************************)
(*                                                                     *)
(*                    ______________                                   *)
(*                                                                     *)
(*      Fabrice Le Fessant, projet SOR/PARA INRIA Rocquencourt         *)
(*                                                                     *)
(*                 mail: fabrice.le_fessant@inria.fr                   *)
(*                 web:  http://www-sor.inria.fr/~lefessan             *)
(*                                                                     *)
(*  Copyright 1997 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(*                                                                     *)
(***********************************************************************)

(*         File written in the Objective-CAML language                 *)

open Jtypes;;

let threads = ref 1

let migrate orig dest =
  let (dst_locator, dest_space, _,_,_) =  Obj.magic dest in
  if (Obj.magic dest_space) == None then Agent.halt_location orig
  else
    let (src_locator, _, _, _, _) = Obj.magic orig in
    let (modules, m) = Agent.marshal dest_space orig in
    Debug.print ("DEST : "^(string_of_int (Obj.tag (Obj.repr dest))));
    let (ts,_,_) as message = 
      (Jmarshal.marshal dest_space
         (LocationMigration (modules, m, dest)));
    in
  (* ici, on doit creer des stubs pour tous les objets migrant *)
    Agent.insert_migration_stubs dest_space ts m;
    Space.send dest_space message
      ;;

let go_thread =
  Thread.create
    (fun () ->
       while true do
         let (origine, destination) = Jprims.wait_on_stub () in
         Printexc.catch (fun () -> migrate origine destination) ();
         Jprims.set_critical_section false;
       done)
    ()
;;


let needed () = 
  Debug.print "Set GO THREAD ...........................";
  Jprims.set_go_thread go_thread;;


