(***********************************************************************)
(*                                                                     *)
(*                    Join : automata.ml                               *)
(*                                                                     *)
(*      Fabrice Le Fessant, projet SOR/PARA INRIA Rocquencourt         *)
(*                                                                     *)
(*                 mail: fabrice.le_fessant@inria.fr                   *)
(*                 web:  http://www-sor.inria.fr/~lefessan             *)
(*                                                                     *)
(*  Copyright 1997 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

(*         File written in the Objective-CAML language                 *)

open Jtypes;;

external filter_channel : (('a => 'b) as 'c) -> 'c = "%identity"

let go = Jprims.go;;
let halt () = 
  let location = Jprims.self_location () in
  Agent.halt_location location

type location = Jtypes.location

(* Join-calculus *)

let bar process1 process2 =
  Thread.safe_create process1 (); process2 ()
;;

let launch process = Thread.safe_create process (); ();;

let server () = Jprims.server ()
let here = Jprims.self_location ()

let new_state (automata, local_automata) = ();;
(*
  let state = ref 0 in
  for i = 0 to (Array.length local_automata.buffers) -1 do
    if local_automata.buffers.(i) <> [] then
      state := (!state) lor (1 lsl i)
  done;
    local_automata.state <- !state
*)
  
Space.join_handler := Handler.handler;
Space.join_initialised := true;
();; (* on autorise les connections *)
  


(* *)

(* L'interface *)

type process = Jtypes.process;;
type automata = Jtypes.automata;;
type channel = Jtypes.channel;;
type module_data = Reloc.module_data;;
type thread_stub = Jtypes.thread_stub;;
type reply_event = thread_stub
type space_id = Jtypes.space_id


let send_asynchrone_and_exit channel v = 
  Channel.send_asynchrone channel v;
  Thread.exit ();;
let send_asynchrone channel v = 
  Channel.send_asynchrone channel v

let send_synchrone channel v = channel v;;
let create_empty_automata_in = Jprims.create_empty_automata_in
let create_empty_automata n = 
  Jprims.create_empty_automata_in (Jprims.self_location ()) n 
let fill_automata = Jprims.fill_def;;

let reply_to th v = 
  Channel.reply_to th (Reply (Obj.repr v))

let reply_to_and_exit th v = 
  Channel.reply_to_and_exit th (Reply (Obj.repr v))

let send_exception exn thread_stubs = Channel.send_exception exn thread_stubs

let _ = Go.needed ()

let getRemoteService = Services.getRemoteService;;
let setLocalService = Services.setLocalService;;
let getLocalService = Services.getLocalService;;
let getSpaceService = Services.getSpaceService

let create_agent = Agent.create;;
let agent_thread location f = Thread.create_in location f ()
let create_location module_data f =
  let location = create_agent module_data
  in
  agent_thread location f;
  location

let exit () = Thread.exit ()

let tail_synchrone = Channel.tail_synchrone

let attach (module_data:Reloc.module_data) =
  Jprims.attach_to_location module_data

let reuse_thread thread_stub clos = Jprims.reuse_thread clos () thread_stub

let print i =
  print_int i;
  flush stdout

let definition () =
  let (automata,defun,declare) = create_empty_automata (* join.mlg 14 *)
      1 2 (1,0)
  in
  let process = function 
      (x,(reply_defun,())) ->
        reply_to_and_exit (* join.mlg 17 *) reply_defun x
  in
  fill_automata automata [|Obj.magic process|] (0,2,(3,0),(3,0),((0,2),0,1));
  ((declare,defun): (<<'a>>) * (unit => 'a))

let _ = Dgc.init ()
let dgc_stop = Dgc.stop
let dgc_restart = Dgc.restart

external test_channel' : 's -> string -> unit = "test_good_channel"
let test_channel c s = test_channel' s c

(*
  Toutes les locations devront etre des objets. Les plus simples
deriveront de la classe agent.
*)

(*
class location () : 'a =
  method go (location:< >) = go location
  method kill = halt (); ()
end


let create_agent module_data =
  let l = new location () in
  Jprims.set_location_module l module_data;
  l
*)

let goo oo = go (Obj.magic oo: location)
let fail = Thread.fail

let new_port = Space.new_port
