(***********************************************************************)
(*                                                                     *)
(*                    ______________                                   *)
(*                                                                     *)
(*      Fabrice Le Fessant, projet SOR/PARA INRIA Rocquencourt         *)
(*                                                                     *)
(*                 mail: fabrice.le_fessant@inria.fr                   *)
(*                 web:  http://www-sor.inria.fr/~lefessan             *)
(*                                                                     *)
(*  Copyright 1997 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(*                                                                     *)
(***********************************************************************)

(*         File written in the Objective-CAML language                 *)

open Jtypes;;

(* extern.c *)

external marshal_message :
  'a -> string * Obj.t array = "join_marshal_message";;

(* intern.c *)

external unmarshal_message :
  string -> Obj.t array -> 'a = "join_unmarshal_message";;

(* scheduler.c *)
external wait_on_stub' : unit -> 'a = "stub_thread_wait";;
external wait_on_gc': unit -> unit = "thread_wait_gc";;
external exit_wait_on_stub' : unit -> 'a = "stub_thread_wait";;
external reply_thread_stub : thread_stub -> 'b -> unit = "stub_thread_reply";;
external set_go_thread : thread -> unit = "set_go_thread";;
external schedule_go_thread' : unit -> unit = "schedule_go_thread";;
external extract_location_threads :
  location -> thread_stub array = "extract_location_threads";;
external marshal_thread :
  thread_stub -> thread_marshalled = "marshal_thread";;
external unmarshal_thread :
  thread_stub * thread_marshalled -> unit = "unmarshal_thread";;
external self_location : unit -> location = "thread_self_location";;
external set_location : location -> unit = "thread_set_location";;
external set_local : 'a -> 'a = "set_local";;
external print_stats: unit -> unit = "print_stats"
external trigger_def_and_wait' : automata * int * 'a -> 'b = "trigger_def_and_wait";;
external exit_trigger_def_and_wait' : automata * int * 'a -> 'b = "exit_trigger_def_and_wait";;
external server : unit -> unit = "set_server"
(* local.c *)

       (* stubs *)
external test_stub: 'a -> string -> unit = "test_good_stub"
external is_local_stub : 'a -> bool = "is_local_stub";;
external get_stub_info' : 'a -> 
  int * (locator option) * space * location * bool * gc_stub * bool
    = "get_stub_info";;
          (* (tag,locator,space,location,localp,reserved) *)
external get_stub_stub: 'a -> gc_stub = "get_stub_stub"
external set_stub_stub: 'a -> gc_stub -> unit = "set_stub_stub"
external set_stub_locator: 'a -> (locator option) -> 
  unit = "set_stub_locator"
external set_stub_space : 'a -> space -> unit = "set_stub_pos"
external create_space_stub : int -> space -> Obj.t = "create_stub";;
external create_location_stub : int -> location -> Obj.t = "create_stub";;
external halt_stub: 'a -> unit = "halt_stub"

       (* objects *)
external copy: Obj.t -> int -> int -> Obj.t = "obj_copy"
external create_object_stub : space ->  Mop.bucket array -> int -> 
  Obj.t = "create_object_stub"
external set_replicat: Obj.t -> Obj.t -> unit = "set_replicat"
external create_new_object : space -> location -> int -> Reloc.module_data ->
  Obj.t = "create_new_object";;
external set_location_module : < .. > -> Reloc.module_data -> unit =
  "set_location_module"
external get_location_module: < .. > -> Reloc.module_data =
  "get_location_module"
       (* Threads *)
external get_thread_stub : thread -> thread_stub = "get_thread_stub";;
external self_stub : unit -> thread_stub = "self_stub";;
external reuse_thread : ('a -> unit) -> 'a -> thread_stub ->
  unit = "reuse_thread"

       (* Locations *)
external create_new_location : space -> location -> Reloc.module_data ->
  location = "create_new_location";;
external create_first_location : space -> unit -> unit -> 
  location = "create_new_location";;
external marshal_location : space -> location -> thread_stub array * 
    automata array * location array * Reloc.module_data
    * (Reloc.module_data list) * (Obj.t array)
    = "marshal_location";;
external unmarshal_location : location -> location -> Reloc.module_data ->
  Reloc.module_data list -> (Obj.t array) ->
  unit = "unmarshal_location";;
external go' : location -> unit = "go_location";;
external halt_location' : location -> 'a = "halt_location";;

       (* interfaces *)
external attach_to_location: Reloc.module_data -> unit = "attach_to_location"
external get_interface: Reloc.module_data -> interface = "get_interface"
external set_interface: Reloc.module_data -> interface -> unit
    = "set_interface"

        (* channels *)
external test_channel': 'a -> string -> unit = "test_good_channel"
external create_empty_automata_in : location -> int -> int -> 'b -> 'a
    = "create_empty_automata_in";;
external fill_def : automata -> process array -> 'a ->
  unit = "fill_def";;
external trigger_def' : automata * int * 'a -> unit = "trigger_def";;
external unmarshal_def : automata * automata_marshalled -> 
  unit = "unmarshal_def";;
external marshal_def : automata -> automata_marshalled = "marshal_def";;

       (* Sends *)
external set_locator_send :
  (space -> 'a -> unit) -> unit = "set_locator_send";;
external send_location : location -> protocole -> unit = "location_send";;

external breakpoint : int -> 'a -> unit = "breakpoint"

(* Wrappers : C functions which call schedule_thread() must be called from
   a ML wrapper with exactly one parameter *)
let wait_on_stub () = wait_on_stub' ();;
let trigger_def_and_wait args = trigger_def_and_wait' args
let exit_wait_on_stub () = exit_wait_on_stub' ();;
let exit_trigger_def_and_wait args = exit_trigger_def_and_wait' args
let trigger_def args = trigger_def' args
let schedule_go_thread () = schedule_go_thread' ();;

external set_critical_section:  bool -> bool = "set_critical_section"
external thread_dirty: thread -> unit = "thread_dirty"

let go l =  (* on entre immediatement dans une section critique. En fait
  le thread de migration va etre schedule dans location_go et la section 
               critique sera quittee dans cet autre thread *)
  set_critical_section true;
  go' l;;


let get_stub_locator stub =
  match (Obj.magic stub:stub).locator with
    None -> failwith "No locator"
  | Some locator -> locator

let critical_section f =
  let old = set_critical_section true in
  try
    let r = f () in
    set_critical_section  old;
    r
  with
    e -> 
      set_critical_section old;
      raise e

let wait_on_gc ()= wait_on_gc' ()
let test_channel c s = test_channel' s c

let halt_location location = halt_location' location

external register_send_sync : code_t -> env_t -> unit = "register_send_sync"
