(***********************************************************************)
(*                                                                     *)
(*                    JOCAML                                           *)
(*                                                                     *)
(*      Fabrice Le Fessant, projet SOR/PARA INRIA Rocquencourt         *)
(*                                                                     *)
(*                 mail: fabrice.le_fessant@inria.fr                   *)
(*                 web:  http://www-sor.inria.fr/~lefessan             *)
(*                                                                     *)
(*  Copyright 1997 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(*                                                                     *)
(*         File written in the Objective-CAML language                 *)
(***********************************************************************)

open Thread

type port = int;;
type time = float;;
type process = Obj.t -> unit;; 
type timestamp = int;;
type interface
type thread = Thread.thread
type location = Thread.location
type stub_ts;;

type automata;;



type thread_marshalled;;
type automata_marshalled;;
         
type space_id = (Unix.inet_addr * port) * time
type uid =
    Uid of space_id * int

type connection =
  { conn_inc : in_channel;
    conn_outc : out_channel;
    conn_thread : thread
  }
;;
type otype =
    Location of space_id * ((Mop.class_id * ((string*int*int) list) * int) option)
  | Automata of locator (* location locator, of course *)
  | Thread of locator (* location locator, of course *)

and locator =
    {
    mutable age: int;
    mutable distance: int;
    mutable uid: uid;
    mutable otype: otype
  } 

type obj_locator =
    RelocModule of Reloc.module_data 
  | DummyModule of string * int
  | RelocException of string * string * int
  | StubLocator of locator

type message = int *  string * obj_locator array

type gc_stub = (Obj.t*space*int) option

and space =
  { space_id : space_id;
    queue: message Fifo.t;
    mutable sender_thread: thread;
    mutable threshold : int;
    ext_buckets : (Mop.class_id , Mop.bucket array) Hashtbl.t;
    mutable next_ts: int;
    mutable set: int;
    mutable last_ts: int;
    mutable modules: string list;
    scions_table : (uid*int, gc_scion) Hashtbl.t;
    mutable link : connection option;
    stubs_table : (uid*int, gc_stub) Whashtbl.t
  }
and gc_scion =
    {
    mutable scion_ts: int;
    source: space;
    scion: scion
  } 
and scion =
    {
    mutable cyclic_ts: int;
    scion_age: int;
    object: Obj.t;
    mutable gc_scion_counter: int;
    mutable old_stub: gc_stub
  }

(* failed si local = None et 
   locator = None ou locator = Some { locator.age = -1 } 
*)
type stub = (* a mask to read stubs data *)
    {
    mutable locator:  locator option;
    mutable position: Obj.t option; (* space or location *)
    mutable local:    Obj.t option; (* ldef, lloc ou self *)
    mutable failstruct:  Obj.t list;
    mutable gc_stub:  gc_stub
  } 

type code_t
type env_t
type channel = code_t * env_t * automata * int;;

type thread_stub

type location_marshalled =
    LocationBloc of location *
        (thread_stub * thread_marshalled) array *
        (automata * automata_marshalled) array *
        location_marshalled array *
        (Reloc.module_data * interface) list *
        Reloc.module_data *
        (Obj.t array)
;;

type reply_type = 
    Reply of Obj.t 
  | Exception of exn;;

type protocole =
    GetService of string * Obj.t * thread_stub
  | Synchrone of (*automata*) uid * int * Obj.t * thread_stub
  | Invoke of (*Obj.t*) uid * string * int * (Obj.t list) * thread_stub
  | Asynchrone of (*automata*) uid * int * Obj.t
  | LocationMigration of string list * location_marshalled * location
  | ReplyThread of (*thread_stub*) uid * Obj.t
  | Life of timestamp * ((uid*int) list)
  | Failed of uid list
  | Close of (* scions *) (uid * int) list * (* stubs *) ((uid*int) list)

let failed_age = 1 lsl 30
