(***********************************************************************)
(*                                                                     *)
(*                    ______________                                   *)
(*                                                                     *)
(*      Fabrice Le Fessant, projet SOR/PARA INRIA Rocquencourt         *)
(*                                                                     *)
(*                 mail: fabrice.le_fessant@inria.fr                   *)
(*                 web:  http://www-sor.inria.fr/~lefessan             *)
(*                                                                     *)
(*  Copyright 1997 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(*                                                                     *)
(***********************************************************************)

(*         File written in the Objective-CAML language                 *)

(*
  L'algo est bancal.
  
  Tout stub pouvant etre reference depuis l'exterieur possede au
  moins un gc_stub. Si l'objet est local, le gc_stub pointe vers
  l'espace courant.
*)

open Jtypes;;
open Reloc

    (*
  les recherches dans les tables de hashage sont nombreuses, et portent
souvent sur la recherche d'une meme cle dans plusieurs tables differentes.
Il faudra optimiser ces recherches repetees (une seule table de hashage
avec plosieurs champs pour une meme cle OU on hashe une fois la valeur
puis on recherche dans toutes les tables en utilisant cette valeur)
*)

let rec loc2ptr src_space ts locator =
  locator.distance <- locator.distance + 1; 
  Debug.print1 'l' "loc2ptr: %s\n" (Space.string_of_uid locator.uid);

  Debug.print "<-------- loc2ptr";
  try
    let pointer =
      match Whashtbl.find Space.stubs_table locator.uid with
        None -> 
          Whashtbl.remove Space.stubs_table locator.uid;
          raise Not_found
      | Some pointer -> pointer
    in
    let (Some old_locator,position,local,_,old_gc_stub) = Obj.magic pointer 
    in
    if (old_locator.age < locator.age) || 
    (old_locator.distance > locator.distance) then
      begin
        match locator.otype with
          Automata in_locator -> loc2ptr src_space ts in_locator;()
        | Thread in_locator   -> loc2ptr src_space ts in_locator;()
        | Location (space_id,oo_desc) ->
            if old_locator.age < locator.age then
              begin
                Debug.print "3";
                old_locator.otype <- locator.otype;
                Jprims.set_stub_space pointer (Space.get_space space_id)
              end
      end;
  (* 2.1 *)
    if old_locator.age < locator.age then
    (* the new locator is younger than the old ... *)
      begin
        if locator.age <> failed_age then (* migration *)
          begin
            let gc_stub = Some (pointer,src_space,locator.age) in
            Jprims.set_stub_stub pointer gc_stub;
            Whashtbl.add src_space.stubs_table (old_locator.uid,locator.age)
              (Some gc_stub);
          end
        else (* pas de gcstub pour les failed proxies *)
          Jprims.halt_stub pointer;
        begin try
          let scion = Hashtbl.find Space.scions_table
              (old_locator.uid,old_locator.age)
          in
          scion.old_stub <- old_gc_stub;
          Hashtbl.remove Space.scions_table (locator.uid,old_locator.age);
        with Not_found -> () end;
        old_locator.age <- locator.age;
        old_locator.distance <- locator.distance;
      end
    else
      if old_locator.distance > locator.distance then
        begin
          Debug.print "5";
          old_locator.distance <- locator.distance;
          let gc_stub = Some (pointer,src_space,locator.age) in
          Jprims.set_stub_stub pointer gc_stub;
          Whashtbl.add src_space.stubs_table (locator.uid,locator.age) (Some gc_stub)
        end;
    Debug.print "---> loc2ptr";
    pointer
  with
    Not_found ->
      Debug.print "Not found";
      let pointer =
        match locator.otype with
          Automata in_locator ->
            let location = Obj.magic (loc2ptr src_space ts in_locator) in
            locator.otype <- Automata
                 (Jprims.get_stub_locator location);
            Jprims.create_location_stub Tag.automata location
        | Thread in_locator ->
            let location = Obj.magic (loc2ptr src_space ts in_locator) in
            locator.otype <- Thread
                 (Jprims.get_stub_locator location);
            Jprims.create_location_stub Tag.thread location
        | Location (space_id,None) ->
            Jprims.create_space_stub Tag.location
              (if space_id = Space.local_id then
                      (* a new location which is migrating to this space *)
                Space.local_space
              else
                Space.get_space space_id)
        | Location (space_id,Some(class_id,class_sig,obj_size)) ->
            let space = if space_id = Space.local_id then
                      (* a new object which is migrating to this space *)
              Space.local_space
            else
              Space.get_space space_id in
            Debug.print0 'u' "Unmarshal object\n";
            Jprims.create_object_stub space
              (Channel.remote_ext_buckets space class_id class_sig) obj_size
      in
      let gc_stub = Some (pointer,src_space,locator.age) in
      Whashtbl.add Space.stubs_table locator.uid (Some pointer);
      Jprims.set_stub_locator pointer (Some locator);
      Jprims.set_stub_stub pointer gc_stub;
      Whashtbl.add src_space.stubs_table (locator.uid,locator.age) (Some gc_stub);
      Debug.print "---> loc2ptr";
      pointer

exception MessageRefused;;

let unmarshal (space : space) message =
  Debug.print0 'l' "Unmarshal ...";
  let (ts, buffer, locators) = message in
  space.last_ts <- ts;
  if ts < space.threshold then raise MessageRefused;
  let pointers =
    Array.map 
      (function 
          DummyModule (name,size) ->
            Debug.print ("find module"^name);
            Obj.repr
              (let m = find_module_data name size 
              in
              Debug.print2 'r' "%s ---> %d\n" m.module_name
                m.module_slot;
              m
                )
        | RelocModule module_data -> 
            let m = find_module_data module_data.module_name
                (Obj.size (Obj.repr module_data))
            in
            Debug.print1 'x' "install module %s\n" module_data.module_name;
            install_module module_data m;
            Obj.repr m
        | RelocException (exn_name,mod_name,mod_num) ->
            Obj.repr (find_exn exn_name mod_name mod_num)
        | StubLocator locator -> 
             Debug.print ("locator");
             loc2ptr space ts locator) locators
  in
  let x =     Jprims.unmarshal_message buffer pointers
  in
  Debug.print0 'l' "... Unmarshaled";
  x
;;
