/***********************************************************************/
/*                                                                     */
/*                           Join-Calculus                             */
/*                                                                     */
/*            Fabrice Le Fessant, projet PARA, INRIA Rocquencourt      */
/*                                                                     */
/*  Copyright 1998 Institut National de Recherche en Informatique et   */
/*  Automatique.  Distributed only by permission.                      */
/*                                                                     */
/***********************************************************************/

#include "callback.h"
#include "config.h"
#include "misc.h"
#include "mlvalues.h"
#include "stacks.h"
#include "fail.h"
#include "io.h"
#include "roots.h"
#include "alloc.h"
#include "memory.h"
#include "signals.h"
#include "gc_ctrl.h"
#include "intext.h"
#include "weak.h"
#include "gc.h"

#include "join.h"
#include "threads.h"

/*********

   module Stub

*********/

value is_local_stub(value stub) /* ML */
{
  return Val_int(Field(stub,STUB_LOCAL) != Val_unit);
}

value is_failed_stub(value stub)
{
  return Val_int(Field(stub,STUB_POS) == Val_unit);
}

value get_stub_info(value stub) /* ML */
{
  value res;

  Begin_root(stub);
  res = alloc(7,0);
  Field(res,0) =Val_int(Tag_val(stub));         /* tag */
  Field(res,1) = Field(stub,STUB_LOCATOR);  /* locator */
  Field(res,2) = Field(stub,STUB_POS);      /* space */
  Field(res,3) = Field(stub,STUB_POS);      /* location */
  Field(res,4) = is_local_stub(stub);           /* localp */
  Field(res,5) = Field(stub,STUB_STUB);     /* gc_stub */
  Field(res,6) = is_failed_stub(stub);
  End_roots();

  return res; /* tag*scion_id*position*local*locator*failed */
}

value get_stub_stub(value stub) /* ML */
{
  return Field(stub,STUB_STUB);
}

value set_stub_stub(value stub,value gc_stub) /* ML */
{
  Good_stub("set_stub_stub",stub);
  
  Assign(Field(stub,STUB_STUB),gc_stub);

  return Val_unit;
}

value set_stub_locator(value stub,value locator) /* ML */
{
  Assign(Field(stub,STUB_LOCATOR),locator);

  return Val_unit;
}

value set_stub_pos(value stub,value position) /* ML */
{
  Assign(Field(stub,STUB_POS),position);

  return Val_unit;
}

value create_stub(value tag,value pos) /* ML */
{
  value res;

  Begin_root(pos); /* (pos) */
  res = obj_block(tag,Val_int(STUB_SIZE));
  Assign(Field(res,STUB_POS),pos);
  Assign(Field(res,STUB_RESERV),Val_unit);
  
  End_roots(); /* (pos) */
  
  Good_stub("create_stub",res);

  return res;
}

/*********

   module Objects

*********/

value obj_copy(value obj,value debut,value len) /* ML */
{
  value res,i;

  Begin_root(obj);
  res = alloc(0,len);
  End_roots();
  for(i=0;i<len;i++)
    Field(res,i) = Field(obj,debut+i);
  
  return res;
}


value create_object_stub(value space,
			 value ext_buckets,value obj_size) /* ML */
{
  value res;
  
  Begin_roots2(space,ext_buckets); /* (pos,class) */
  res = obj_block (Val_int(Object_tag), obj_size);
  Assign(Field(res,STUB_POS),space);
  Assign(Field(res,5), ext_buckets); /* TMV exterieure */
  End_roots(); /* (pos,class) */

  return res;
}

value set_replicat(value stub,value contenu) /* ML */
{
  value i;

  for(i=0;i<Wosize_val(contenu);i++)
    Assign(Field(stub,i+6),Field(contenu,i));
  
  return Val_unit;
}

/*********

   module Thread

*********/

value get_thread_stub(thread_t th)    /* ML */
{
  value res;

  if(th->stub != Val_unit) {
    return th->stub;
  } else {
    value location = th->location;
    local_location_t lloc = (local_location_t)Field(location,STUB_LOCAL);
    if(lloc->next_free_stub>1){
      lloc->next_free_stub -= 2;
      res = Field(lloc->free_stubs,Int_val(lloc->next_free_stub));
      Assign(Field(lloc->free_stubs,Int_val(lloc->next_free_stub)),Val_unit);
    } else {
      DPRINT("get_thread_stub");
      
      Begin_root(th); /* (thread) */
      res = create_stub(Val_int(Thread_tag),location);
      End_roots(); /* (thread) */
    }
    Assign(Field(res,STUB_LOCAL),(value)th);
    Assign(th->stub,res);
    return res;
  }
}

value self_stub(value unit)
{
  return get_thread_stub(curr_thread);
}

value reuse_thread_loc(value location,value clos,value args,value thread_stub)
     /* ML */
{
  if(is_local_stub(thread_stub)!=Val_unit){
    return local_thread_reuse(location,clos,args,
			      (thread_t)Field(thread_stub,STUB_LOCAL));  
  } else{
    return gen_thread_new(location,clos,Val_true,args);
  } 
}


value reuse_thread(value clos,value args,value thread_stub) /* ML */
{
  return reuse_thread_loc((value)curr_thread->location,clos,args,thread_stub);
}

/*********

   module Location

*********/
/*
external create_new_object : space -> location -> int -> Reloc.module_data ->
  Obj.t = "create_new_object";;
*/

value set_location_module(value obj, value module)
{
  Assign(Field(Field(obj,STUB_POS),LOCATION_MODULE),module);
  return Val_unit;
}

value get_location_module(value obj)
{
  return Field(Field(obj,STUB_POS),LOCATION_MODULE);
}


value create_location(value space,value father,value module_data,
		      int size, int tag)
{
  value l,ll;
  value tmp;

  Begin_roots3(space,father,module_data);
  /* (space,father,module_data,ll) */

  ll =alloc(LOCATION_SIZE,Local_tag);
  
  Field(ll,LOCATION_FATHER) = father;
  Field(ll,LOCATION_DEFS) = Val_unit;
  Field(ll,LOCATION_LOCS) = Val_unit;
  Field(ll,LOCATION_MODULE) = module_data;
  Field(ll,LOCATION_ATTACH) = Val_unit;
  Field(ll,LOCATION_FREE_STUBS) = Val_unit;
  Field(ll,LOCATION_FREE_STUBS_SIZE) = Val_int(0);
  Field(ll,LOCATION_NEXT_FREE_STUB) = Val_int(0);
  Begin_root(ll);
  
  tmp = weakset_create(Val_int(DEF_WEAKSET_SIZE));
  Assign(Field(ll,LOCATION_DEFS),tmp);
  tmp = weakset_create(Val_int(LOC_WEAKSET_SIZE));
  Assign(Field(ll,LOCATION_LOCS),tmp);
  tmp = weakset_info(Field(ll,LOCATION_LOCS));
  
  l = obj_block(Val_int(tag),Val_int(size));
  
  Good_stub("create_location",l);
  Assign(Field(l,STUB_LOCAL),ll);

  if((value)father != Val_unit){
    Assign(Field(l,STUB_POS),Field(father,STUB_POS));
    Begin_root(l); /* (l) */
    tmp = weakset_add(Field(Field(father,STUB_LOCAL),LOCATION_LOCS),l);
    Assign(Field(Field(father,STUB_LOCAL),LOCATION_LOCS),tmp);
    tmp = weakset_info(Field(Field(father,STUB_LOCAL),LOCATION_LOCS));
    End_roots(); /* (l) */
  } else {
    Assign(Field(l,STUB_POS),space);
  }
  End_roots(); /* (space,father,module_data,ll) */
  End_roots();
  
  Good_stub("create_new_location",l);

  return l;
}

value create_new_location(value space,value father, 
			  value module_data) /* ML */
{
  return create_location(space,father,module_data,STUB_SIZE, Location_tag);
}

value create_new_object(value space,value location,value size,
			value module_data) /*  ML  */
{
  return create_location(space,location,module_data, Int_val(size), 
			 Object_tag);
}

/*
 
     STUB
   5 TMV externe
   6 TMV interne
   variables d'instances

*/
#define IN_BUCKET 6
value marshal_location(value dest_space,value location) /* ML */
{
  value res = Val_unit;
  value tmp;
  value size = Wosize_val(location);

  Good_stub("marshal_location",location);
  Assign(Field(location,STUB_POS),dest_space);

  if(Field(Field(location,STUB_POS),LOCATION_FATHER)== Val_unit)
    failwith("Attempt to marshal site location");

  if(
     weakset_remove(Field(Field(Field(Field(location,STUB_LOCAL),LOCATION_FATHER),STUB_LOCAL),LOCATION_LOCS),
		    location)
     ==Val_false){
    res = weakset_info(Field(Field(Field(Field(location,STUB_LOCAL),LOCATION_FATHER),STUB_LOCAL),LOCATION_LOCS));
    failwith("Go marshal: Unable to remove location");
  }

  Begin_roots2(res,location); /* (res,location) */
  res = alloc(6,0);
  Field(res,0) = Val_unit;
  Field(res,1) = Val_unit;
  Field(res,2) = Val_unit;
  Assign(Field(res,3),Field(Field(location,STUB_LOCAL),LOCATION_MODULE));
  Assign(Field(res,4),Field(Field(location,STUB_LOCAL),LOCATION_ATTACH));
  Field(res,5) = Atom(0);

  tmp = extract_location_threads((value)location);
  Assign(Field(res,0),tmp);  /* les threads */
  tmp = weakset_array(Field(Field(location,STUB_LOCAL),LOCATION_DEFS));
  Assign(Field(res,1),tmp);  /* les defs */
  tmp = weakset_array(Field(Field(location,STUB_LOCAL),LOCATION_LOCS));
  Assign(Field(res,2),tmp);  /* les sous-locations */
  End_roots(); /* (res,location) */

  /* si la location est un objet */
  if(size > STUB_SIZE){
    value vars = alloc(size-IN_BUCKET,0); 
    int i;

    for(i=IN_BUCKET;i<size;i++){
      Assign(Field(vars,i-IN_BUCKET),Field(location,i));
      Assign(Field(location,i),Val_unit);
    }
    Assign(Field(res,5),vars);
  }

  Assign(Field(location,STUB_LOCAL), Val_unit);

  return res;
}

value unmarshal_location (value father_v,value location_v,
			  value module_data,value attach, 
			  value vars)   /* ML */
{
  location_t father = (location_t) father_v;
  local_location_t ll;
  value tmp;
 
  Good_stub("unmarshal_location",location_v);
  
  Begin_roots5(father,location_v,module_data,attach,vars);

  if(Wosize_val(location_v) > STUB_SIZE){
    int size = Wosize_val(vars);
    int i;
    /* il s'agit d'un objet */
    for(i=0;i<size;i++)
      Assign(Field(location_v,i+IN_BUCKET),Field(vars,i));
  }

  tmp = weakset_add(father->lloc->loc_weakset, location_v);
  Assign(father->lloc->loc_weakset,tmp);
    
  ll = (local_location_t)alloc(LOCATION_SIZE,Local_tag);
  ll->father = father;
  ll->def_weakset = Val_unit;
  ll->loc_weakset = Val_unit;
  ll->module = module_data;
  ll->attach = attach;
  ll->free_stubs = Val_unit;
  ll->free_stubs_size = Val_int(0);
  ll->next_free_stub = Val_int(0);

  Begin_root(ll); /* (ll) */
  tmp = weakset_create(Val_int(DEF_WEAKSET_SIZE));
  Assign(ll->def_weakset,tmp);
  tmp = weakset_create(Val_int(LOC_WEAKSET_SIZE));
  Assign(ll->loc_weakset,tmp);
  End_roots(); /* (ll) */
  {
    location_t location = (location_t) location_v;

    Assign(location->space,father->space);
    Assign(location->lloc,ll);
  }
  End_roots(); /* (father,location_v,module_data) */

 return Val_unit;
}

/*
  On attache une unite de relocation a une location, sous la forme
d'une liste de (unite de relocation,interface). On migre la location avec 
ses attaches, puis, quand on la demarshalle, on place les interfaces sur
les unites de relocation, permettant ainsi les appels non resolus.
 */
value attach_to_location(value module_data) /* ML */
{
  value cons;

  Begin_root(module_data);
  cons = alloc(2,0);
  End_roots();
  {
    local_location_t location = ((location_t)curr_thread->location)->lloc;
    
    Field(cons,0) = module_data;
    Field(cons,1) = Field(location,LOCATION_ATTACH);
    Assign(Field(location,LOCATION_ATTACH),cons);
  }
  return Val_unit;
}

value go_location(value dest) /* ML */
{
  location_t l,self;
  value tmp;

  self = (location_t) curr_thread->location;
  l=(location_t)dest;

  if(self->lloc->father == l) return Val_unit; /* deja fait */
  if((value)(self->lloc->father) == Val_unit)
    failwith("Go: migration of space location");

  if((value) Field(dest,STUB_LOCAL) != Val_unit){ /* migration locale */
    while((value)l!=Val_unit){
      if(l == self) 
	failwith("Go: migration to children location");
      l=l->lloc->father;
    }
    l=(location_t)dest;
    if(weakset_remove(self->lloc->father->lloc->loc_weakset,(value)self)
       ==Val_false)
      failwith("Local Go: Unable to remove location");
    Assign(self->lloc->father,l);
    Begin_root(l);
    tmp = weakset_add(l->lloc->loc_weakset,(value)self);
    Assign(l->lloc->loc_weakset,tmp);
    End_roots();
  } else {
    value res;

    Begin_root(dest);
    res = alloc(2,0);
    End_roots();

    Field(res,0) = curr_thread->location;
    Field(res,1) = dest;
    return schedule_go_thread(res);
  }
  return Val_unit;
}

/*********

   module Interfaces

 ********/
value get_interface(value module_data) /* ML */
{
  if(Field(module_data,INTERFACE) == Val_int(-1))
    failwith("Location attached to unset Module");
  return Field(global_data,Int_val(Field(module_data,INTERFACE)));
}

value next_free_global = 0;
value end_free_global = 0;

extern void realloc_global(value);

value set_interface(value module_data,value interface) /* ML */
{
  if(Field(module_data,LONGUEUR) == Val_int(0))
    failwith("Can't attach interface to dummy module");
  if(Field(module_data,INTERFACE) != Val_int(-1)) 
    return Val_unit; /* already done */

  /* il faut ajouter un slot a la table globale pour l'interface.
     Ceci n'est possible que si on ne se trouve pas dans le toplevel (qui
     manipule aussi la table globale ! ).
   */

  if(next_free_global == end_free_global){
    /* no slots reserved for interfaces */
    next_free_global = Wosize_val(global_data);
    end_free_global = next_free_global + 20;
    realloc_global(Val_int(end_free_global));
  }
  Assign(Field(global_data,next_free_global),interface);
  Field(module_data,INTERFACE) = Val_int(next_free_global++);
  
  return Val_unit;
}

/*********

   module Channel

*********/

value code_t_root = Val_unit;
value env_t_root = Val_unit;

value register_send_sync(value code_t, value env_t)
{
  code_t_root = code_t;
  env_t_root = env_t;
  register_global_root(&code_t_root);
  register_global_root(&env_t_root);

  return Val_unit;
}

value create_empty_automata_in(value location,
			       value ndefs_v,value nchans_v,value indexes) /* ML */
{
  value tmp,i;
  value nchans = Int_val(nchans_v);
  value res=Val_unit;
  value def=Val_unit;
  value ldef=Val_unit;

  Begin_roots2(location,indexes);  /* (res,def,ldef) */
  
  res = alloc(Wosize_val(indexes)+1,0);
  
  for(i=0;i<=Wosize_val(indexes);++i)
    Field(res,i) = Val_unit;
  
  Begin_root(res);
  
  DPRINT("create_empty_automata_in\n");
  
  def = create_stub(Val_int(Automata_tag),location);

  Begin_root(def);
  
  Assign(Field(res,0),def);
  for(i=0;i<Wosize_val(indexes);++i){
    tmp = alloc(4,Closure_byte_tag); /* les canaux */
    Init(Field(tmp,0), code_t_root);
    Init(Field(tmp,1), env_t_root);
    Field(tmp,2) = def;
    Field(tmp,3) = Field(indexes,i);
    Assign(Field(res,i+1),tmp);
  }
  tmp = weakset_add(((location_t)location)->lloc->def_weakset,def);
  Assign(((location_t)location)->lloc->def_weakset,tmp);

  ldef = alloc(3+nchans,0);/* local def: status + structure + processes + queues */
  
  for(i=0;i<3+nchans;i++)
    Field(ldef,i) = Val_unit;

  Assign(Field(def,STUB_LOCAL),ldef);
  End_roots();            /* (res,def,ldef) */
  End_roots();
  End_roots();
  Good_stub("create_empty_automata_in",def);
    
  return res;
}

value fill_def(value def,value processes,value structure) /* ML */
{
  value ldef = Field(def,STUB_LOCAL);
  
  Good_stub("fill_def",def);
  Assign(Field(ldef,1),structure);
  Assign(Field(ldef,2),processes);
  
  return Val_unit;
}

value trigger_def (value params)
{
  value def = Field(params,0); /* BLOCK */
  value ord_v = Field(params,1);
  value arg = Field(params,2);  /* BLOCK */
  value ldef = Field(def,STUB_LOCAL); /* BLOCK */
  value ord = Int_val(ord_v);
  value index = ord & 127;
  value dispatch = ord >> 10;
  value structure = Field(ldef,1);   /* BLOCK */
  value nmult = Int_val(Field(structure,0));
  value state = Field(ldef,0);
  value bit = 1 << (index + 1);
  value cons;

  Good_stub("trigger_def",def);

  if(! (state & bit)){    /* pas ce message */
    value tests = Field(structure,2+index); /* BLOCK */
    value ntests =   Wosize_val(tests);
    value new_state = state | bit;
    value inverse = ~new_state;
    value nchans = Int_val(Field(structure,1)); 
    value i;
    value processes = Field(ldef,2);   /* BLOCK */

    for(i=0;i<ntests;i+=2){
      if(!(Field(tests,i) & inverse)){
      /* we must trigger this process */
	/* first we have to construct its args */
	value match = Field(tests,i+1);
	value nargs;
	value process;
	value thread_num;

	if(Is_long(match)){
	  match = Field(structure,2+nchans+Int_val(match));
	  Assign(Field(tests,i+1),match);
	}
	nargs = Wosize_val(match)-1;
	process = Field(match,0);
	
	if(nargs==1){ /* un seul argument, c'est le message ! */
	  if(index < nmult) /* dispatch */
	    process = Field(process,dispatch);

	  thread_num = Int_val(Field(process,1));
	  process = Int_val(Field(process,0));
	  if (!thread_num){
	    gen_thread_new(Field(def,STUB_POS),
			   Field(processes,process),
			   Val_true,arg);
	  } else 
            if(thread_num!=-1){
              reuse_thread_loc(Field(def,STUB_POS),
                               Field(processes,process),
                               arg,Field(arg,0));
            } else {
              local_thread_reuse(Field(def,STUB_POS),
                                 Field(processes,process),
                                 arg,curr_thread);
            }
          return Val_unit;
	} else {
	  value args;
	  value chan;
	  value list;

	  Begin_roots5(match,processes,process,arg,ldef);
	  args = alloc(nargs,0);
	  End_roots();
	  for(i=0;i<nargs;i++){
	    chan = Int_val(Field(match,i+1));
	    if(chan==index){ /* message actuel */
	      if(index<nmult) /* dispatch */
		process = Field(process,dispatch);
	      Assign(Field(args,i),arg);
	    } else { /* message queue */
	      list = Field(ldef,3+chan);
	      if(chan<nmult) /* dispatch */
		process = Field(process, Int_val(Field(list,2)));
	      Assign(Field(args,i),Field(list,0));
	      list = Field(list,1);
	      if(list == Val_unit)
		state &= ~(1 << (chan+1));
	      Assign(Field(ldef,3+chan),list);
	    }
	  }
	  Field(ldef,0) = state;
	  thread_num = Int_val(Field(process,1));
	  process = Int_val(Field(process,0));
	  if (!thread_num){
	    gen_thread_new(Field(def,STUB_POS),
			   Field(processes,process),
			   Val_true,args);
	  } else
            if(thread_num != -1) {
              value thread_stub = Field(Field(args,thread_num-1),0);
              reuse_thread_loc(Field(def,STUB_POS),
                               Field(processes,process),
                               args,
                               thread_stub);
            } else {
              local_thread_reuse(Field(def,STUB_POS),
                                 Field(processes,process),
                                 args,curr_thread);
            }
	  return Val_unit;
	}
      }
    }
    Field(ldef,0) = new_state;
  }
  Begin_roots2(ldef,arg); /* ldef, arg */
  if(index < nmult){
    cons = alloc(3,0);
    Field(cons,2) = Val_int(dispatch);
  } else
    cons = alloc(2,0);
  Field(cons,0) = arg;
  Field(cons,1) = Field(ldef,3+index);
  Assign(Field(ldef,3+index),cons);
  End_roots();

  return Val_unit;
}

value marshal_def (value def) /* ML */
{
  value tmp = Field(def,STUB_LOCAL);
  
  Assign(Field(def,STUB_LOCAL),Val_unit);

  return tmp; /* ldef */
}

value unmarshal_def (value paire) /* ML */
{ /* paire = (def,ldef) */
  local_location_t lloc;
  value tmp;
  value def;

  Assign(Field(Field(paire,0),STUB_LOCAL), Field(paire,1));
  
  def = Field(paire,0);
  lloc = (local_location_t) Field(Field(def,STUB_POS),STUB_LOCAL);
  Begin_root(lloc);
  tmp = weakset_add(((local_location_t)lloc)->def_weakset,def);
  Assign(((local_location_t)lloc)->def_weakset,tmp);
  End_roots();
  
  Good_stub("unmarshal_def",def);
  
  
  return Val_unit;
}

/*********

   module Space

*********/

/*
  On ne veut pas que la pile du thread contiennent des donnees ne pouvant pas
migrer. En utilisant un callback a chaque fois, on est sur que la pile est
nettoyee au retour de la fonction.
*/
value locator_send;

value set_locator_send(value send_callback) /* ML */
{
  locator_send = send_callback;
  register_global_root(&locator_send);
  
  return Val_unit;
}

value location_send(value location,value message) /* ML */
{
  if(is_local_stub(location)==Val_true)
    failwith ("Unable to send to local location");

  return callback2(locator_send,Field(location,STUB_POS),message);
}

value breakpoint(value num,value v)/* ML */
{
  return Val_unit;
}

value test_good_stub(value stub,value string)
{
  Good_stub(String_val(string),stub);
  return Val_unit;
}

#ifdef DEBUG
#define Good_channel(msg,channel) good_channel(channel,msg,__FILE__,__LINE__)
#else
#define Good_channel(msg,channel)
#endif

void good_channel(value channel,char* msg,char *file,int line)
{
  value tmp;
  
  if(Wosize_val(channel)!=2)
    failed_message(msg,"Channel is not a pair",file,line);
  if(!Is_long(Field(channel,1)))
    failed_message(msg,"Bad second element",file,line);
  tmp = Field(channel,0); /* automate */
  DPRINT4("%s: Test def (%s:%d) : %x\n",msg,file,line,tmp);
  
  if(Tag_val(tmp)!=Automata_tag)
    failed_message(msg,"Not an automata",file,line);
  good_stub(tmp,msg,file,line);
}


value test_good_channel(value channel,value string)
{
  Good_channel(String_val(string),channel);
  
  return Val_unit;
}

#include <stdio.h>
value prerr_string(value string)
{
  fprintf(stderr,"%s",String_val(string));
  return Val_unit;
}

value join_sys_exit(value exn_bucket)
{
  fprintf(stderr,"Exited in send_exception (UNRECOVERABLE): \n");
  fatal_uncaught_exception(exn_bucket);
  exit(2);
  return Val_unit;
}

 
value halt_stub(value stub)
{
  value locator = Field(stub,STUB_LOCATOR);
  
  if(locator != Val_unit){
    locator = Field(locator,0);
    Field(locator,LOCATOR_AGE) = Val_int(1 << 30);
  }

  Assign(Field(stub,STUB_POS),Val_unit);  
  Assign(Field(stub,STUB_LOCAL),Val_unit);
  Assign(Field(stub,STUB_STUB),Val_unit);
  
  return Val_unit;
}


value halt_location(value location)
{
  value tmp = Val_unit;
  int i;
  

  Begin_roots2(location,tmp);
  
  Good_stub("marshal_location",location);
  Assign(Field(location,STUB_POS),Val_unit);

  if(weakset_remove(Field(Field(Field(Field(location,STUB_LOCAL),
                                      LOCATION_FATHER)
                                ,STUB_LOCAL),LOCATION_LOCS),location)
     ==Val_false) failwith("Halt location: Unable to remove location");

  /* les threads */
  tmp = extract_location_threads(location);
  for(i=0;i<Wosize_val(tmp);i++) {
    remove_thread((thread_t)Field(Field(tmp,i),STUB_LOCAL));
    halt_stub(Field(tmp,i));
  }
    
  /* sous-locations */
  tmp = weakset_array(Field(Field(location,STUB_LOCAL),LOCATION_LOCS));
  for(i=0;i<Wosize_val(tmp);i++) halt_location(Field(tmp,i));
  
  /* les automates */
  tmp = weakset_array(Field(Field(location,STUB_LOCAL),LOCATION_DEFS));
  for(i=0;i<Wosize_val(tmp);i++) halt_stub(Field(tmp,i));

  halt_stub(location);
  End_roots();

  return curr_thread->retval;
}

