(***********************************************************************)
(*                                                                     *)
(*                     the join-calculus language                      *)
(*                                                                     *)
(*   Cedric Fournet and Luc Maranget, projet PARA, INRIA Rocquencourt  *)
(*                                                                     *)
(*  Copyright 1997 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

type 'a t = { mutable pos : int; mutable data : 'a Weak.t };;

exception Found of int;;

let create () = {pos = 0; data = Weak.create 0};;

let apply f t =
  for i = 0 to Weak.length t.data - 1 do
    match Weak.get t.data i with
      Some x -> f x i
    | None -> ()
  done
;;

let fold f v0 t =
  let r = ref v0 in
  for i = 0 to Weak.length t.data - 1 do
    match Weak.get t.data i with
      Some x -> r := f !r x
    | None -> ()
  done;
  !r
;;
  
let remove test t =
  apply (fun x i -> if test x then Weak.set t.data i None) t
;;

let iter f t = apply (fun x i -> f x) t;;

let map f t =
  let t' = {pos = t.pos; data = Weak.create (Weak.length t.data)} in
  apply (fun x i -> Weak.set t'.data i (Some (f x))) t; t'
;;
  
let clear t = t.pos <- 0; t.data <- Weak.create 0;;

let length t = fold (fun x _ -> x + 1) 0 t;;

let size t = Weak.length t.data;;
  
let add t x =
  let size = Weak.length t.data in
  let empty_pos =
    try
      for i = t.pos to size - 1 do
        if Weak.get t.data i = None then raise (Found i)
      done;
      for i = 0 to t.pos - 1 do
        if Weak.get t.data i = None then raise (Found i)
      done;
      let previous = t.data in
      t.data <- Weak.create (max 2 (2 * size));
      Weak.blit previous 0 t.data 0 size;
      size
    with
      Found i -> i
  in
  Weak.set t.data empty_pos (Some x); t.pos <- empty_pos + 1
;;

let map_to_list f t = fold (fun xs x -> f x :: xs) [] t;;

let map_of_list f l =
  let t = create () in List.iter (fun x -> add t (f x)) l; t
;;










