type ('a,'b) t =
    { 
    size: int;
    table: ('a,int) Hashtbl.t;
    mutable array: 'b Weak.array ref
    } 

let create n =
  { 
  size = n;
  table = Hashtbl.create n;
  array = ref ( Weak.array_create n)
  } 

let add t a b =
  let wk = Weak.ref_add t.array b in
  Hashtbl.add t.table a wk
    
let find t a = Weak.array_get (!(t.array)) (Hashtbl.find t.table a)
    
let remove t a = 
  try
    let wk = Hashtbl.find t.table a in
    Weak.array_free !(t.array) wk;
    Hashtbl.remove t.table a
  with Not_found -> ()

let h_to_list t =
  let l = ref [] in
  Hashtbl.iter (fun a b -> l:=(a,b):: !l) t;
  !l

let h_from_list n l =
  let t = Hashtbl.create n in
  List.iter (fun (a,b) -> Hashtbl.add t a b) l;
  t

let iter f t =
  let l = h_to_list t.table in
  List.iter (fun (a,b) -> f a (Weak.array_get !(t.array) b)) l

let to_list t =
  List.map (fun (a,b) -> (a,Weak.array_get !(t.array) b)) (h_to_list t.table)

let from_list n l =
  let t = create n in
  List.iter (fun (a,b) -> add t a b) l;
  t

let clear t =
  Hashtbl.clear t.table;
  t.array <- ref ( Weak.array_create t.size)
