open Mop

let new_method = Mop.new_method

let set_initializer = Mop.set_initializer
let inheritance = Mop.inheritance
let get_method = Mop.get_method
let set_method = Mop.set_method
let get_method_label = Mop.get_method_label
let get_variable = Mop.get_variable
let hide_variable = Mop.hide_variable
let get_private_variable = Mop.get_private_variable
let create_class = Mop.create_class
let get_class_info = Mop.get_class_info

(* Parameters *)
let params = Mop.params

(* Objects *)
let create_object table =
  let Class_id(module_name,class_no) = 
    (Obj.magic table.buckets.(0) : class_info).class_type 
  in
  let module_data = Reloc.get_module_data module_name
  in
  let obj = 
    let self_loc = Jprims.self_location () in
    Jprims.create_new_object Space.local_space 
      self_loc table.size module_data
  in
  Obj.set_field obj 5 (Obj.repr table.ext_buckets);     (* TMV *)
  Obj.set_field obj 6 (Obj.repr table.buckets);     (* TMV *)
  let initialization = Obj.magic List.hd (List.hd table.init) in
  (* sur un autre thread *)
  Jprims.set_critical_section  true;
  let thread_stub = Jprims.self_stub () in
  Thread.create_in (Obj.magic obj:Jtypes.location) 
    (function () ->
      Channel.reply_to_and_exit thread_stub
        (try
          Jtypes.Reply (initialization (Obj.obj obj))
        with
          e -> Jtypes.Exception e)
        ) ();
  Jprims.set_critical_section false;
  match  Jprims.wait_on_stub () with
    Jtypes.Reply v -> Obj.magic v
  | Jtypes.Exception e -> raise e


let create_class class_info public_methods creator class_init module_data =
  Debug.print0 'o' "Mop.create_class...";
  let class_info =
    Mop.create_class class_info public_methods creator class_init module_data
  in
  (* il faut reconstruire class_info.table.ext_buckets a partir de
       class_info.table.buckets et de class_info.table.messages
  *)
  Debug.print0 'o' "...Mop.create_class\n";
  class_info.table.ext_buckets 
    <- Channel.create_ext_buckets class_info.table.messages ;
  Debug.print0 'o' "...Oo.create_class\n"

let copy (o: < .. >) = 
  let o' = Obj.repr o in
  let (locator, position, local, _, gc_stub) = Obj.magic o' in
  if (Obj.magic local) == None then
    failwith "Can not copy non local object";
  let sz = Obj.size o' in
  let r = 
    let self_loc = Jprims.self_location () in
    Jprims.create_new_object Space.local_space 
      self_loc sz (Jprims.get_location_module o)
  in
  for i = 5 to sz - 1 do Obj.set_field r i (Obj.field o' i) done;
  Obj.obj r 
