
#include "callback.h"
#include "config.h"
#include "misc.h"
#include "mlvalues.h"
#include "stacks.h"
#include "fail.h"
#include "io.h"
#include "roots.h"
#include "alloc.h"
#include "memory.h"
#include "signals.h"
#include "gc_ctrl.h"
#include "intext.h"
#include "threads.h"
#include <stdio.h>

extern void byte_do_local_roots(scanning_action,value*,value*,value);


#if ! (defined(HAS_SELECT) && \
       defined(HAS_SETITIMER) && \
       defined(HAS_GETTIMEOFDAY) && \
       (defined(HAS_WAITPID) || defined(HAS_WAIT4)))
#include "Cannot compile libthreads, system calls missing"
#endif

#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <fcntl.h>
#ifdef HAS_UNISTD
#include <unistd.h>
#endif
#ifdef HAS_SYS_SELECT_H
#include <sys/select.h>
#endif

#ifndef FD_ISSET
/* Assume old-style BSD 4.2 fd sets */
typedef int fd_set;
#define FD_SETSIZE (sizeof(int) * 8)
#define FD_SET(fd,fds) (*(fds) |= 1 << (fd))
#define FD_CLR(fd,fds) (*(fds) &= ~(1 << (fd)))
#define FD_ISSET(fd,fds) (*(fds) & (1 << (fd)))
#define FD_ZERO(fds) (*(fds) = 0)
#endif

#ifndef HAS_WAITPID
#define waitpid(pid,status,opts) wait4(pid,status,opts,NULL)
#endif

#define NO_THREADS (thread_t)Val_unit

/* Identifier for next thread creation */
value next_ident = Val_int(0);

/* le thread courant */
/* not registered for GC since used to access stacks during compaction */
thread_t curr_thread = NO_THREADS;

/* les threads qui doivent tourner dans ce cycle, et
les threads qui devront tourner au prochain cycle */
thread_t runqueue = NO_THREADS;
thread_t nextqueue = NO_THREADS;
 /* les threads qui sont bloques (avec et sans select) */
thread_t blockedqueue = NO_THREADS;
thread_t selectqueue = NO_THREADS;

/* drapeau de section critique */
value critical_section = Val_false;

/* recyclage de threads */
thread_t *old_threads = Val_unit;
int next_old_thread = 0;
int first_old_thread = 0;
int old_threads_size = 0;

/* terminaison */
int terminaison = 0; /* nbre de threads restant a la terminaison (0=aucun) */
int nbre_threads = 0;

/* statistiques */
int stat_nthreads_started = 0;
int stat_nthreads_alloc = 0;

/* pour le GC */
static void (*prev_scan_roots_hook) (scanning_action);
static void thread_scan_roots(scanning_action action)
{
  thread_t th,start;

  check_runqueue();
  
  /* Scan all active descriptors */
  start = curr_thread;
  (*action)((value) curr_thread, (value *) &curr_thread);
  /* Don't scan curr_thread->sp, this has already been done.
     Don't scan local roots either, for the same reason. */
  for (th = start->next; th != start; th = th->next) {
    byte_do_local_roots(action, th->sp, th->stack_high, (value)NULL);
  }
  /* Hook */
  if (prev_scan_roots_hook != NULL) (*prev_scan_roots_hook)(action);
}


/* Initialisation */
value thread_initialize(value unit)       /* ML */
{
  struct itimerval timer;

  /* Create a descriptor for the current thread */
  curr_thread =
    (thread_t) alloc_shr(sizeof(struct thread_struct) / sizeof(value),
			 Local_tag);
  curr_thread->ident = next_ident;
  next_ident = Val_int(Int_val(next_ident) + 1);
  curr_thread->stack_low = stack_low;
  curr_thread->stack_high = stack_high;
  curr_thread->stack_threshold = stack_threshold;
  curr_thread->sp = extern_sp;
  curr_thread->trapsp = trapsp;
  curr_thread->status = RUNNABLE;
  curr_thread->readfds = NO_FDS;
  curr_thread->writefds = NO_FDS;
  curr_thread->exceptfds = NO_FDS;
  curr_thread->delay = NO_DELAY;
  curr_thread->joining = NO_JOINING;
  curr_thread->waitpid = NO_WAITPID;
  curr_thread->retval = Val_unit;

  /* join stuff */
  curr_thread->stub = Val_unit;
  curr_thread->location= Val_unit;
  curr_thread->reused = Val_true;
  curr_thread->sync = Val_unit;

  /* Initialize GC */
  prev_scan_roots_hook = scan_roots_hook;
  scan_roots_hook = thread_scan_roots;

  /* Initialize interval timer */
  timer.it_interval.tv_sec = 0;
  timer.it_interval.tv_usec = Thread_timeout;
  timer.it_value = timer.it_interval;
  setitimer(ITIMER_VIRTUAL, &timer, NULL);

  nbre_threads++;
  curr_thread->next = curr_thread;
  curr_thread->prev = curr_thread;

  check_runqueue();
  
  return Val_unit;
}


/* cree un thread dans une location a partir d'une closure */
value gen_thread_new(value location,value clos,
		     value param,value clos_arg) /* ML */
{
  thread_t th;

  /* Allocate the thread and its stack */

  if(first_old_thread < next_old_thread){ /* on utilise un vieux thread */
    
  if(location == Val_unit || 
     Field(Field(location,STUB_LOCAL), LOCATION_THREAD_NEXT) == Val_int(0)){
    /* pas de threads a reutiliser dans la location */
    Begin_roots3(location,clos,clos_arg);
    th = (thread_t) alloc_shr(sizeof(struct thread_struct) / sizeof(value), 
			      Local_tag);
    End_roots();
    thread_alloues++;
    th->stub = Val_unit; /* rien par defaut */
    Init(th->location,location);
    th->reused = Val_unit;
    th->sync = Val_unit;
  } else {
    value lloc = Field(location,STUB_LOCAL);
    value reused = Field(lloc,LOCATION_THREADS_REUSED);
    int next = Int_val(Field(lloc,LOCATION_THREAD_NEXT))-1;
    
    Field(lloc,LOCATION_THREAD_NEXT) = Val_int(next);
    th = (thread_t) Field(reused,next);
    Assign(Field(reused,next),Val_unit);
  }
  
  th->ident = next_ident;
  next_ident = next_ident + 2;
  DPRINT2("NEW THREAD %d with %x\n",Int_val(th->ident),location);

  /* is there a free stack in the buffer ? */
  if(first_stack_to_use<next_stack_free){
    /*    DPRINT("New thread not allocated\n");*/
    th->stack_low = stacks_buffer[first_stack_to_use++];
  } else { /* no, we must alloc this stack */
    /*   DPRINT("New thread allocated\n");*/
    stack_allouees++;
    th->stack_low = (value *) stat_alloc(Thread_stack_size);
  }

  DPRINT1("new thread : stacklow %lx\n", th-> stack_low);

  th->stack_high = th->stack_low + Thread_stack_size / sizeof(value);
  th->stack_threshold = th->stack_low + Stack_threshold / sizeof(value);
  th->sp = th->stack_high;
  /*  DPRINT1("NEW THREAD stack high : %d\n",th->sp);*/
  th->trapsp = th->stack_high;
  /* Set up a return frame that pretends we're applying clos to ().
     This way, when this thread is activated, the RETURN will take us
     to the entry point of the closure. */
  if(param == Val_false)
    th->sp -= 4;
  else {
    th->sp -= 5;
    Assign(th->sp[4],clos_arg);
  }
  th->sp[0] = Val_unit;
  th->sp[1] = (value) Code_val(clos);
  th->sp[2] = clos;
  th->sp[3] = Val_long(0);      /* no extra args */
  /* Fake a C call frame */
  th->sp--;
  th->sp[0] = Val_unit;         /* a dummy environment */
  /* The thread is initially runnable */
  th->status = RUNNABLE;
  th->readfds = NO_FDS;
  th->writefds = NO_FDS;
  th->exceptfds = NO_FDS;
  th->delay = NO_DELAY;
  th->joining = NO_JOINING;
  th->waitpid = NO_WAITPID;
  th->retval=Val_unit;

  /* Insert thread in doubly linked list of threads */
  Init(th->prev,(value) curr_thread->prev);
  Init(th->next ,(value) curr_thread);
  Assign(curr_thread->prev->next, th);
  Assign(curr_thread->prev, th);
  nbre_threads++;

  check_runqueue();
  
  return (value) th;
}

value thread_new(value clos)          /* ML */
{
  return gen_thread_new((value)curr_thread->location,clos,Val_false,Val_unit);
}

value gen_thread_in(value location,value clos)          /* ML */
{
  return gen_thread_new(location,clos,Val_false,Val_unit);
}























/***********************************************************************/
/*                                                                     */
/*                           Objective Caml                            */
/*                                                                     */
/*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         */
/*                                                                     */
/*  Copyright 1996 Institut National de Recherche en Informatique et   */
/*  Automatique.  Distributed only by permission.                      */
/*                                                                     */
/***********************************************************************/

/* $Id: new_scheduler.c,v 1.1.1.1 1998/04/20 11:32:13 lefessan Exp $ */

/* The thread scheduler */

/*

1 - Inutile de tester tous les threads a chaque appel du scheduler. On
  peut le faire une seule fois par cycle:
   - on teste tous les threads en placant ce qui peuvent tourner dans
      une runqueue.
   - on fait tourner tour a tour tous les threads de la runqueue
   - on place immediatement les threads preemptes dans le prochaine 
      runqueue
   - on place les threads en attente de terminaison (JOIN) et ceux
      suspendus (SUSPENDED) dans deux queues speciales dont ils
      sont extraits sur demande (kill ou wake_up) .

 A chaque cycle, on teste donc uniquement les threads bloques (IO, DELAY 
et GC)

2. Securiser les acces aux descripteurs de threads pour empecher la
reutilisation des threads extruder par les fonctions du module Thread
(modification de th->reused) et le transmettre lors du marshaller.

3. Ameliorer le recyclage :
  - Recyclage des threads au niveau de l'espace independament des locations
      et des piles les contenant
       (acceleration de la creation/destruction de threads)
  - Recyclage des stubs avec locators des threads morts au niveau des
      locations (diminution de l'activite du GC distribue)

*/ 

/* il faut verifier ce fichier pas a pas pour le rendre compatible avec
le compacteur qui deplace les objets */

#include "callback.h"
#include "config.h"
#include "misc.h"
#include "mlvalues.h"
#include "stacks.h"
#include "fail.h"
#include "io.h"
#include "roots.h"
#include "alloc.h"
#include "memory.h"
#include "signals.h"
#include "gc_ctrl.h"
#include "intext.h"
#include "threads.h"
#include <stdio.h>

#if ! (defined(HAS_SELECT) && \
       defined(HAS_SETITIMER) && \
       defined(HAS_GETTIMEOFDAY) && \
       (defined(HAS_WAITPID) || defined(HAS_WAIT4)))
#include "Cannot compile libthreads, system calls missing"
#endif

#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <fcntl.h>
#ifdef HAS_UNISTD
#include <unistd.h>
#endif
#ifdef HAS_SYS_SELECT_H
#include <sys/select.h>
#endif

#ifndef FD_ISSET
/* Assume old-style BSD 4.2 fd sets */
typedef int fd_set;
#define FD_SETSIZE (sizeof(int) * 8)
#define FD_SET(fd,fds) (*(fds) |= 1 << (fd))
#define FD_CLR(fd,fds) (*(fds) &= ~(1 << (fd)))
#define FD_ISSET(fd,fds) (*(fds) & (1 << (fd)))
#define FD_ZERO(fds) (*(fds) = 0)
#endif

#ifndef HAS_WAITPID
#define waitpid(pid,status,opts) wait4(pid,status,opts,NULL)
#endif

/* The thread currently active */
thread_t curr_thread = NULL;
/* Identifier for next thread creation */
value next_ident = Val_int(0);
value critical_section = Val_false;
/*
  Threads with normal stacks are kept in a free list to prevent spending too
much time in allocating them. We should extend this mecanism for all data
which are desallocated explicitely (stacks+???).
*/

value *stacks_buffer[STACKS_BUFFER_SIZE];
int next_stack_free=0;
int first_stack_to_use=0;
value stack_allouees = 0;
int nbre_threads = 0;
int terminaison = 1; /* we detect terminaison */
static int nthreads = 0;
static int  thread_alloues = 0;

/* Scan the stacks of the other threads */

static void (*prev_scan_roots_hook) (scanning_action);
extern void byte_do_local_roots(scanning_action,value*,value*,value);

static void thread_scan_roots(scanning_action action)
{
  thread_t th,start;

  check_runqueue();
  
  /* Scan all active descriptors */
  start = curr_thread;
  (*action)((value) curr_thread, (value *) &curr_thread);
  /* Don't scan curr_thread->sp, this has already been done.
     Don't scan local roots either, for the same reason. */
  for (th = start->next; th != start; th = th->next) {
    byte_do_local_roots(action, th->sp, th->stack_high, (value)NULL);
  }
  /* Hook */
  if (prev_scan_roots_hook != NULL) (*prev_scan_roots_hook)(action);
}

/* Initialize the thread machinery */
static long last_gc;

value thread_initialize(value unit)       /* ML */
{
  struct itimerval timer;
  /* Create a descriptor for the current thread */
  last_gc = stat_major_collections;

  curr_thread =
    (thread_t) alloc_shr(sizeof(struct thread_struct) / sizeof(value),
			 Local_tag);
  thread_alloues++;
  
  curr_thread->ident = next_ident;
  next_ident = Val_int(Int_val(next_ident) + 1);
  curr_thread->next = curr_thread;
  curr_thread->prev = curr_thread;
  curr_thread->stack_low = stack_low;
  curr_thread->stack_high = stack_high;
  curr_thread->stack_threshold = stack_threshold;
  curr_thread->sp = extern_sp;
  curr_thread->trapsp = trapsp;
  curr_thread->status = RUNNABLE;
  curr_thread->readfds = NO_FDS;
  curr_thread->writefds = NO_FDS;
  curr_thread->exceptfds = NO_FDS;
  curr_thread->delay = NO_DELAY;
  curr_thread->joining = NO_JOINING;
  curr_thread->waitpid = NO_WAITPID;
  curr_thread->retval = Val_unit;

  curr_thread->stub = Val_unit;
  curr_thread->location= Val_unit;
  curr_thread->reused = Val_unit;
  curr_thread->sync = Val_unit;
  /* Initialize GC */
  prev_scan_roots_hook = scan_roots_hook;
  scan_roots_hook = thread_scan_roots;
  /* Initialize interval timer */
  timer.it_interval.tv_sec = 0;
  timer.it_interval.tv_usec = Thread_timeout;
  timer.it_value = timer.it_interval;
  setitimer(ITIMER_VIRTUAL, &timer, NULL);

  check_runqueue();
  
  return Val_unit;
}

/* Create a thread */

/*
La generation de nouveaux threads est tres couteuse. En particulier, on cree
de nouveaux threads pour effectuer des appels synchrones distants, qui ne sont
plus utilises par la suite, et necessitent cependant l'utilisation couteuse du
GC distribue. Pourtant, apres le KILL, et a condition que le thread ne soit pas
manipuler autrement (Thread.join ...), on peut se permettre de reutiliser un
thread entierement ...

Chaque location contient donc un tableau de threads a reutiliser apres un
KILL. Ces threads sont toujours les meme, sauf quand l'un d'eux est utilise
autrement que dans le Join, et apparait donc en surface ...

local_location_t {

   thread_t[MAX_REUSED_THREADS] reuse_threads;
   int next_free;
}

 */

value gen_thread_new(value location,value clos,
		     value param,value clos_arg) /* ML */
{
  thread_t th;
  /* Allocate the thread and its stack */

  nbre_threads++;
  if(location == Val_unit || 
     Field(Field(location,STUB_LOCAL), LOCATION_THREAD_NEXT) == Val_int(0)){
    /* pas de threads a reutiliser dans la location */
    Begin_roots3(location,clos,clos_arg);
    th = (thread_t) alloc_shr(sizeof(struct thread_struct) / sizeof(value), 
			      Local_tag);
    End_roots();
    thread_alloues++;
    th->stub = Val_unit; /* rien par defaut */
    Init(th->location,location);
    th->reused = Val_unit;
    th->sync = Val_unit;
  } else {
    value lloc = Field(location,STUB_LOCAL);
    value reused = Field(lloc,LOCATION_THREADS_REUSED);
    int next = Int_val(Field(lloc,LOCATION_THREAD_NEXT))-1;
    
    Field(lloc,LOCATION_THREAD_NEXT) = Val_int(next);
    th = (thread_t) Field(reused,next);
    Assign(Field(reused,next),Val_unit);
  }
  
  th->ident = next_ident;
  next_ident = next_ident + 2;
  DPRINT2("NEW THREAD %d with %x\n",Int_val(th->ident),location);

  /* is there a free stack in the buffer ? */
  if(first_stack_to_use<next_stack_free){
    /*    DPRINT("New thread not allocated\n");*/
    th->stack_low = stacks_buffer[first_stack_to_use++];
  } else { /* no, we must alloc this stack */
    /*   DPRINT("New thread allocated\n");*/
    stack_allouees++;
    th->stack_low = (value *) stat_alloc(Thread_stack_size);
  }

  DPRINT1("new thread : stacklow %lx\n", th-> stack_low);

  th->stack_high = th->stack_low + Thread_stack_size / sizeof(value);
  th->stack_threshold = th->stack_low + Stack_threshold / sizeof(value);
  th->sp = th->stack_high;
  /*  DPRINT1("NEW THREAD stack high : %d\n",th->sp);*/
  th->trapsp = th->stack_high;
  /* Set up a return frame that pretends we're applying clos to ().
     This way, when this thread is activated, the RETURN will take us
     to the entry point of the closure. */
  if(param == Val_false)
    th->sp -= 4;
  else {
    th->sp -= 5;
    Assign(th->sp[4],clos_arg);
  }
  th->sp[0] = Val_unit;
  th->sp[1] = (value) Code_val(clos);
  th->sp[2] = clos;
  th->sp[3] = Val_long(0);      /* no extra args */
  /* Fake a C call frame */
  th->sp--;
  th->sp[0] = Val_unit;         /* a dummy environment */
  /* The thread is initially runnable */
  th->status = RUNNABLE;
  th->readfds = NO_FDS;
  th->writefds = NO_FDS;
  th->exceptfds = NO_FDS;
  th->delay = NO_DELAY;
  th->joining = NO_JOINING;
  th->waitpid = NO_WAITPID;
  th->retval=Val_unit;

  /* Insert thread in doubly linked list of threads */
  Init(th->prev,(value) curr_thread->prev);
  Init(th->next ,(value) curr_thread);
  Assign(curr_thread->prev->next, th);
  Assign(curr_thread->prev, th);

  /*  DPRINT("NEW THREAD inserted\n"); */

  check_runqueue();
  
  /* Return thread */
  return (value) th;
}

value thread_new(value clos)          /* ML */
{
  return gen_thread_new((value)curr_thread->location,clos,Val_false,Val_unit);
}

value gen_thread_in(value location,value clos)          /* ML */
{
  return gen_thread_new(location,clos,Val_false,Val_unit);
}

/* Return the thread identifier */

value thread_id(value th)             /* ML */
{
  return ((struct thread_struct *)th)->ident;
}

/* Return the current time as a floating-point number */

static double timeofday(void)
{
  struct timeval tv;
  gettimeofday(&tv, NULL);
  return (double) tv.tv_sec + (double) tv.tv_usec * 1e-6;
}


/* Find a runnable thread and activate it */

#define FOREACH_THREAD(x) x = curr_thread; do { x = x->next;
#define END_FOREACH(x) } while (x != curr_thread)

static value alloc_process_status(int pid, int status);
static void add_fdlist_to_set(value fdl, fd_set *set);
static value inter_fdlist_set(value fdl, fd_set *set);


value schedule_thread(void)
{
  thread_t run_thread, th;
  fd_set readfds, writefds, exceptfds;
  double delay, now;
  int need_select, need_wait;

  /* Don't allow preemption during a callback */
  check_runqueue();
  
  if (callback_depth > 0){  
    DPRINT1( "thread_schedule exited: callback depth : %d \n",callback_depth);
    return curr_thread->retval;
  }
  if(critical_section==Val_true) {
    DPRINT("Scheduler exited in critical section\n"); 
    /*
    if(curr_thread->status & BLOCKED_IO){
      DPRINT("Select in critical section\n");
    }
    */
    fatal_error("Enter scheduler in critical section");
    
    return curr_thread->retval;
  }

  /* Save the status of the current thread */
  curr_thread->stack_low = stack_low;
  curr_thread->stack_high = stack_high;
  curr_thread->stack_threshold = stack_threshold;
  curr_thread->sp = extern_sp;
  curr_thread->trapsp = trapsp;

  DPRINT1("sched stacklow %lx\n",stack_low);

try_again:
  /* Build fdsets and delay for select.
     See if some join or wait operations succeeded. */
  FD_ZERO(&readfds);
  FD_ZERO(&writefds);
  FD_ZERO(&exceptfds);
  delay = DELAY_INFTY;
  now = -1.0;
  need_select = 0;
  need_wait = 0;

  FOREACH_THREAD(th)
  if (th->status & (BLOCKED_IO - 1)) {
      add_fdlist_to_set(th->readfds, &readfds);
      add_fdlist_to_set(th->writefds, &writefds);
      add_fdlist_to_set(th->exceptfds, &exceptfds);
      need_select = 1;
    }
    if (th->status & (BLOCKED_DELAY - 1)) {
      double th_delay;
      if (now < 0.0) now = timeofday();
      th_delay = Double_val(th->delay) - now;
      if (th_delay <= 0) {
        th->status = RUNNABLE;
        th->retval = RESUMED_DELAY;
      } else {
        if (th_delay < delay) delay = th_delay;
      }
    }
    if (th->status & (BLOCKED_JOIN - 1)) {
      if (((thread_t)(th->joining))->status == KILLED) {
        th->status = RUNNABLE;
        th->retval = RESUMED_JOIN;
      }
    }
    if (th->status & (BLOCKED_WAIT - 1)) {
      int status, pid;
      pid = waitpid(Int_val(th->waitpid), &status, WNOHANG);
      if (pid > 0) {
        th->status = RUNNABLE;
        Begin_root(th); /*++*/
        Assign(th->retval, alloc_process_status(pid, status));
        End_roots();/*++*/
      } else {
        need_wait = 1;
      }
    } else
      if (th->status & (BLOCKED_GC - 1)){
	if(stat_major_collections>last_gc){
	  th->status = RUNNABLE;
        } 
      }
  END_FOREACH(th);

  last_gc = stat_major_collections;

  /* Find if a thread is runnable. */
  run_thread = NULL;
  FOREACH_THREAD(th)
    if (th->status == RUNNABLE) { run_thread = th; break; }
  END_FOREACH(th);

  /* Do the select if needed */
  if (need_select || run_thread == NULL) {
    struct timeval delay_tv, * delay_ptr;
    int retcode;
    /* If a thread is blocked on wait, don't block forever */
    if (need_wait && delay > Thread_timeout * 1e-6) {
      delay = Thread_timeout * 1e-6;
    }
    /* Convert delay to a timeval */
    /* If a thread is runnable, just poll */
    if (run_thread != NULL) {
      delay_tv.tv_sec = 0;
      delay_tv.tv_usec = 0;
      delay_ptr = &delay_tv;
    }
    else if (delay != DELAY_INFTY) {
      delay_tv.tv_sec = (unsigned int) delay;
      delay_tv.tv_usec = (delay - (double) delay_tv.tv_sec) * 1E6;
      delay_ptr = &delay_tv;
    }
    else {
      delay_ptr = NULL;
    }
    enter_blocking_section(); /* alloc */
    retcode = select(FD_SETSIZE, &readfds, &writefds, &exceptfds, delay_ptr);
    leave_blocking_section(); /* alloc */
    if (retcode > 0) {
      /* Some descriptors are ready. 
         Mark the corresponding threads runnable. */
      FOREACH_THREAD(th)
        if (th->status & (BLOCKED_IO - 1)) {
	  value r = Val_unit, w = Val_unit, e = Val_unit;
          Begin_roots4(th,r,w,e);
            r = inter_fdlist_set(th->readfds, &readfds);
	    w = inter_fdlist_set(th->writefds, &writefds);
	    e = inter_fdlist_set(th->exceptfds, &exceptfds);
	    if (r != NO_FDS || w != NO_FDS || e != NO_FDS) {
	      value retval = alloc(3, TAG_RESUMED_IO);
	      Field(retval, 0) = r;
	      Field(retval, 1) = w;
	      Field(retval, 2) = e;
	      Assign(th->retval, retval);
	      th->status = RUNNABLE;
	      if (run_thread == NULL) run_thread = th; /* Found one. */
	    }
          End_roots();
        }
      END_FOREACH(th);
    }
    /* If we get here with run_thread still NULL, one of the following
       may have happened:
       - a delay has expired
       - a wait() needs to be polled again
       - the select() failed (e.g. was interrupted)
       In these cases, we go through the loop once more to make the
       corresponding threads runnable. */
    if (run_thread == NULL &&
        (delay != DELAY_INFTY || need_wait || retcode == -1))
      goto try_again;
  }

  /* If we haven't something to run at that point, we're in big trouble. */
  if (run_thread == NULL) invalid_argument("Thread: deadlock");

  /* Free everything the thread was waiting on */
  Assign(run_thread->readfds, NO_FDS);
  Assign(run_thread->writefds, NO_FDS);
  Assign(run_thread->exceptfds, NO_FDS);
  Assign(run_thread->delay, NO_DELAY);
  Assign(run_thread->joining, NO_JOINING);
  run_thread->waitpid = NO_WAITPID;

  /* Activate the thread */
  curr_thread = run_thread;


  DPRINT1("Scheduling to thread : %d\n",Int_val(curr_thread->ident)); 
  if(curr_thread->stub == Val_unit) 
    DPRINT(" no stub \n");
  stack_low = curr_thread->stack_low;
  stack_high = curr_thread->stack_high;
  stack_threshold = curr_thread->stack_threshold;
  extern_sp = curr_thread->sp;
  trapsp = curr_thread->trapsp;
  if(extern_sp<stack_threshold) {
    DPRINT("Reallocating stack for new scheduled thread\n");
    realloc_stack();
  }
  check_runqueue();
  
  return curr_thread->retval;
}

/* Reschedule without suspending the current thread */

value set_critical_section(value str,value section) /* ML */
{
  value old = critical_section;
  
  critical_section = section;
  if(critical_section==Val_true)
    DPRINT1("--------------  ENTERING CRITICAL SECTION(%s)\n",String_val(str));
  else
    DPRINT1("--------------  LEAVING CRITICAL SECTION(%s)\n",String_val(str));

  check_runqueue();
  
  return old;
}

value thread_yield(value unit)        /* ML */
{
  Assert(curr_thread != NULL);
  if(critical_section == Val_true)
    return Val_unit;
  curr_thread->retval = Val_unit;
  return schedule_thread();
}

/* Suspend the current thread */

value thread_sleep(value unit)        /* ML */
{
  Assert(curr_thread != NULL);
  check_callback("thread_sleep");
  curr_thread->status = SUSPENDED;
  return schedule_thread();
}

/* Suspend the current thread on a select() request */

value thread_select(value arg)        /* ML */
{
  double date;
  /* Don't do an error if we're not initialized yet
     (we can be called from thread-safe Pervasives before initialization),
     just return immediately. */
  if (curr_thread == NULL) return RESUMED_WAKEUP;
  /* si on est dans un callback, on doit bloquer jusqu'a la fin du select */
  Assign(curr_thread->readfds, Field(arg, 0));
  Assign(curr_thread->writefds, Field(arg, 1));
  Assign(curr_thread->exceptfds, Field(arg, 2));
  date = Double_val(Field(arg, 3));
  if (date >= 0.0) {
    date += timeofday();
    Assign(curr_thread->delay, copy_double(date));
    curr_thread->status = BLOCKED_IO | BLOCKED_DELAY;
  } else {
    curr_thread->status = BLOCKED_IO;
  }
  if(callback_depth==0 && critical_section==Val_false)
    return schedule_thread();
  else {
    fd_set readfds, writefds, exceptfds;
    double delay, now;
    int need_select, need_wait;

    fprintf(stderr,"Warning: blocked in callback\n");

  select_try_again:
    /* Build fdsets and delay for select.
    See if some join or wait operations succeeded. */
    FD_ZERO(&readfds);
    FD_ZERO(&writefds);
    FD_ZERO(&exceptfds);
    delay = DELAY_INFTY;
    now = -1.0;
    need_select = 0;
    need_wait = 0;

    add_fdlist_to_set(curr_thread->readfds, &readfds);
    add_fdlist_to_set(curr_thread->writefds, &writefds);
    add_fdlist_to_set(curr_thread->exceptfds, &exceptfds);
    need_select = 1;

    if (curr_thread->status & (BLOCKED_DELAY - 1)) {
      double th_delay;
      if (now < 0.0) now = timeofday();
      th_delay = Double_val(curr_thread->delay) - now;
      if (th_delay <= 0) {
        curr_thread->status = RUNNABLE;
        curr_thread->retval = RESUMED_DELAY;
      } else {
        if (th_delay < delay) delay = th_delay;
      }
    }
    if (curr_thread->status != RUNNABLE){
      struct timeval delay_tv, * delay_ptr;
      int retcode;

      /* Convert delay to a timeval */
      if (delay != DELAY_INFTY) {
        delay_tv.tv_sec = (unsigned int) delay;
        delay_tv.tv_usec = (delay - (double) delay_tv.tv_sec) * 1E6;
        delay_ptr = &delay_tv;
      }
      else {
        delay_ptr = NULL;
      }

      enter_blocking_section();
      retcode = select(FD_SETSIZE, &readfds, &writefds, &exceptfds, delay_ptr);
      leave_blocking_section();

      if (retcode > 0) {
        /* Some descriptors are ready. 
        Mark the corresponding threads runnable. */
        value r = Val_unit, w = Val_unit, e = Val_unit;
        Begin_roots3(r,w,e);
        r = inter_fdlist_set(curr_thread->readfds, &readfds);
        w = inter_fdlist_set(curr_thread->writefds, &writefds);
        e = inter_fdlist_set(curr_thread->exceptfds, &exceptfds);
        if (r != NO_FDS || w != NO_FDS || e != NO_FDS) {
          value retval = alloc(3, TAG_RESUMED_IO);
          Field(retval, 0) = r;
          Field(retval, 1) = w;
          Field(retval, 2) = e;
          Assign(curr_thread->retval, retval);
          curr_thread->status = RUNNABLE;
        }
        End_roots();
      }
      /* If we get here with run_thread still NULL, one of the following
      may have happened:
      - a delay has expired
       - a wait() needs to be polled again
       - the select() failed (e.g. was interrupted)
       In these cases, we go through the loop once more to make the
       corresponding threads runnable. */
      if (curr_thread->status!= RUNNABLE &&
          (delay != DELAY_INFTY || retcode == -1))
        goto select_try_again;
    }

    /* If we haven't something to run at that point, we're in big trouble. */
    if (curr_thread->status != RUNNABLE)
      invalid_argument("Thread: deadlock");

    /* Free everything the thread was waiting on */
    Assign(curr_thread->readfds, NO_FDS);
    Assign(curr_thread->writefds, NO_FDS);
    Assign(curr_thread->exceptfds, NO_FDS);
    Assign(curr_thread->delay, NO_DELAY);
    Assign(curr_thread->joining, NO_JOINING);
    curr_thread->waitpid = NO_WAITPID;
    
    return curr_thread->retval;
  }
}

/* Primitives to implement suspension on buffered channels */

value thread_inchan_ready(value vchan) /* ML */
{
  struct channel * chan = Channel(vchan);
  return Val_bool(chan->curr < chan->max);
}

value thread_outchan_ready(value vchan, value vsize) /* ML */
{
  struct channel * chan = Channel(vchan);
  long size = Long_val(vsize);
  /* Negative size means we want to flush the buffer entirely */
  if (size < 0) {
    return Val_bool(chan->curr == chan->buff);
  } else {
    int free = chan->end - chan->curr;
    if (chan->curr == chan->buff)
      return Val_bool(size < free);
    else
      return Val_bool(size <= free);
  }
}

/* Suspend the current thread for some time */

value thread_delay(value time)          /* ML */
{
  double date = timeofday() + Double_val(time);
  Assert(curr_thread != NULL);
  check_callback("thread_delay");
  curr_thread->status = BLOCKED_DELAY;
  Assign(curr_thread->delay, copy_double(date));
  return schedule_thread();
}

/* Suspend the current thread until another thread terminates */

value thread_join(value th)          /* ML */
{
  check_callback("thread_join");
  Assert(curr_thread != NULL);
  if (((thread_t)th)->status == KILLED) return Val_unit;
  curr_thread->status = BLOCKED_JOIN;
  Assign(curr_thread->joining, th);
  return schedule_thread();
}

/* Suspend the current thread until a Unix process exits */

value thread_wait_pid(value pid)          /* ML */
{
  Assert(curr_thread != NULL);
  check_callback("thread_wait_pid");
  curr_thread->status = BLOCKED_WAIT;
  curr_thread->waitpid = pid;
  return schedule_thread();
}

/* Suspend the current thread until a major GC finishes */

value thread_wait_gc(value unit)          /* ML */
{
  Assert(curr_thread != NULL);
  check_callback("thread_wait_gc");
  curr_thread->status = BLOCKED_GC;
  return schedule_thread();
}

/* Reactivate another thread */

value thread_wakeup(value thread)     /* ML */
{
  thread_t th = (thread_t) thread;
  switch (th->status) {
  case SUSPENDED:
    th->status = RUNNABLE;
    th->retval = RESUMED_WAKEUP;
    break;
  case KILLED:
    failwith("Thread.wakeup: killed thread");
  case RUNNABLE:
  case BLOCKED_IO:
    /*DPRINT("Thread.wakeup: thread not suspended");*/
    break;
  default:
    
    /*failwith("Thread.wakeup: thread not suspended");*/
  }
  return Val_unit;
}

/* Return the current thread */

value thread_self(value unit)         /* ML */
{
  Assert(curr_thread != NULL);
  return (value) curr_thread;
}

/* Kill a thread */

value thread_kill(value thread)       /* ML */
{
  value retval = Val_unit;
  thread_t th = (thread_t) thread;
  /* Don't paint ourselves in a corner */

  check_runqueue();
  
  if(critical_section==Val_true){
    Begin_root(th);
    set_critical_section(copy_string("thread_kill"),Val_false);
    End_roots();
  }
  
  if(th->sync!=Val_unit){
    /* we are exiting from a synchronous call ... */
    DPRINT1("exiting from a synchronous call ...%ld",Int_val(th->sync));
    if(th != curr_thread) 
      failwith("Exit sync: th<>curr_thread");
    th->stack_high = stack_high;
    extern_sp = ((value*)(th->stack_high)) - Int_val(th->sync);
    DPRINT1("NEW TRAP : %d\n", Int_val(extern_sp[-2]));
    trapsp =((value*)(th->stack_high)) - Int_val(extern_sp[-2]);
    th->sync = extern_sp[-1];
    DPRINT1("Return %ld",th->retval);
    return th->retval;
  }
  DPRINT("killing the thread ...");
  if (th == th->next) failwith("Thread.kill: cannot kill the last thread");
  /* This thread is no longer waiting on anything */
  if (th->status == KILLED) failwith("Thread.kill: killed thread");
  th->status = KILLED;

  /*DPRINT("THREAD KILL\n");*/

  
  if(terminaison){
    /* at the end, there are three threads: the main one, one waiting
    on the port, and one for migrations */
    nbre_threads--;
    if(nbre_threads<nthreads+1){
      if(th->next->status & (BLOCKED_END-1)){
	/* printf("LAUNCHING MASTER THREAD\n"); */
	th->next->status = RUNNABLE;
      } else
	if(th->next->next->status & (BLOCKED_END-1)){
	  /* printf("LAUNCHING MASTER THREAD\n"); */
	  th->next->next->status = RUNNABLE;
	} else
	  fprintf(stderr,"No threads waiting\n");
    }
  }

  /* If this is the current thread, activate another one */
  if (th == curr_thread){
    Begin_root(th);
    retval = schedule_thread();
    End_roots();
  }
  /* Remove thread from the doubly-linked list */
  Assign(th->prev->next, th->next);
  Assign(th->next->prev, th->prev);
  /* Free its resources */

  if(th->stack_high == th->stack_low + Thread_stack_size / sizeof(value)){
    /* is there a free slot in the buffer for this stack ? */
    if(first_stack_to_use>0){
      /*DPRINT("Saving stack\n");*/
      stacks_buffer[--first_stack_to_use] = th->stack_low;
    } else 
      if (next_stack_free<STACKS_BUFFER_SIZE){
	/* DPRINT("Saving stack\n"); */
	stacks_buffer[next_stack_free++] = th->stack_low;
      } else {
	/* we must free this stack */
	/* DPRINT("Freeing thread stack\n"); */
	stat_free((char *) th->stack_low);
      }
  } else {
    /*    DPRINT("Freeing thread stack\n");*/
    stat_free((char *) th->stack_low);
  }
 
  th->stack_low = NULL;
  th->stack_high = NULL;
  th->stack_threshold = NULL;
  th->sp = NULL;
  th->trapsp = NULL;

  check_runqueue();

  if(th->stub!=Val_unit) {
    if(th->location != Val_unit && th->reused==Val_unit){
    /* ce thread est un bon candidat pour etre sauvegarde pour utilisation
       ulterieure */
      value lloc = Field(th->location,STUB_LOCAL);
      value reused = Field(lloc,LOCATION_THREADS_REUSED);
      int next = Int_val(Field(lloc,LOCATION_THREAD_NEXT));
      
      if(reused == Val_unit){
        int i;
	Begin_roots2(th,lloc);
	
	/* il faut alloue le tableau */
	reused = alloc(MAX_THREADS_REUSED,0);
	Assign(Field(lloc,LOCATION_THREADS_REUSED), reused);
	for(i=0;i<MAX_THREADS_REUSED;i++)
	  Field(reused,i) = Val_unit;

	End_roots();
      }
      
      if(next == MAX_THREADS_REUSED)
	/* pas de place */
	Assign(Field(th->stub,STUB_LOCAL),th->stub);
      else {
	Assign(Field(reused,next),th);
	Field(lloc,LOCATION_THREAD_NEXT) = Val_int(next+1);
      }
      
    } else
    /* sinon, on peut toujours se contenter de supprimer le pointeur 
	 depuis son stub pour qu'il soit un peu collecter */
      Assign(Field(th->stub,STUB_LOCAL),Val_unit);
  }
  
  return retval;
}

/* Set a list of file descriptors in a fdset */

static void add_fdlist_to_set(value fdl, fd_set *set)
{
  for (/*nothing*/; fdl != NO_FDS; fdl = Field(fdl, 1)) {
    FD_SET(Int_val(Field(fdl, 0)), set);
  }
}

/* Build the intersection of a list and a fdset (the list of file descriptors
   which are both in the list and in the fdset). */

static value inter_fdlist_set(value fdl, fd_set *set)
{
  value res = Val_unit;
  value cons;

  Begin_roots2(fdl, res);
    for (res = NO_FDS; fdl != NO_FDS; fdl = Field(fdl, 1)) {
      int fd = Int_val(Field(fdl, 0));
      if (FD_ISSET(fd, set)) {
        cons = alloc(2, 0);
	Field(cons, 0) = Val_int(fd);
	Field(cons, 1) = res;
	res = cons;
	FD_CLR(fd, set); /* wake up only one thread per fd ready */
      }
    }
  End_roots();
  return res;
}

/* Auxiliary function for allocating the result of a waitpid() call */

#if !(defined(WIFEXITED) && defined(WEXITSTATUS) && defined(WIFSTOPPED) && \
      defined(WSTOPSIG) && defined(WTERMSIG))
/* Assume old-style V7 status word */
#define WIFEXITED(status) (((status) & 0xFF) == 0)
#define WEXITSTATUS(status) (((status) >> 8) & 0xFF)
#define WIFSTOPPED(status) (((status) & 0xFF) == 0xFF)
#define WSTOPSIG(status) (((status) >> 8) & 0xFF)
#define WTERMSIG(status) ((status) & 0x3F)
#endif

#define TAG_WEXITED 0
#define TAG_WSIGNALED 1
#define TAG_WSTOPPED 2

static value alloc_process_status(int pid, int status)
{
  value st, res;

  if (WIFEXITED(status)) {
    st = alloc(1, TAG_WEXITED);
    Field(st, 0) = Val_int(WEXITSTATUS(status));
  }
  else if (WIFSTOPPED(status)) {
    st = alloc(1, TAG_WSTOPPED);
    Field(st, 0) = Val_int(WSTOPSIG(status));
  }
  else {
    st = alloc(1, TAG_WSIGNALED);
    Field(st, 0) = Val_int(WTERMSIG(status));
  }
  Begin_root(st);
    res = alloc(2, TAG_RESUMED_WAIT);
    Field(res, 0) = Val_int(pid);
    Field(res, 1) = st;
  End_roots();
  return res;
}

value suspend_on_terminaison(value nth)  /* ML */
{
  nthreads = Int_val(nth);
  if(terminaison){
    if(nbre_threads<nthreads+1) return Val_unit;
    check_callback("suspend_on_terminaison");
    curr_thread->status = BLOCKED_END;
    return schedule_thread();
  }
  return Val_unit;
}

#include <errno.h>
#ifndef EAGAIN
#define EAGAIN (-1)
#endif
#ifndef EINTR
#define EINTR (-1)
#endif
#ifndef EWOULDBLOCK
#define EWOULDBLOCK (-1)
#endif

value try_again(value err)
{
  fprintf(stderr,"try again (errno = %d) EAGAIN(%d) EINTR(%d) EWOULDBLOCK(%d) ? \n",Int_val(err),EAGAIN,EINTR,EWOULDBLOCK);
  if(err == Val_int(EAGAIN) || err == Val_int(EINTR) || 
     err == Val_int(EWOULDBLOCK)) {
    fprintf(stderr,"yeeeeh, try again ... \n");
    return Val_true;
  }
  
  return Val_false;
}

/* prevent a thread structure from being recycled */

value dirty_thread(value thread) /* ML */
{
  thread_t th = (thread_t) th;
  
  th->reused = Val_false;
  return Val_unit;
}




