#ifndef THREADS_H
#define THREADS_H

#include "stub.h"

/* The thread descriptors */
typedef struct thread_struct    *thread_t;

struct thread_struct {
  value ident;                  /* Unique id (for equality comparisons) */
  struct thread_struct * next;  /* Double linking of threads */
  struct thread_struct * prev;
  value * stack_low;            /* The execution stack for this thread */
  value * stack_high;
  value * stack_threshold;
  value * sp;
  value * trapsp;
  value status;                 /* RUNNABLE, KILLED. etc (see below) */
  value readfds, writefds, exceptfds;
                /* Lists of file descriptors on which we're doing select() */
  value delay;                  /* Time until which this thread is blocked */
  value joining;                /* Thread we're trying to join */
  value waitpid;                /* PID of process we're waiting for */
  value retval;                 /* Value to return when thread resumes */

  value location;           /* Location to which this thread belongs */
  value stub;                   /* A stub for this thread */
  value reused;                 /* offset from the tail of the stack
				   to find the next frame */
  value sync;                   /* sync > 0 means that this thread is used 
				   for a synchronous call in the same 
				   location */
  value passe;              /* to minimize scheduler work */
};



/* Initial size of stack when a thread is created (4 Ko) */
#define Thread_stack_size (Stack_size / 4)

/* Max computation time before rescheduling, in microseconds (50ms) */
#define Thread_timeout 50000

#define RUNNABLE Val_int(0)
#define KILLED Val_int(1)
#define SUSPENDED Val_int(2)
#define BLOCKED_IO Val_int(4)
#define BLOCKED_DELAY Val_int(8)
#define BLOCKED_JOIN Val_int(16)
#define BLOCKED_WAIT Val_int(32)

#define BLOCKED_GC Val_int(64)
#define BLOCKED_END Val_int(128)
#define BLOCKED_REPLY Val_int(256)
#define BLOCKED_FAIL Val_int(512)

#define BLOCKED_ANY ((BLOCKED_IO-1) + (BLOCKED_DELAY - 1)\
   + (BLOCKED_JOIN - 1) + (BLOCKED_WAIT - 1)\
   + (BLOCKED_GC - 1) + (BLOCKED_FAIL-1))

#define RESUMED_WAKEUP Val_int(0)
#define RESUMED_DELAY Val_int(1)
#define RESUMED_JOIN Val_int(2)

#define TAG_RESUMED_IO 0
#define TAG_RESUMED_WAIT 1

#define NO_FDS Val_unit
#define NO_DELAY Val_unit
#define NO_JOINING Val_unit
#define NO_WAITPID Val_int(0)

#define DELAY_INFTY 1E30        /* +infty, for this purpose */

#ifdef DEBUG
#  define check_runqueue()\
{\
  thread_t th = curr_thread;\
  \
  Assert(Is_in_heap(th));\
  th = th->next;\
  while(th != curr_thread){\
    Assert(Is_in_heap(th));\
    th = th->next;\
    }\
}
#else
#  define check_runqueue()
#endif

/* Since context switching is not allowed in callbacks, a thread that
   blocks during a callback is a deadlock. */
    /*    fatal_error("Thread: deadlock during callback");*/           

#define check_callback(msg) \
  if (callback_depth > 0){  \
     fprintf(stderr, "scheduler exited: callback depth : %d in %s \n",callback_depth,msg);\
	fatal_error("Thread: deadlock during callback");\
  }


extern value next_ident;

#define STACKS_BUFFER_SIZE 90
/*
extern value *stacks_buffer[STACKS_BUFFER_SIZE];
extern int next_stack_free;
extern int first_stack_to_use;
*/
extern void free_stack(thread_t);
extern value *get_stack();
extern value stack_allouees;
extern int nbre_threads;
extern int terminaison; /* we detect terminaison */

typedef struct location_stub    *location_t;
typedef struct local_loc_struct *local_location_t;
/* Types.local_location */
#define LOCATION_FATHER  0
#define LOCATION_DEFS    1
#define LOCATION_LOCS    2
#define LOCATION_MODULE  3
#define LOCATION_ATTACH  4
#define LOCATION_FREE_STUBS      5
#define LOCATION_FREE_STUBS_SIZE 6
#define LOCATION_NEXT_FREE_STUB  7
#define LOCATION_SIZE    8

#define MAX_STUBS_REUSED 100

struct local_loc_struct {
  location_t father;            /* the location in which is is */
  value      def_weakset;       /* set of contained automatas */
  value      loc_weakset;       /* set of contained locations */
  value      module;            /* inutile a priori */
  value      attach;            /* reloc attach */
  value      free_stubs;
  value      free_stubs_size;
  value      next_free_stub;
};



/* byterun/obj.c */
value obj_block(value tag, value size);

/* scheduler.c */
extern thread_t curr_thread;
extern value local_thread_reuse(value location,value clos,value args
			 ,thread_t thread);

extern value extract_location_threads (value location_v);
extern value gen_thread_new(value location,value clos,value param,value clos_arg);
extern value schedule_go_thread(value);
extern value set_critical_section(value);
extern value schedule_thread();
extern value thread_kill(value th);
extern void remove_thread(thread_t th);


/* local.c */
extern value get_thread_stub(thread_t thread);
extern value trigger_def (value args);

#endif
