open Obj

external byte_code_start : unit -> unit = "mixte_byte_start"
external c_import_module : string -> t = "c_import_module"
external is_mixte_value : t -> bool = "is_mixte_value"
external register_named_value: string -> t -> unit = "register_named_value"

type 'a return =
    Value of 'a
  | Exception of exn
  | ExnFailure of string
  | ExnOut_of_memory 
  | ExnSys_error of int * string
  | ExnEnd_of_file
  | ExnDivision_by_zero
  | ExnNot_found
  | ExnMatch_failure of (string * int * int)
  | ExnAssert_failure of (string * int * int)
  | ExnInvalid_argument of string
  | ExnStack_overflow
  | ExnExit

external c_callback : t -> t -> t return = "mixte_callback"

let rec wrappe f =
  if is_mixte_value f then f
  else
    begin
      repr (function x -> 
        match c_callback f x with
          Value v -> wrappe v
        | Exception e -> raise e
        | ExnFailure s -> failwith s
        | ExnOut_of_memory -> raise Out_of_memory
        | ExnSys_error (e,s) -> raise (Sys_error (e,s))
        | ExnEnd_of_file -> raise End_of_file
        | ExnDivision_by_zero -> raise Division_by_zero
        | ExnNot_found -> raise Not_found
        | ExnMatch_failure (s,i1,i2) -> raise (Match_failure (s,i1,i2))
        | ExnAssert_failure (s,i1,i2) -> raise (Assert_failure (s,i1,i2))
        | ExnInvalid_argument s -> invalid_arg s
        | ExnStack_overflow -> raise Stack_overflow
        | ExnExit -> raise Exit
             )
    end
let callback clos x =
  try
    Value (clos x)
  with
  | Failure s -> ExnFailure s
  | Out_of_memory -> ExnOut_of_memory
  | Sys_error (e,s) -> ExnSys_error (e,s)
  | End_of_file -> ExnEnd_of_file
  | Division_by_zero -> ExnDivision_by_zero
  | Not_found -> ExnNot_found
  | Match_failure (s,i1,i2) -> ExnMatch_failure (s,i1,i2)
  | Assert_failure (s,i1,i2) -> ExnAssert_failure (s,i1,i2)
  | Invalid_argument s -> ExnInvalid_argument s
  | Stack_overflow -> ExnStack_overflow
  | Exit -> ExnExit
  | e -> Exception e

let callback_and_wrappe clos f = callback clos (wrappe f)

let init = ref true

let export_module module_name m =
  register_named_value ("module+"^module_name) m;
  if !init then 
    begin at_exit byte_code_start; init := false end

let soft_wrappe module_name i f =
  if is_mixte_value f then f
  else
    begin
      repr (function x -> 
        let m = c_import_module ("module+"^module_name)
        in
        match c_callback (field m i) x with
          Value v -> wrappe v
        | Exception e -> raise e
        | ExnFailure s -> failwith s
        | ExnOut_of_memory -> raise Out_of_memory
        | ExnSys_error (e,s) -> raise (Sys_error (e,s))
        | ExnEnd_of_file -> raise End_of_file
        | ExnDivision_by_zero -> raise Division_by_zero
        | ExnNot_found -> raise Not_found
        | ExnMatch_failure (s,i1,i2) -> raise (Match_failure (s,i1,i2))
        | ExnAssert_failure  (s,i1,i2) -> raise (Assert_failure (s,i1,i2))
        | ExnInvalid_argument s -> invalid_arg s
        | ExnStack_overflow -> raise Stack_overflow
        | ExnExit -> raise Exit
              )
    end

let import_module module_name =
  let m = c_import_module ("module+"^module_name)
  in
  let len = size m in
  let block = new_block Tag.interface len in
    for i = 0 to len - 1 do
      set_field block i (soft_wrappe module_name i (field m i))
    done;
    block

external mixte_register : (('a -> 'b) -> 'a -> 'b return) ->
  (('a -> 'b) -> 'a -> 'b return) -> unit = "mixte_register"

let _ = 
  mixte_register callback callback_and_wrappe
