(* pas de types polymorphes pour l'instant *)

type mtype_fullname = string
type mtype_kind =
    MTrecord
  | MTconcret
  | MTabstract
 
type mtype_name =
    MType_int
  | MType_string
  | MType_float
  | MType_bool
  | MType_unit
  | MType_array
  | MType_list
  | MType_metatype
  | MType_name of mtype_fullname * mtype_kind
  
type mtype_desc =
    MTnil
  | MTvar of int
  | MTchannel of mtype_expr * (mtype_expr option)
  | MTarrow of mtype_expr * mtype_expr
  | MTtuple of mtype_expr list
  | MTconstr of mtype_name * mtype_expr list
  | MTobject of (string * mtype_expr) list
and mtype_expr = mtype_desc ref

let meta_to_type meta =
  let (expr,_) = Marshal.from_string meta 0
  in
    (Obj.magic expr: mtype_expr)

let alias expr =
  let aliases = ref []
  and counter = ref 0
  in
  let rec iter (expr:mtype_expr) =
    try
      let alias = List.assq !expr !aliases
      in
	match !alias with
	    None -> 
	      alias := Some !counter;
	      incr counter
	  | Some _ -> ()
    with
	Not_found ->
	  aliases := (!expr,ref None):: !aliases;
	  match !expr with
	    MTnil -> ()
          | MTvar i -> ()
          | MTarrow (e1,e2) -> iter e1; iter e2
          | MTchannel(e1,None) -> iter e1 
          | MTchannel(e1,Some e2) -> iter e1; iter e2  
	  | MTtuple elist -> List.iter iter elist 
	  | MTconstr (_,elist) -> List.iter iter elist 
	  | MTobject list -> List.iter (function (_,e) -> iter e) list
  in
    iter expr;
    List.fold_left (function list -> function ((desc:mtype_desc),alias) -> 
		      match !alias with
			  None -> list
			| Some v -> (desc,(v,ref false))::list) [] !aliases

let print_type (mtype:mtype_expr) =
  let aliases = alias mtype
  in
  let rec iter expr =
    try
      let (v,p) = List.assq !expr aliases
      in
	if !p then
	  begin
	    print_string "'";
	    print_int v
	  end
    	else
	  begin
	    print_char '(';
	    print expr;
	    print_string ") as '";
	    print_int v;
	    p := true
	  end
    with
	Not_found ->
	  print expr
  and print_list n sep elist = 
    match elist with
      [] -> ()
    | hd::tl ->
        if n then print_string sep;
        print hd;
        print_list true sep tl  
    
  and print (expr:mtype_expr) =
    match !expr with
	MTnil -> ()
    | MTvar i -> print_string  "'a"; print_int i
    | MTarrow (e1,e2) -> 
	print_char '(';
	iter e1; 
	print_string " -> ";
	iter e2;
	  print_char ')'
    | MTchannel (e,None) ->
        print_string "<<";
        print e;
        print_string ">>";
    | MTchannel (e1,Some e2) ->
	print_char '(';
	iter e1; 
	print_string " => ";
	iter e2;
	print_char ')'
          
    | MTtuple elist -> 
	  print_char '(';
          print_list false " * " elist;
	  print_char ')'
      | MTconstr (name,elist) -> 
          begin
            match elist with
              [] -> ()
            | [hd] -> print hd
            | elist ->
	        print_char '(';
                print_list false "," elist;
	        print_char  ')';
          end;
          print_char ' ';
	  begin
	    match name with
		MType_int -> print_string "int"
	      | MType_string -> print_string "string"
	      | MType_float -> print_string "float"
	      | MType_bool -> print_string "bool"
	      | MType_unit -> print_string "unit"
	      | MType_array -> print_string "array"
	      | MType_list -> print_string "list"
              | MType_metatype -> print_string "metatype"
	      | MType_name (name,kind) -> 
		  begin
		    match kind with
			MTabstract -> print_string "abstract "
		      | MTconcret -> print_string "concret "
		      | MTrecord -> print_string "record "
		  end;
		  print_string name
	  end;
      | MTobject list ->
	  print_char '<';
	  List.iter (function (s,e) -> 
		       print_string (s^":");
		       iter e; print_string " ; ") list;
	  print_char '>'
  in
    iter mtype;
    print_newline ()


exception NotSubtype

let le exp1 exp2 =
  let env = ref []
  in
  let rec le exp1 exp2 =
    let constraints =
      try
      	List.assq !exp1 !env
      with
	  Not_found -> 
	    let c = ref []
	    in
	      env := (!exp1,c):: !env;
	      c
    in
      if not (List.memq !exp2 !constraints) then
	begin
	  constraints := !exp2 :: !constraints;
	  subtype exp1 exp2
	end
  and occur_check exp1 exp2 = 
    if exp1 == exp2 then raise NotSubtype;
    match !exp2 with
      MTnil -> ()
    | MTvar _ -> ()
    | MTchannel (e1,None) -> 
        occur_check exp1 e1
    | MTchannel (e1,Some e2) ->
        occur_check exp1 e1;
        occur_check exp1 e2
    | MTarrow (e1,e2) ->
        occur_check exp1 e1;
        occur_check exp1 e2
    | MTtuple elist ->
        List.iter (occur_check exp1)  elist
    | MTconstr (_,elist) ->
        List.iter (occur_check exp1) elist
    | MTobject elist ->
        List.iter (fun (_,e) -> occur_check exp1 e) elist

  and subtype exp1 exp2 =
    if !exp1 == !exp2 then () else
    match (!exp1,!exp2) with
      (MTnil,MTnil) -> ()
    | (MTvar i,ty) -> 
        occur_check exp1 exp2;
        exp1 := ty
    | (ty,MTvar i) -> 
        occur_check exp2 exp1;
        exp2 := ty
    | (MTarrow(left1,right1),MTarrow(left2,right2)) ->
	le left2 left1; le right1 right2
    | (MTarrow(left1,right1),MTchannel(left2,Some right2)) ->
	le left2 left1; le right1 right2
    | (MTtuple list1,MTtuple list2) ->
	List.iter 
	  (function (exp1,exp2) ->
	    le exp1 exp2) (List.combine list1 list2)
    | (MTchannel(e1,None),MTchannel(e2,None)) ->
	le e1 e2
    | (MTchannel(list1,Some right1),MTchannel(list2,Some right2)) ->
	le list1 list2;
        le right1 right2
    | (MTconstr (name1,list1), MTconstr (name2,list2)) -> 
	if name1 <> name2 then raise NotSubtype
	List.iter 
	  (function (exp1,exp2) ->
	    le exp1 exp2) (List.combine list1 list2);
	List.iter 
	  (function (exp1,exp2) ->
	    le exp1 exp2) (List.combine list2 list1)	  
    | (MTobject list1,MTobject list2) ->
	List.fold_left 
	  (function list1 ->
	    function (meth2,exp2) ->
	      let rec iter =
		function [] -> 
		  raise NotSubtype
		  | (meth1,exp1)::list1 ->
		      if meth1 = meth2 then
			begin
			  le exp1 exp2;
			  list1
			end
		      else
			iter list1
	      in
	      iter list1) list1 list2;
	()
    | _ -> raise NotSubtype
  in
  try 
    le exp1 exp2; true
  with
    NotSubtype -> false


let coerce (v:'a) (v':'a metatype) = (v':'a metatype)
let translate  (v:'a metatype) = 
  (Marshal.from_string (Obj.magic v) 0 : mtype_expr)
