(***********************************************************************)
(*                                                                     *)
(*                           Objective Caml                            *)
(*                                                                     *)
(*         Jerome Vouillon, projet Cristal, INRIA Rocquencourt         *)
(*                                                                     *)
(*  Copyright 1996 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

(* $Id: mop.mli,v 1.1.1.1 1998/04/20 11:32:18 lefessan Exp $ *)

(* Module [Oo]: object-oriented extension *)


type class_id = Class_id of string * int


val copy : (< .. > as 'a) -> 'a
        (* [Oo.copy o] returns a copy of object [o], that is a fresh
           object with the same methods and instance variables as [o]  *)

(*--*)

(*** For system use only, not for the casual user *)

(* Methods *)
type label
val new_method: string -> label

type t
type item
type bucket
type vars
type obj_init
type meths
type 'a option

type table =
 { mutable messages: (string*int*int) list;
   mutable buckets: bucket array;
   mutable ext_buckets: bucket array;
   mutable methods: meths;
   mutable saved_methods: meths list;
   mutable meth_defs: (label * item) list;
   mutable vars: vars;
   mutable saved_vars: vars list;
   mutable saved_var_lst: int option list;
   mutable size: int;
   mutable init: obj_init list list }

type class_info = {mutable obj_init: t;
   mutable class_init: table -> unit;
   mutable table: table;
   mutable class_sig: (string*int*int) list;
   mutable class_type: class_id;
   mutable methodes : (string,label) Hashtbl.t
  }

val set_initializer: table -> obj_init -> unit
val inheritance: table -> class_info -> string list -> string list -> unit
val get_method: table -> label -> item
val set_method: table -> string -> int -> item -> unit
val get_method_label: table -> string -> label
val get_variable: table -> string -> int
val hide_variable: table -> string -> unit
val get_private_variable: table -> string -> int
val oo_create_class:
        class_info -> string list -> (table -> t) ->
        (table -> unit) -> class_id -> class_info
val create_class:
        class_info -> string list -> (table -> t) ->
        (table -> unit) -> Reloc.module_data -> class_info
val get_class_info: 'a -> class_info
(* Objects *)
type object
val create_object: table -> t

val lookup :   object -> int -> item

val create_ext_buckets :
    (string * int * int) list -> (int -> string -> int -> item) -> 
      bucket array

(* Parameters *)
type params = {
    mutable compact_table : bool;
    mutable copy_parent : bool;
    mutable clean_when_copying : bool;
    mutable retry_count : int;
    mutable bucket_small_size : int
  } 

val params : params

(* Statistics
type stats =
  { classes: int; labels: int; methods: int; inst_vars: int; buckets: int;
    distrib : int array; small_bucket_count: int; small_bucket_max: int }
val stats: unit -> stats
val show_buckets: unit -> unit
 *)

val invoke_methode : item -> 'a -> 'c list -> Obj.t 
val wrappe_methode : ('a -> 'b list -> 'c) -> int -> 'a -> 'c
