(***********************************************************************)
(*                                                                     *)
(*                           Objective Caml                            *)
(*                                                                     *)
(*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         *)
(*                                                                     *)
(*  Copyright 1996 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

(* $Id: obj.ml,v 1.2 1998/05/07 15:38:07 lefessan Exp $ *)

(* Operations on internal representations of values *)

type t

external repr : 'a -> t = "%identity"
external obj : t -> 'a = "%identity"
external magic : 'a -> 'b = "%identity"
external is_block : t -> bool = "obj_is_block"
external tag : t -> int = "obj_tag"
external size : t -> int = "%obj_size"
external field : t -> int -> t = "%obj_field"
external set_field : t -> int -> t -> unit = "%obj_set_field"
external new_block : int -> int -> t = "obj_block"

(*
let marshal (obj: t) =
  Marshal.to_string obj []
let unmarshal str pos =
  (Marshal.from_string str pos, pos + Marshal.total_size str pos)
*)

let create_block tag tab =
  let tab = repr tab in
  let len = size tab in
  let block = new_block tag len in
    for i = 0 to len - 1 do
      set_field block i (field tab i)
    done;
    block

let rec print_spaces offset =
  for i = 1 to offset do
    print_char ' '
  done

let rec univ_print offset x =
  if is_block x then
    if offset > 40 then
      print_string "<block>"
    else
      begin
      print_spaces offset;
      print_char '[';
      print_int (tag x);
      print_char ']';
      print_char '(';
      print_newline ();
      for i = 0 to size x - 1 do
        univ_print (offset + 2) (field x i);
      done;
      print_spaces offset;
      print_char ')';
    end
  else
    begin
      print_spaces offset;
      print_int (magic x)
    end;
  print_newline()

let print x =
  univ_print 0 (repr x)
