(***********************************************************************)
(*                                                                     *)
(*                    ______________                                   *)
(*                                                                     *)
(*      Fabrice Le Fessant, projet SOR/PARA INRIA Rocquencourt         *)
(*                                                                     *)
(*                 mail: fabrice.le_fessant@inria.fr                   *)
(*                 web:  http://www-sor.inria.fr/~lefessan             *)
(*                                                                     *)
(*  Copyright 1997 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(*                                                                     *)
(*         File written in the Objective-CAML language                 *)
(***********************************************************************)

open Misc
open Config

open Emitcode

let topo = Topo.create ()
let ids = ref []

let process_file file_name =
  let ic = open_in_bin file_name in
    try
      let buffer = String.create (String.length cmo_magic_number) in
        really_input ic buffer 0 (String.length cmo_magic_number);
        if buffer = cmo_magic_number then
          begin
            let compunit_pos = input_binary_int ic in  (* Go to descriptor *)
              seek_in ic compunit_pos;
              let compunit = (input_value ic : compilation_unit) in
                close_in ic;
                ids := file_name :: !ids;
                Topo.add topo file_name;
                List.iter (fun (name,_) -> 
                             let s = name ^ ".cmo" in
                               s.[0] <- Char.lowercase s.[0];
                               if s <> file_name then
                                 Topo.inf topo s file_name
                                 )
                  compunit.cu_imports;
          end
        else
          failwith "Not an object"
    with x -> close_in ic; raise x

let rec print_list =
  function
      [] -> ()
    | s :: reste ->
        print_string s; print_char ' ';
        print_list reste

let _ = 
  Arg.parse [] process_file "";
  match Topo.order topo with
    Topo.Sorted ->
      print_list (Topo.sort_list topo !ids);
      print_newline ()
  | Topo.Cycle list ->
      printnl_string "Cyclic dependency :";
      print_list list;
      print_newline ()
      
