open Metatype
open Types
open Btype
open Ctype
open Typemisc;;


let rec translate_path = function
    Path.Pident ident -> Ident.name ident
  | Path.Pdot(p, s, n) -> (translate_path p)^"."^s
  | Path.Papply(p1, p2) ->
      (translate_path p1)^"("^(translate_path p2)^")"


exception Found of mtype_name
let aliases = ref []
let var = ref 0

let rec type_to_meta env pexp_loc ty =
    try
      List.assq ty !aliases
    with
        Not_found ->
          let expr = ref MTnil
          in
            aliases := (ty,expr):: !aliases;
            expr := desc_to_desc ty.desc env pexp_loc;
            (expr:mtype_expr)

and desc_to_desc desc env pexp_loc =
  match desc with
      Tvar -> incr var; MTvar !var
    | Tarrow (ty1,ty2) -> MTarrow (type_to_meta env pexp_loc ty1,
                                   type_to_meta env pexp_loc ty2)
    | Tchannel (tylist,r) ->
        MTchannel (type_to_meta env pexp_loc tylist,
                   match r with
                     Tasync -> None
                   | Tsync ty -> Some (type_to_meta env pexp_loc ty)
                         )
    | Ttuple tylist -> MTtuple (List.map (type_to_meta env pexp_loc) tylist)
    | Tconstr (path,tylist,_) -> 
        let decl = Env.find_type path env
        in
          (match decl.type_manifest with
              None -> 
                MTconstr(
                  (
                  try
                    List.iter (function (p,mt) ->
                                 if (Path.same path p) then
                                   raise (Found mt)
                    ) 
                      [
                        Predef.path_int,MType_int;
                        Predef.path_string,MType_string;
                        Predef.path_float,MType_float;
                        Predef.path_bool,MType_bool;
                        Predef.path_unit,MType_unit;
                        Predef.path_array,MType_array;
                        Predef.path_list,MType_list
                      ];
                    MType_name (translate_path path, 
                                match decl.type_kind with
                                    Type_abstract -> MTabstract
                                  | Type_variant _ -> MTconcret
                                  | Type_record _ -> MTrecord
                    )
                  with
                      Found mt -> mt
                ),List.map (type_to_meta env pexp_loc) tylist)
            | Some ty -> 
                !(type_to_meta env pexp_loc ty)
          )

    | Tobject (fields,_) -> 
        let (fields,rest) = flatten_fields fields
        in
          if rest.desc <> Tnil then
            failwith "type_to_meta : < fields; .. >";
          MTobject 
            (List.map 
               (function (methname,_,methtype) ->
                  (methname, type_to_meta env pexp_loc methtype)
               ) (Sort.list (function (m1,_,t1) -> function (m2,_,t2) ->
                              m1 <= m2) fields))
    | Tlink ty -> !(type_to_meta env pexp_loc ty)
    | _ -> failwith "type_to_meta : _ "

                
let to_meta env pexp_loc ty =
  aliases := [];
  var := 0;
  Location.print_warning pexp_loc "VARTYPE replaced by type ";
  let s =
    Marshal.to_string (Obj.repr (let mtype = type_to_meta env pexp_loc ty
    in
    print_type mtype;
    mtype
      )) []
  in
  s

